#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define GAMEFILE     1
#define GAMESTART    2
#define CDLFILE      3
#define CDLSTART     4
#define NOBYTES      5
#define OUTFILE      6
#define PRINTADDR    7
#define ORG          8

#define ARGSIZE      9

#define OUTBYTE      ".BYTE"
#define OUTBYTEL     ".byte"
#define OUTWORD      ".WORD"
#define OUTWORDL     ".word"

#define tolower(c)   ((c >= 'A') && (c <= 'Z'))

const char *bytetypes[] = {
".BYTE", ".WORD",
};

const char *formats[] = {
"\t%s $%02X\n", "\t%s $%04X\n",
};

const char title[] = "cdldis <game-file> <game-startaddr> \
<cdl-file> <cdl-startaddr> <# bytes> <out-file> <print-addr> <org>\n";

/*the opsize table is used to quickly grab the instruction sizes (in bytes)*/
const int  opsize[256] = {
/*0x00*/	1,2,0,0,0,2,2,0,1,2,1,0,0,3,3,0,
/*0x10*/	2,2,0,0,0,2,2,0,1,3,0,0,0,3,3,0,
/*0x20*/	3,2,0,0,2,2,2,0,1,2,1,0,3,3,3,0,
/*0x30*/	2,2,0,0,0,2,2,0,1,3,0,0,0,3,3,0,
/*0x40*/	1,2,0,0,0,2,2,0,1,2,1,0,3,3,3,0,
/*0x50*/	2,2,0,0,0,2,2,0,1,3,0,0,0,3,3,0,
/*0x60*/	1,2,0,0,0,2,2,0,1,2,1,0,3,3,3,0,
/*0x70*/	2,2,0,0,0,2,2,0,1,3,0,0,0,3,3,0,
/*0x80*/	0,2,0,0,2,2,2,0,1,0,1,0,3,3,3,0,
/*0x90*/	2,2,0,0,2,2,2,0,1,3,1,0,0,3,0,0,
/*0xA0*/	2,2,2,0,2,2,2,0,1,2,1,0,3,3,3,0,
/*0xB0*/	2,2,0,0,2,2,2,0,1,3,1,0,3,3,3,0,
/*0xC0*/	2,2,0,0,2,2,2,0,1,2,1,0,3,3,3,0,
/*0xD0*/	2,2,0,0,0,2,2,0,1,3,0,0,0,3,3,0,
/*0xE0*/	2,2,0,0,2,2,2,0,1,2,1,0,3,3,3,0,
/*0xF0*/	2,2,0,0,0,2,2,0,1,3,0,0,0,3,3,0
};

void *sortonthefly(void *ptr, size_t num, size_t size,
					int (*cp) (const void *, const void *)) {
size_t low = 0, high = num - 1, half, doing;
void *holder;
int test;
	while(high > low) {
		half = (low + high) / 2;
		test = (*cp)(ptr + size * (num-1), ptr + size * half);
		//onaji nara
		if(!test) {
			high = half;
			break;
		} else if(test<0)
			high = half;
		else
			low = half + 1;
	}
	holder = malloc(size);
	memcpy(holder, ptr + size * (num-1), size);
	for(doing = num - 1; doing > high; doing--)
        memcpy(ptr + size * doing, ptr + size * (doing - 1), size);
    memcpy(ptr + size * high, holder, size);
    free(holder);
    return ptr + size * high;
}

void lower(char *tochange, int numbytes) {
int i;
    for(i = 0; i < numbytes; i++) {
        if(!tochange[i]) break;
        if(tolower(tochange[i])) 
            tochange[i] -='A'-'a';
    }
}

int main(int argc, char *argv[]) {
    FILE *game = NULL, *cdl = NULL, *out = NULL;
    int size, gamest, cdlst, i, j, opcode[3], org, temp, tmp, cflag, badop;
    char str[30], chr[5];
    const char *ctemp, *ctemp2;
    if(argc < ARGSIZE) {
        printf("%s", title);
        return 0;
    }
    if(argc > ARGSIZE) {
        printf("%sToo many arguments!\n", title);
        return 1;
    }
    game = fopen(argv[GAMEFILE], "rb");
    if(game == NULL) {
        perror("Error opening file!\n");
        return 2;
    }
    printf("Loaded %s\n", argv[GAMEFILE]);
    cdl = fopen(argv[CDLFILE], "rb");
    if(cdl == NULL) {
        fclose(game);
        perror("Error opening file!\n");
        return 2;
    }
    printf("Loaded %s\n", argv[CDLFILE]);
    out = fopen(argv[OUTFILE], "wt");
    if(out == NULL) {
        fclose(game);
        fclose(cdl);
        perror("Error opening file!\n");
        return 2;
    }
    printf("Writing %s\n", argv[OUTFILE]);
    printf("Working...\n");
    size = gamest = cdlst = i = 0;
    sscanf(argv[NOBYTES], "%X", &size);
    sscanf(argv[GAMESTART], "%X", &gamest);
    sscanf(argv[CDLSTART], "%X", &cdlst);
    sscanf(argv[ORG], "%X", &org);
    fseek(game, gamest, SEEK_SET);
    fseek(cdl, cdlst, SEEK_SET);
    if(argv[PRINTADDR][0]!='y') 
        fprintf(out, "\t$ = $%04x\n\n", org);
    while(i < size) {
        if(argv[PRINTADDR][0]=='y') 
            fprintf(out, "$%06X:", org + i);
        opcode[0] = fgetc(game);
        if(opcode[0] < 0) break;
        temp = fgetc(cdl);
        if(temp < 0) break;
        cflag = 1;
        i++;
        switch(temp & 1) {
            case 1:
                for(j = 1; j < opsize[opcode[0]]; j++) {
                    opcode[j] = fgetc(game);
                    cflag = 0;
                    if(opcode[j] < 0) break;
                    cflag = 1;
                    fgetc(cdl);                    
                    i++;
                }
                if(!cflag) break;
                badop = 0;
                switch(opcode[0]) {
                    #include "dasm.h"
                }
                if(badop)
                    printf("bad opcode at %X!\n", org + i - opsize[opcode[0]] - 1);
                /*if(argv[PRINTADDR][0]!='y') {
                    if(badop) 
                        lower(str, strlen(BADOP));
                    else
                        lower(str, 3);                        
                }*/
                if(argv[PRINTADDR][0]!='y') 
                    lower(str, strlen(str));
                fprintf(out, "\t%s\n", str);
                break;
            default:
                if(temp & 0x80) {
                    fgetc(cdl);
                    i++;
                    cflag = 0;
                    opcode[1] = fgetc(game);
                    if(opcode[1] < 0) break;
                    cflag = 1;
                    temp = opcode[0] | (opcode[1] << 8);
                    ctemp = bytetypes[1];
                    ctemp2 = formats[1];
                } else {
                    temp = opcode[0];
                    ctemp = bytetypes[0];
                    ctemp2 = formats[0];
                }
                sprintf(str, ctemp2, ctemp, temp);
                if(argv[PRINTADDR][0]!='y')
                    lower(str, strlen(str));
                fprintf(out, str);
        }
        if(!cflag) break;
    }
    temp = fclose(game);
    temp |=fclose(out);
    temp |=fclose(cdl);
    if(temp) {
        perror("Error closing files!\n");
        return 3;
    }
    printf("Finished!  %X bytes read!\n", i);
    return 0;
}













