#include <stdio.h>

#define IN      1
#define OUT     2
#define ORG     3
#define START   4
#define END     5
#define LABEL   6
#define DSIZE   7
#define ROWSIZE 8
#define COUNT   9

#define LOW     1
#define HIGH    2

const char title[] = 
"quikdis2 <in> <out> <org> <start> <end> <label> <dsize> <rowsize>\n\n";

const char *types[] = {
    ".byte", ".word",
};

const char *formats[] = {
    "$%02x", "$%04x",
};

int main(int argc, char *argv[]) {
    FILE *in = NULL, *out = NULL;
    int temp = 0, size = 0, dsize = 0, i = 0, temp2 = 0, j = 0, row = 0;
    const char *type = types[0], *format = formats[0];
    
    if(argc < COUNT) {
        printf(title);
        exit(0);
    }
    if(argc > COUNT) {
        printf(title);
        printf("Too many parameters!\n");
        exit(1);
    }
    sscanf(argv[START], "%X", &temp2);
    sscanf(argv[END], "%X", &i);
    size = i - temp2;
    if(size < 0) {
        printf("End comes before start\n\n");
        exit(3);
    }
    sscanf(argv[ROWSIZE], "%X", &row);
    if(row <= 0) {
        printf("Need at least one item per row!");
        exit(3);
    }
    sscanf(argv[DSIZE], "%X", &dsize);
    if((dsize < LOW) || (dsize > HIGH)) {
        printf("Datasize too big or too small--try %i to %i\n\n", LOW, HIGH);
        exit(3);
    }
    type = types[dsize - 1];
    format = formats[dsize - 1];
    temp = size % dsize;
    if(temp) {
        printf("Warning: Modulus of %X on size of to be read to datasize", temp);
        printf("\nSubtracting modulus from total size...\n");
        size - temp;
    }
    size /=dsize;
    in = fopen(argv[IN], "rb");
    if(in == NULL) {
        perror("Could not open file IN!\n\n");
        exit(2);
    }
    out = fopen(argv[OUT], "at");
    if(out == NULL) {
        fclose(in);
        perror("Could not open file OUT!\n\n");
        exit(2);
    }
    sscanf(argv[ORG], "%X", &j);
    temp = fseek(in, temp2 - j, SEEK_CUR);
    if(temp) {
        fclose(in);
        fclose(out);
        perror("File error!\n");
        exit(4);
    }
    fprintf(out, "%s:\n", argv[LABEL]);
    j = 0;
    for(i = 0; i < size; i++) {
        temp2 = 0;
        temp = fread(&temp2, dsize, 1, in);
        if(temp!=1) {
            fclose(in);
            fclose(out);
            perror("File error!\n");
            exit(4);
        }
        if(!j) fprintf(out, "\t%s ", type);
        fprintf(out, format, temp2);
        j++;
        if((j >= row)||(size-1==i)) {
            if(j > row) 
                printf("Something happened to cause counter to go over row size...\n");
            j = 0;
            fprintf(out, "\n");
        } else {
            fprintf(out, ", ");
        }
    }
    fprintf(out, "\n");
    temp = fclose(in) | fclose(out);
    if(temp) {
        perror("Error closing files at the very end of program!\n\n\n\n");
        exit(5);
    }
    printf("Finished! ^_^  %X bytes done!", size * dsize);
    exit(0);
}




















