

Start:
	lda FDS_Main_Ctrl_Mir
	and #%11110111
	sta FDS_Main_Ctrl
	lda WorldNumber
	pha
	ldy #ColdBootOffset
	ldx #$05
CheckNextDigit:
	lda TopScoreDisplay,x
	cmp #10
	bcs DoColdBoot
	dex
	bpl CheckNextDigit
	lda WarmBootValidation
	cmp #WarmBootCheckValue
	bne DoColdBoot
	ldy #WarmBootOffset
DoColdBoot:
	jsr InitializeMemory
	sta SND_DELTA_REG+1
	sta OperMode
	sta WorldSelectEnableFlag
	pla
	sta WorldNumber
	lda #WarmBootCheckValue
	sta WarmBootValidation
	sta PseudoRandomBitReg
	lda #%00001111
	sta SND_MASTERCTRL_REG
	lda #%00000110
	sta PPU_CTRL_REG1
	jsr ResetAllSprites
	jsr ClearScreen
	inc DisableScreenFlag
	lda #%11000000
	sta FDS_IRQ_Stat
	cli
	lda Mirror_PPU_CTRL_REG1
	ora #%10000000
	jsr DoPPUReg1
eternity:
	lda temp
	jmp eternity

VRAM_Address_Table:
	.word VRAM_Buffer1
	.word $6b8d, $6bb1, $6bd5
	.word $6bf9, $c62b
	.word VRAM_Buffer2, VRAM_Buffer2
	.word $6c35, $6c1d, $6c25, $6c2d
	.word $6c3d, $6c51, $c876, $c87f
	.word $c893, $c8ab, $c8c1, $c8d7
	.word $c8eb, $c8fc, $c913, $c92b
	.word $c943, $c149, $c10b, $c95c
	.word $c511, $c97d, $c9c0

VRAM_Buffer_Start:
	.byte <VRAM_Buffer1_Offset
	.byte <VRAM_Buffer2_Offset
	
NMI:
	lda Mirror_PPU_CTRL_REG1
	and #%01111110
	sta Mirror_PPU_CTRL_REG1
	sta PPU_CTRL_REG1
	sei
	lda Sprite0HitDetectFlag
	beq SkipTimerSet
	lda #<IRQTime
	sta FDS_IRQTimerLow
	lda #>IRQTime
	sta FDS_IRQTimerHigh
	lda #%00000010
	sta FDS_IRQTimerCtrl
	inc IRQWaitFlag
SkipTimerSet:
	lda Mirror_PPU_CTRL_REG2
	and #%11100110
	ldy DisableScreenFlag
	bne ScreenOff
	lda Mirror_PPU_CTRL_REG2
	ora #%00011110
ScreenOff:
	sta Mirror_PPU_CTRL_REG2
	and #%11100111
	sta PPU_CTRL_REG2
	ldx PPU_STATUS
	lda #$00
	jsr InitScroll
	sta PPU_SPR_ADDR
	lda #$02
	sta SPR_DMA
	lda VRAM_Buffer_AddrCtrl
	asl
	tax
	lda VRAM_Address_Table,x
	sta address_lo
	inx
	lda VRAM_Address_Table,x
	sta address_hi
	jsr UpdateScreen
	ldy #$00
	ldx VRAM_Buffer_AddrCtrl
	cpx #$06
	bne DontAdd
	iny

DontAdd:
	ldx VRAM_Buffer_Start,y
	lda #$00
	sta VRAM_Buffer1_Offset,x
	sta VRAM_Buffer1,x
	sta VRAM_Buffer_AddrCtrl
	lda Mirror_PPU_CTRL_REG2
	sta PPU_CTRL_REG2
	cli
SoundEngine_JSR:
	jsr SoundEngine

	jsr ReadJoypads
	jsr PauseRoutine
	jsr UpdateTopScore
	lda GamePauseStatus
	lsr
	bcs PauseSkip
	lda TimerControl 
	beq DecTimers
	dec TimerControl 
	bne NoDecTimers

DecTimers:
	ldx #$14
	dec IntervalTimerControl
	bpl DecTimersLoop
	lda #$14
	sta IntervalTimerControl
	ldx #$23

DecTimersLoop:
	lda Timers,x
	beq SkipExpTimer
	dec Timers,x

SkipExpTimer
	dex
	bpl DecTimersLoop

NoDecTimers:
	inc FrameCounter
PauseSkip:
	ldx #$00
	ldy #$07
	lda PseudoRandomBitReg
	and #$02
	sta temp
	lda PseudoRandomHelper
	and #$02
	eor temp
	clc
	beq RotPRandomBit
	sec

RotPRandomBit:
	ror PseudoRandomBitReg,x
	inx
	dey
	bne RotPRandomBit
	lda GamePauseStatus
	lsr
	bcs WaitForIRQ
	lda Sprite0HitDetectFlag
	beq TestWorld
	jsr ResetAllButOne
	jsr SpriteShuffler

TestWorld:
	lda WorldNumber
	cmp #World9+1
	bcc NormalOp
	jsr TerminateGame

NormalOp:
	jsr OperModeExecutionTree

WaitForIRQ:
	lda IRQWaitFlag
	bne WaitForIRQ
	lda PPU_STATUS
	lda Mirror_PPU_CTRL_REG1
	ora #%10000000
	sta Mirror_PPU_CTRL_REG1
	sta PPU_CTRL_REG1
	rti

IRQ:
	sei
	php
	pha
	txa
	pha
	tya
	pha
	lda FDS_IO_Status
	pha
	and #%00000010
	bne DelayFDS
	pla
	and #$01
	beq EndIRQ
	lda Mirror_PPU_CTRL_REG1
	and #%11110111
	ora NumberOfPlayers
	sta Mirror_PPU_CTRL_REG1
	sta PPU_CTRL_REG1
	lda #$00
	sta FDS_IRQTimerCtrl
	lda HorizontalScroll
	sta PPU_SCROLL_REG
	lda VerticalScroll
	sta PPU_SCROLL_REG
	lda #$00
	sta IRQWaitFlag
	jmp EndIRQ
DelayFDS:
	pla
	jsr FDS_Delay	
EndIRQ:
	pla
	tay
	pla
	tax
	pla
	plp
	cli
	rti

PauseRoutine:
	lda OperMode
	cmp #VictoryModeValue
	beq ChkPauseTimer
	cmp #GameModeValue
	bne ExitPause
	lda OperMode_Task
	cmp #$04
	bne ExitPause
ChkPauseTimer:
	lda GamePauseTimer
	beq ChkStart
	dec GamePauseTimer
	rts
ChkStart:
	lda SavedJoypad1Bits
	and #%00010000
	beq ClrPauseStatus
	lda GamePauseStatus
	and #%10000000
	bne ExitPause
	lda #$2b
	sta GamePauseTimer
	lda GamePauseStatus
	tay
	iny
	sty PauseSoundQueue
	eor #%00000001
	ora #%10000000
	bne SetPauseStatus
ClrPauseStatus:
	lda GamePauseStatus
	and #%01111111
SetPauseStatus:
	sta GamePauseStatus
ExitPause:
	rts

SpriteShuffler:
	ldy AreaType
	lda #$28
	sta temp
	ldx #$0e
ShuffleLoop:
	lda SprDataOffset,x
	cmp temp
	bcc NextSprOffset
	ldy SprShuffleAmtOffset
	clc
	adc SprShuffleAmt,y
	bcc StrSprOffset
	clc
	adc temp
StrSprOffset:
	sta SprDataOffset,x
NextSprOffset:
	dex
	bpl ShuffleLoop
	ldx SprShuffleAmtOffset
	inx
	cpx #$03
	bne SetAmtOffset
	ldx #$00
SetAmtOffset:
	stx SprShuffleAmtOffset
	ldx #$08
	ldy #$02
SetMiscOffset:
	lda SprDataOffset+5,y
	sta Misc_SprDataOffset-2,x
	clc
	adc #$08
	sta Misc_SprDataOffset-1,x
	clc
	adc #$08
	sta Misc_SprDataOffset,x
	dex
	dex
	dex
	dey
	bpl SetMiscOffset
	rts

OperModeExecutionTree:
	lda OperMode
	jsr JumpEngine

	.word TitleScreenMode
	.word $7a37
	.word $6298
	.word $7057

ResetAllSprites:
	ldy #$00
	.byte $2c          ;bit instruction (comes out to "bit $04a0")
	
ResetAllButOne:
	ldy #$04
	lda #$f8
DoNextSprite:
	sta Sprite_Data,y
	iny
	iny
	iny
	iny
	bne DoNextSprite
	rts

	jsr $62bc
	lda $0772
	beq $62b6
	ldx $075f
	cpx #$07
	bne $62af
	cmp #$05
	beq $6297
	cmp #$0d
	beq $6297
	ldx #$00
	stx $08
	jsr $8c23
	jsr $be0f
	jmp $bbc4

	lda $075f
	cmp #$07
	beq $62d5
	lda $0772
	jsr $6c7d
	.byte $21
	.byte $9c
	.byte $ff
	.byte $62
	.byte $34
	.byte $63
	.byte $6d
	.byte $63
	.byte $af
	.byte $63
	.byte $d2
	.byte $63
	.byte $ad
	.byte $72
	.byte $07
	.byte $20
	.byte $7d
	.byte $6c
	.byte $21
	.byte $9c
	.byte $ff
	.byte $62
	.byte $34
	.byte $63
	.byte $81
	.byte $c0
	.byte $88
	.byte $c0
	.byte $8e
	.byte $c0
	.byte $fe
	.byte $c5
	.byte $42
	.byte $c6
	.byte $af
	.byte $63
	.byte $86
	.byte $c6
	.byte $ca
	.byte $c6
	.byte $10
	.byte $c7
	.byte $27
	.byte $c7
	.byte $38
	.byte $c7
	.byte $01
	.byte $02
	.byte $04
	.byte $08
	.byte $10
	.byte $20
	.byte $40
	.byte $80
	ldx $071b
	inx
	stx $34
	ldy $075f
	lda $62f7,y
	ora $07fa
	sta $07fa
	lda $07fb
	beq $6322
	lda $075f
	cmp #$03
	bcc $6322
	lda #$07
	sta $075f
	lda #$08
	sta $fc
	inc $0772
	rts

	lda $0770
	bne $6326
	lda #$05
	jmp $65a1

	ldy #$00
	sty $35
	lda $6d
	cmp $34
	bne $6344
	lda $86
	cmp #$60
	bcs $6347
	inc $35
	iny
	tya
	jsr $7c3e
	lda $071a
	cmp $34
	beq $6368
	lda $0768
	clc
	adc #$80
	sta $0768
	lda #$01
	adc #$00
	tay
	jsr $7b20
	jsr $7acb
	inc $35
	lda $35
	beq $63ab
	rts

	lda $0749
	bne $6391
	lda $0719
	beq $637f
	cmp #$08
	bcs $6391
	cmp #$01
	bcc $6391
	tay
	beq $638a
	cpy #$03
	bcs $63a4
	cpy #$02
	bcs $6391
	tya
	clc
	adc #$0c
	sta $0773
	lda $0749
	clc
	adc #$04
	sta $0749
	lda $0719
	adc #$00
	sta $0719
	cmp #$06
	bcc $63ae
	lda #$08
	sta $07a1
	inc $0772
	rts

	lda $07a1
	cmp #$06
	bcs $63d1
	jsr $9f57
	lda $07ec
	ora $07ed
	ora $07ee
	bne $63d1
	lda #$30
	sta $0780
	lda #$06
	sta $07a1
	inc $0772
	rts

	lda $07a1
	bne $63fc
	lda #$00
	sta $0760
	sta $075c
	sta $0772
	lda $075f
	clc
	adc #$01
	cmp #$08
	bcc $63ee
	lda #$08
	sta $075f
	jsr $c2a4
	inc $0757
	lda #$01
	sta $0770
	rts

	.byte $ff
	.byte $ff
	.byte $f6
	.byte $fb
	.byte $f7
	.byte $fb
	.byte $f8
	.byte $fb
	.byte $f9
	.byte $fb
	.byte $fa
	.byte $fb
	.byte $f6
	.byte $50
	.byte $f7
	.byte $50
	.byte $f8
	.byte $50
	.byte $f9
	.byte $50
	.byte $fa
	.byte $50
	.byte $fd
	.byte $fe
	.byte $ff
	.byte $41
	.byte $42
	.byte $44
	.byte $45
	.byte $48
	.byte $31
	.byte $32
	.byte $34
	.byte $35
	.byte $38
	.byte $00
	lda $0110,x
	beq $63fc
	cmp #$0b
	bcc $642f
	lda #$0b
	sta $0110,x
	tay
	lda $012c,x
	bne $6439
	sta $0110,x
	rts

	dec $012c,x
	cmp #$2b
	bne $645e
	cpy #$0b
	bne $644b
	inc $075a
	lda #$40
	sta $fe
	lda $6415,y
	lsr
	lsr
	lsr
	lsr
	tax
	lda $6415,y
	and #$0f
	sta $0134,x
	jsr $87e8
	ldy $06e5,x
	lda $16,x
	cmp #$12
	beq $6489
	cmp #$0d
	beq $6489
	cmp #$05
	beq $6481
	cmp #$0a
	beq $6489
	cmp #$0b
	beq $6489
	cmp #$09
	bcs $6481
	lda $1e,x
	cmp #$02
	bcs $6489
	ldx $03ee
	ldy $06ec,x
	ldx $08
	lda $011e,x
	cmp #$18
	bcc $6495
	sbc #$01
	sta $011e,x
	lda $011e,x
	sbc #$08
	jsr $b267
	lda $0117,x
	sta $0203,y
	clc
	adc #$08
	sta $0207,y
	lda #$02
	sta $0202,y
	sta $0206,y
	lda $0110,x
	asl
	tax
	lda $63fd,x
	sta $0201,y
	lda $63fe,x
	sta $0205,y
	ldx $08
	rts

	lda $073c
	jsr $6c7d
	.byte $eb
	.byte $64
	.byte $fb
	.byte $64
	.byte $a7
	.byte $65
	.byte $af
	.byte $65
	.byte $f8
	.byte $65
	.byte $7a
	.byte $67
	.byte $17
	.byte $66
	.byte $d0
	.byte $c5
	.byte $7a
	.byte $67
	.byte $59
	.byte $66
	.byte $1f
	.byte $65
	.byte $3f
	.byte $65
	.byte $98
	.byte $65
	.byte $2a
	.byte $63
	.byte $73
	.byte $c5
	.byte $8a
	.byte $c5
	jsr $6287
	jsr $6c92
	lda $0770
	beq $6528
	ldx #$03
	jmp $6525

	lda $0744
	pha
	lda $0756
	pha
	lda #$00
	sta $0756
	lda #$02
	sta $0744
	jsr $654d
	pla
	sta $0756
	pla
	sta $0744
	jmp $6613

	.byte $01
	.byte $02
	.byte $03
	.byte $04
	ldy $074e
	ldx $651b,y
	stx $0773
	jmp $6613

	.byte $00
	.byte $09
	.byte $0a
	.byte $04
	.byte $22
	.byte $22
	.byte $0f
	.byte $0f
	.byte $0f
	.byte $22
	.byte $0f
	.byte $0f
	.byte $22
	.byte $16
	.byte $27
	.byte $18
	.byte $22
	.byte $37
	.byte $27
	.byte $16
	ldy $0744
	beq $654a
	lda $6527,y
	sta $0773
	inc $073c
	ldx $0300
	ldy #$00
	lda $0756
	cmp #$02
	bne $655b
	ldy #$04
	lda #$03
	sta $00
	lda $6537,y
	sta $0304,x
	iny
	inx
	dec $00
	bpl $655f
	ldx $0300
	ldy $0744
	bne $6576
	ldy $074e
	lda $652f,y
	sta $0304,x
	lda #$3f
	sta $0301,x
	lda #$10
	sta $0302,x
	lda #$04
	sta $0303,x
	lda #$00
	sta $0308,x
	txa
	clc
	adc #$07
	sta $0300
	rts

	lda $0733
	cmp #$01
	bne $65a4
	lda #$0b
	sta $0773
	jmp $6613

	lda #$00
	jsr $671b
	jmp $6613

	jsr $87ed
	ldx $0300
	lda #$20
	sta $0301,x
	lda #$73
	sta $0302,x
	lda #$03
	sta $0303,x
	jsr $65e6
	sta $0304,x
	lda #$28
	sta $0305,x
	ldy $075c
	iny
	tya
	sta $0306,x
	lda #$00
	sta $0307,x
	txa
	clc
	adc #$06
	sta $0300
	jmp $6613

	ldy $075f
	lda $07fb
	beq $65f5
	tya
	and #$03
	clc
	adc #$09
	tay
	iny
	tya
	rts

	lda $0759
	beq $6610
	lda #$00
	sta $0759
	lda #$02
	jsr $671b
	jsr $6782
	lda #$00
	sta $0774
	rts

	inc $073c
	inc $073c
	rts

	lda $0770
	beq $6653
	cmp #$03
	beq $6644
	lda $0752
	bne $6653
	ldy $074e
	cpy #$03
	beq $6631
	lda $0769
	bne $6653
	jsr $bc7f
	lda #$01
	jsr $6604
	lda $075f
	cmp #$08
	bne $6613
	inc $0774
	rts

	lda #$03
	jsr $671b
	lda $075f
	cmp #$08
	beq $6613
	jmp $6326

	lda #$09
	sta $073c
	rts

	inc $0774
	jsr $70c1
	lda $071f
	bne $665c
	dec $071e
	bpl $666c
	inc $073c
	lda #$06
	sta $0773
	rts

	.byte $20
	.byte $43
	.byte $05
	.byte $16
	.byte $0a
	.byte $1b
	.byte $12
	.byte $18
	.byte $20
	.byte $52
	.byte $0b
	.byte $20
	.byte $18
	.byte $1b
	.byte $15
	.byte $0d
	.byte $24
	.byte $24
	.byte $1d
	.byte $12
	.byte $16
	.byte $0e
	.byte $20
	.byte $68
	.byte $05
	.byte $00
	.byte $24
	.byte $24
	.byte $2e
	.byte $29
	.byte $23
	.byte $c0
	.byte $7f
	.byte $aa
	.byte $23
	.byte $c2
	.byte $01
	.byte $ea
	.byte $ff
	.byte $21
	.byte $cd
	.byte $07
	.byte $24
	.byte $24
	.byte $29
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $21
	.byte $4b
	.byte $09
	.byte $20
	.byte $18
	.byte $1b
	.byte $15
	.byte $0d
	.byte $24
	.byte $24
	.byte $28
	.byte $24
	.byte $22
	.byte $0c
	.byte $47
	.byte $24
	.byte $23
	.byte $dc
	.byte $01
	.byte $ba
	.byte $ff
	.byte $22
	.byte $0c
	.byte $07
	.byte $1d
	.byte $12
	.byte $16
	.byte $0e
	.byte $24
	.byte $1e
	.byte $19
	.byte $ff
	.byte $21
	.byte $6b
	.byte $09
	.byte $10
	.byte $0a
	.byte $16
	.byte $0e
	.byte $24
	.byte $18
	.byte $1f
	.byte $0e
	.byte $1b
	.byte $21
	.byte $eb
	.byte $08
	.byte $0c
	.byte $18
	.byte $17
	.byte $1d
	.byte $12
	.byte $17
	.byte $1e
	.byte $0e
	.byte $22
	.byte $0c
	.byte $47
	.byte $24
	.byte $22
	.byte $4b
	.byte $05
	.byte $1b
	.byte $0e
	.byte $1d
	.byte $1b
	.byte $22
	.byte $ff
	.byte $25
	.byte $84
	.byte $15
	.byte $20
	.byte $0e
	.byte $15
	.byte $0c
	.byte $18
	.byte $16
	.byte $0e
	.byte $24
	.byte $1d
	.byte $18
	.byte $24
	.byte $20
	.byte $0a
	.byte $1b
	.byte $19
	.byte $24
	.byte $23
	.byte $18
	.byte $17
	.byte $0e
	.byte $2b
	.byte $26
	.byte $2d
	.byte $01
	.byte $24
	.byte $27
	.byte $d9
	.byte $46
	.byte $aa
	.byte $27
	.byte $e1
	.byte $45
	.byte $aa
	.byte $00
	.byte $02
	.byte $03
	.byte $04
	.byte $01
	.byte $06
	.byte $07
	.byte $08
	.byte $05
	.byte $0b
	.byte $0c
	.byte $0d
	.byte $00
	.byte $27
	.byte $46
	.byte $51
	pha
	tay
	ldx $6717,y
	ldy #$00
	lda $6672,x
	cmp #$ff
	beq $6730
	sta $0301,y
	inx
	iny
	bne $6722
	lda #$00
	sta $0301,y
	pla
	beq $675d
	tax
	dex
	bne $675d
	lda $075a
	clc
	adc #$01
	cmp #$0a
	bcc $674d
	sbc #$0a
	ldy #$9f
	sty $0308
	sta $0309
	jsr $65e6
	sta $0314
	ldy $075c
	iny
	sty $0316
	rts

	pha
	ldy #$ff
	iny
	lda $66e7,y
	sta $0301,y
	bne $6761
	pla
	sec
	sbc #$80
	tax
	lda $670c,x
	sta $031c
	lda #$24
	jmp $6594

	lda $07a0
	bne $678a
	jsr $6287
	lda #$07
	sta $07a0
	inc $073c
	rts

	lda $0726
	and #$01
	sta $05
	ldy $0340
	sty $00
	lda $0721
	sta $0342,y
	lda $0720
	sta $0341,y
	lda #$9a
	sta $0343,y
	lda #$00
	sta $04
	tax
	stx $01
	lda $06a1,x
	and #$c0
	sta $03
	asl
	rol
	rol
	tay
	lda $69e5,y
	sta $06
	lda $69e9,y
	sta $07
	lda $06a1,x
	asl
	asl
	sta $02
	lda $071f
	and #$01
	eor #$01
	asl
	adc $02
	tay
	ldx $00
	lda ($06),y
	sta $0344,x
	iny
	lda ($06),y
	sta $0345,x
	ldy $04
	lda $05
	bne $67f7
	lda $01
	lsr
	bcs $6807
	rol $03
	rol $03
	rol $03
	jmp $680d

	lda $01
	lsr
	bcs $680b
	lsr $03
	lsr $03
	lsr $03
	lsr $03
	jmp $680d

	lsr $03
	lsr $03
	inc $04
	lda $03f9,y
	ora $03
	sta $03f9,y
	inc $00
	inc $00
	ldx $01
	inx
	cpx #$0d
	bcc $67ad
	ldy $00
	iny
	iny
	iny
	lda #$00
	sta $0341,y
	sty $0340
	inc $0721
	lda $0721
	and #$1f
	bne $6844
	lda #$80
	sta $0721
	lda $0720
	eor #$04
	sta $0720
	jmp $689a

	lda $0721
	and #$1f
	sec
	sbc #$04
	and #$1f
	sta $01
	lda $0720
	bcs $685a
	eor #$04
	and #$04
	ora #$23
	sta $00
	lda $01
	lsr
	lsr
	adc #$c0
	sta $01
	ldx #$00
	ldy $0340
	lda $00
	sta $0341,y
	lda $01
	clc
	adc #$08
	sta $0342,y
	sta $01
	lda $03f9,x
	sta $0344,y
	lda #$01
	sta $0343,y
	lsr
	sta $03f9,x
	iny
	iny
	iny
	iny
	inx
	cpx #$07
	bcc $686d
	sta $0341,y
	sty $0340
	lda #$06
	sta $0773
	rts

	.byte $27
	.byte $27
	.byte $27
	.byte $17
	.byte $07
	.byte $17
	.byte $3f
	.byte $0c
	.byte $04
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $00
	.byte $0f
	.byte $07
	.byte $12
	.byte $0f
	.byte $0f
	.byte $07
	.byte $17
	.byte $0f
	.byte $0f
	.byte $07
	.byte $17
	.byte $1c
	.byte $0f
	.byte $07
	.byte $17
	.byte $00
	lda $09
	and #$07
	bne $6915
	ldx $0300
	cpx #$31
	bcs $6915
	tay
	lda $68a6,y
	sta $0301,x
	inx
	iny
	cpy #$08
	bcc $68cc
	ldx $0300
	lda #$03
	sta $00
	lda $074e
	asl
	asl
	tay
	lda $68ae,y
	sta $0304,x
	iny
	inx
	dec $00
	bpl $68e5
	ldx $0300
	ldy $06d4
	lda $68a0,y
	sta $0305,x
	lda $0300
	clc
	adc #$07
	sta $0300
	inc $06d4
	lda $06d4
	cmp #$06
	bcc $6915
	lda #$00
	sta $06d4
	rts

	.byte $45
	.byte $45
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $57
	.byte $58
	.byte $59
	.byte $5a
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	ldy #$41
	lda #$03
	ldx $074e
	bne $6935
	lda #$04
	jsr $6974
	lda #$06
	sta $0773
	rts

	jsr $694a
	inc $03f0
	dec $03ec,x
	rts

	lda #$00
	ldy #$03
	cmp #$00
	beq $6964
	ldy #$00
	cmp #$56
	beq $6964
	cmp #$4f
	beq $6964
	iny
	cmp #$5c
	beq $6964
	cmp #$50
	beq $6964
	iny
	tya
	ldy $0300
	iny
	jsr $6974
	dey
	tya
	clc
	adc #$0a
	jmp $6594

	stx $00
	sty $01
	asl
	asl
	tax
	ldy #$20
	lda $06
	cmp #$d0
	bcc $6985
	ldy #$24
	sty $03
	and #$0f
	asl
	sta $04
	lda #$00
	sta $05
	lda $02
	clc
	adc #$20
	asl
	rol $05
	asl
	rol $05
	adc $04
	sta $04
	lda $05
	adc #$00
	clc
	adc $03
	sta $05
	ldy $01
	lda $6916,x
	sta $0303,y
	lda $6917,x
	sta $0304,y
	lda $6918,x
	sta $0308,y
	lda $6919,x
	sta $0309,y
	lda $04
	sta $0301,y
	clc
	adc #$20
	sta $0306,y
	lda $05
	sta $0300,y
	sta $0305,y
	lda #$02
	sta $0302,y
	sta $0307,y
	lda #$00
	sta $030a,y
	ldx $00
	rts

	.byte $ed
	.byte $7d
	.byte $3d
	.byte $71
	.byte $69
	.byte $6a
	.byte $6b
	.byte $6b
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $27
	.byte $27
	.byte $27
	.byte $27
	.byte $24
	.byte $24
	.byte $24
	.byte $35
	.byte $36
	.byte $25
	.byte $37
	.byte $25
	.byte $24
	.byte $38
	.byte $24
	.byte $24
	.byte $24
	.byte $30
	.byte $30
	.byte $26
	.byte $26
	.byte $26
	.byte $34
	.byte $26
	.byte $24
	.byte $31
	.byte $24
	.byte $32
	.byte $33
	.byte $26
	.byte $24
	.byte $33
	.byte $34
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $24
	.byte $c0
	.byte $24
	.byte $c0
	.byte $24
	.byte $7f
	.byte $7f
	.byte $24
	.byte $b8
	.byte $ba
	.byte $b9
	.byte $bb
	.byte $b8
	.byte $bc
	.byte $b9
	.byte $bd
	.byte $ba
	.byte $bc
	.byte $bb
	.byte $bd
	.byte $60
	.byte $64
	.byte $61
	.byte $65
	.byte $62
	.byte $66
	.byte $63
	.byte $67
	.byte $60
	.byte $64
	.byte $61
	.byte $65
	.byte $62
	.byte $66
	.byte $63
	.byte $67
	.byte $68
	.byte $68
	.byte $69
	.byte $69
	.byte $26
	.byte $26
	.byte $6a
	.byte $6a
	.byte $4b
	.byte $4c
	.byte $4d
	.byte $4e
	.byte $4d
	.byte $4f
	.byte $4d
	.byte $4f
	.byte $4d
	.byte $4e
	.byte $50
	.byte $51
	.byte $86
	.byte $8a
	.byte $87
	.byte $8b
	.byte $88
	.byte $8c
	.byte $88
	.byte $8c
	.byte $89
	.byte $8d
	.byte $69
	.byte $69
	.byte $8e
	.byte $91
	.byte $8f
	.byte $92
	.byte $26
	.byte $93
	.byte $26
	.byte $93
	.byte $90
	.byte $94
	.byte $69
	.byte $69
	.byte $a4
	.byte $e9
	.byte $ea
	.byte $eb
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $2f
	.byte $24
	.byte $3d
	.byte $a2
	.byte $a2
	.byte $a3
	.byte $a3
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $a2
	.byte $a2
	.byte $a3
	.byte $a3
	.byte $99
	.byte $24
	.byte $99
	.byte $24
	.byte $24
	.byte $a2
	.byte $3e
	.byte $3f
	.byte $5b
	.byte $5c
	.byte $24
	.byte $a3
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $9d
	.byte $47
	.byte $9e
	.byte $47
	.byte $47
	.byte $47
	.byte $27
	.byte $27
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $27
	.byte $27
	.byte $47
	.byte $47
	.byte $a9
	.byte $47
	.byte $aa
	.byte $47
	.byte $9b
	.byte $27
	.byte $9c
	.byte $27
	.byte $27
	.byte $27
	.byte $27
	.byte $27
	.byte $52
	.byte $52
	.byte $52
	.byte $52
	.byte $80
	.byte $a0
	.byte $81
	.byte $a1
	.byte $be
	.byte $be
	.byte $bf
	.byte $bf
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $45
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $ab
	.byte $ac
	.byte $ad
	.byte $ae
	.byte $5d
	.byte $5e
	.byte $5d
	.byte $5e
	.byte $c1
	.byte $24
	.byte $c1
	.byte $24
	.byte $c6
	.byte $c8
	.byte $c7
	.byte $c9
	.byte $ca
	.byte $cc
	.byte $cb
	.byte $cd
	.byte $2a
	.byte $2a
	.byte $40
	.byte $40
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $47
	.byte $24
	.byte $47
	.byte $82
	.byte $83
	.byte $84
	.byte $85
	.byte $b4
	.byte $b6
	.byte $b5
	.byte $b7
	.byte $24
	.byte $47
	.byte $24
	.byte $47
	.byte $86
	.byte $8a
	.byte $87
	.byte $8b
	.byte $8e
	.byte $91
	.byte $8f
	.byte $92
	.byte $24
	.byte $2f
	.byte $24
	.byte $3d
	.byte $24
	.byte $24
	.byte $24
	.byte $35
	.byte $36
	.byte $25
	.byte $37
	.byte $25
	.byte $24
	.byte $38
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $39
	.byte $24
	.byte $3a
	.byte $24
	.byte $3b
	.byte $24
	.byte $3c
	.byte $24
	.byte $24
	.byte $24
	.byte $41
	.byte $26
	.byte $41
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $b0
	.byte $b1
	.byte $b2
	.byte $b3
	.byte $77
	.byte $79
	.byte $77
	.byte $79
	.byte $6b
	.byte $70
	.byte $2c
	.byte $2d
	.byte $6c
	.byte $71
	.byte $6d
	.byte $72
	.byte $6e
	.byte $73
	.byte $6f
	.byte $74
	.byte $53
	.byte $55
	.byte $54
	.byte $56
	.byte $53
	.byte $55
	.byte $54
	.byte $56
	.byte $53
	.byte $55
	.byte $54
	.byte $56
	.byte $a5
	.byte $a7
	.byte $a6
	.byte $a8
	.byte $c2
	.byte $c4
	.byte $c3
	.byte $c5
	.byte $57
	.byte $59
	.byte $58
	.byte $5a
	.byte $7b
	.byte $7d
	.byte $7c
	.byte $7e
	.byte $3f
	.byte $00
	.byte $20
	.byte $0f
	.byte $15
	.byte $12
	.byte $25
	.byte $0f
	.byte $3a
	.byte $1a
	.byte $0f
	.byte $0f
	.byte $30
	.byte $12
	.byte $0f
	.byte $0f
	.byte $27
	.byte $12
	.byte $0f
	.byte $22
	.byte $16
	.byte $27
	.byte $18
	.byte $0f
	.byte $10
	.byte $30
	.byte $27
	.byte $0f
	.byte $16
	.byte $30
	.byte $27
	.byte $0f
	.byte $0f
	.byte $30
	.byte $10
	.byte $00
	.byte $3f
	.byte $00
	.byte $20
	.byte $0f
	.byte $29
	.byte $1a
	.byte $0f
	.byte $0f
	.byte $36
	.byte $17
	.byte $0f
	.byte $0f
	.byte $30
	.byte $21
	.byte $0f
	.byte $0f
	.byte $27
	.byte $17
	.byte $0f
	.byte $0f
	.byte $16
	.byte $27
	.byte $18
	.byte $0f
	.byte $1a
	.byte $30
	.byte $27
	.byte $0f
	.byte $16
	.byte $30
	.byte $27
	.byte $0f
	.byte $0f
	.byte $36
	.byte $17
	.byte $00
	.byte $3f
	.byte $00
	.byte $20
	.byte $0f
	.byte $29
	.byte $1a
	.byte $09
	.byte $0f
	.byte $3c
	.byte $1c
	.byte $0f
	.byte $0f
	.byte $30
	.byte $21
	.byte $1c
	.byte $0f
	.byte $27
	.byte $17
	.byte $1c
	.byte $0f
	.byte $16
	.byte $27
	.byte $18
	.byte $0f
	.byte $1c
	.byte $36
	.byte $17
	.byte $0f
	.byte $16
	.byte $30
	.byte $27
	.byte $0f
	.byte $0c
	.byte $3c
	.byte $1c
	.byte $00
	.byte $3f
	.byte $00
	.byte $20
	.byte $0f
	.byte $30
	.byte $10
	.byte $00
	.byte $0f
	.byte $30
	.byte $10
	.byte $00
	.byte $0f
	.byte $30
	.byte $16
	.byte $00
	.byte $0f
	.byte $27
	.byte $17
	.byte $00
	.byte $0f
	.byte $16
	.byte $27
	.byte $18
	.byte $0f
	.byte $1c
	.byte $36
	.byte $17
	.byte $0f
	.byte $16
	.byte $30
	.byte $27
	.byte $0f
	.byte $00
	.byte $30
	.byte $10
	.byte $00
	.byte $3f
	.byte $00
	.byte $04
	.byte $22
	.byte $30
	.byte $00
	.byte $10
	.byte $00
	.byte $3f
	.byte $00
	.byte $04
	.byte $0f
	.byte $30
	.byte $00
	.byte $10
	.byte $00
	.byte $3f
	.byte $00
	.byte $04
	.byte $22
	.byte $27
	.byte $16
	.byte $0f
	.byte $00
	.byte $3f
	.byte $14
	.byte $04
	.byte $0f
	.byte $1a
	.byte $30
	.byte $27
	.byte $00
	.byte $25
	.byte $48
	.byte $10
	.byte $1d
	.byte $11
	.byte $0a
	.byte $17
	.byte $14
	.byte $24
	.byte $22
	.byte $18
	.byte $1e
	.byte $24
	.byte $16
	.byte $0a
	.byte $1b
	.byte $12
	.byte $18
	.byte $2b
	.byte $00
	.byte $25
	.byte $c5
	.byte $16
	.byte $0b
	.byte $1e
	.byte $1d
	.byte $24
	.byte $18
	.byte $1e
	.byte $1b
	.byte $24
	.byte $19
	.byte $1b
	.byte $12
	.byte $17
	.byte $0c
	.byte $0e
	.byte $1c
	.byte $1c
	.byte $24
	.byte $12
	.byte $1c
	.byte $24
	.byte $12
	.byte $17
	.byte $26
	.byte $05
	.byte $0f
	.byte $0a
	.byte $17
	.byte $18
	.byte $1d
	.byte $11
	.byte $0e
	.byte $1b
	.byte $24
	.byte $0c
	.byte $0a
	.byte $1c
	.byte $1d
	.byte $15
	.byte $0e
	.byte $2b
	.byte $00
	
JumpEngine:
	asl
	tay
	pla
	sta jpe_comefrom_lo
	pla
	sta jpe_comefrom_hi
	iny
	lda (jpe_comefrom),y
	sta jpe_jmp_lo
	iny
	lda (jpe_comefrom),y
	sta jpe_jmp_hi
	jmp (jpe_jmp)

ClearScreen:
	lda PPU_STATUS
	lda Mirror_PPU_CTRL_REG1
	ora #%00010000
	and #%11110000
	jsr DoPPUReg1
	lda #$24
	jsr DoClearScreens
	lda #$20
	
DoClearScreens:
	sta PPU_ADDRESS
	lda #$00
	sta PPU_ADDRESS
	ldx #$04
	ldy #$c0
	lda #$24
ClrNextByte:
	sta PPU_DATA
	dey
	bne ClrNextByte
	dex
	bne ClrNextByte
	ldy #$40
	txa
	sta VRAM_Buffer1_Offset
	sta VRAM_Buffer1
ClrAttrTbl:
	sta PPU_DATA
	dey
	bne ClrAttrTbl
	sta HorizontalScroll
	sta VerticalScroll
	jmp InitScroll

ReadJoypads:
	lda #$01
	sta JOYPAD_PORT
	lsr
	tax
	sta JOYPAD_PORT
	jsr ReadFirstJoypad
	inx
ReadFirstJoypad:
	ldy #$08
NextButton:
	pha
	lda JOYPAD_PORT,x
	sta temp
	lsr
	ora temp
	lsr
	pla
	rol
	dey
	bne NextButton
	sta SavedJoypadBits,x
	pha
	and #%00110000
	and JoypadBitMask,x
	beq Save8Bits
	pla
	and #%11001111
	sta SavedJoypadBits,x
	rts
Save8Bits:
	pla
	sta JoypadBitMask,x
	rts

WriteBufferToScreen:
	sta PPU_ADDRESS
	iny
	lda (address),y
	sta PPU_ADDRESS
	iny
	lda (address),y
	asl
	pha
	lda Mirror_PPU_CTRL_REG1
	ora #%00000100
	bcs SetupWrites
	and #%11111011
SetupWrites:
	jsr DoPPUReg1
	pla
	asl
	bcc GetLength
	ora #%00000010
	iny
GetLength:
	lsr
	lsr
	tax
OutputToVRAM:
	bcs RepeatByte
	iny
RepeatByte:
	lda (address),y
	sta PPU_DATA
	dex
	bne OutputToVRAM
	sec
	tya
	adc address_lo
	sta address_lo
	lda #$00
	adc address_hi
	sta address_hi
	lda #$3f
	sta PPU_ADDRESS
	lda #$00
	sta PPU_ADDRESS
	sta PPU_ADDRESS
	sta PPU_ADDRESS
UpdateScreen:
	ldx PPU_STATUS
	ldy #$00
	lda (address),y
	bne WriteBufferToScreen

InitScroll:
	sta PPU_SCROLL_REG
	sta PPU_SCROLL_REG
	rts

DoPPUReg1:
	sta PPU_CTRL_REG1
	sta Mirror_PPU_CTRL_REG1
	rts


	.byte $ef
	.byte $06
	.byte $62
	.byte $06
	.byte $6d
	.byte $02
	.byte $7a
	.byte $03
	.byte $06
	.byte $0c
	.byte $12
	.byte $18
	sta $00
	jsr $6d84
	lda $00
	lsr
	lsr
	lsr
	lsr
	clc
	adc #$01
	and #$0f
	cmp #$06
	bcs $6dd1
	pha
	asl
	tay
	ldx $0300
	lda #$20
	cpy #$00
	bne $6d9b
	lda #$22
	sta $0301,x
	lda $6d6d,y
	sta $0302,x
	lda $6d6e,y
	sta $0303,x
	sta $03
	stx $02
	pla
	tax
	lda $6d75,x
	sec
	sbc $6d6e,y
	tay
	ldx $02
	lda $07d7,y
	sta $0304,x
	inx
	iny
	dec $03
	bne $6dba
	lda #$00
	sta $0304,x
	inx
	inx
	inx
	stx $0300
	rts

	lda $0770
	beq $6ded
	ldx #$05
	lda $0134,x
	clc
	adc $07d7,y
	bmi $6df8
	cmp #$0a
	bcs $6dff
	sta $07d7,y
	dey
	dex
	bpl $6dd9
	lda #$00
	ldx #$06
	sta $0133,x
	dex
	bpl $6df1
	rts

	dec $0133,x
	lda #$09
	bne $6de6
	sec
	sbc #$0a
	inc $0133,x
	jmp $6de6

UpdateTopScore:
	ldx #$05
	ldy #$05
	sec
GetScoreDifference:
	lda PlayerScoreDisplay,x
	sbc TopScoreDisplay,y
	dex
	dey
	bpl GetScoreDifference
	bcc NoTopScore
	inx
	iny
CopyScore:
	lda PlayerScoreDisplay,x
	sta TopScoreDisplay,y
	inx
	iny
	cpy #$06
	bcc CopyScore
NoTopScore:
	rts

	.byte $ff
	.byte $ff
	.byte $04
	.byte $30
	.byte $48
	.byte $60
	.byte $78
	.byte $90
	.byte $a8
	.byte $c0
	.byte $d8
	.byte $e8
	.byte $24
	.byte $f8
	.byte $fc
	.byte $28
	.byte $2c
	ldy #$4b
	jsr $6f08
	ldx #$21
	lda #$00
	sta $0780,x
	dex
	bpl $6e42
	lda $075b
	ldy $0752
	beq $6e53
	lda $0751
	sta $071a
	sta $0725
	sta $0728
	jsr $7b90
	ldy #$20
	and #$01
	beq $6e67
	ldy #$24
	sty $0720
	ldy #$80
	sty $0721
	asl
	asl
	asl
	asl
	sta $06a0
	dec $0730
	dec $0731
	dec $0732
	lda #$0b
	sta $071e
	jsr $c2c3
	lda $07fb
	bne $6e9c
	lda $075f
	cmp #$03
	bcc $6e9f
	bne $6e9c
	lda $075c
	cmp #$03
	bcc $6e9f
	inc $06cc
	lda $075b
	beq $6ea9
	lda #$02
	sta $0710
	lda #$80
	sta $fb
	lda #$01
	sta $0774
	jsr $c241
	inc $0772
	rts

	lda #$00
	sta $0774
	sta $07f9
	sta $07f6
	tay
	sta $0300,y
	iny
	bne $6ec5
	sta $0759
	sta $0769
	sta $0728
	lda #$ff
	sta $03a0
	lda $071a
	and #$01
	sta $077a
	jsr $6f2d
	lda #$38
	sta $06e3
	lda #$48
	sta $06e2
	lda #$58
	sta $06e1
	ldx #$0e
	lda $6e2a,x
	sta $06e4,x
	dex
	bpl $6ef5
	jsr $70bb
	inc $0722
	inc $0772
	rts

InitializeMemory:
	ldx #$07
	lda #$00
	sta $06
InitDoNextPage:
	stx $07
InitDoNextY:
	cpx #$01
	bne InitDoNormal
	cpy #$60
	bcs InitGotoNextY
	cpy #$09
	bcc InitGotoNextY
InitDoNormal:
	sta ($06),y
InitGotoNextY:
	dey
	cpy #$ff
	bne InitDoNextY
	dex
	bpl InitDoNextPage
	rts

	.byte $02
	.byte $01
	.byte $04
	.byte $08
	.byte $10
	.byte $20
	lda $0770
	beq $6f55
	lda $0752
	cmp #$02
	beq $6f46
	ldy #$05
	lda $0710
	cmp #$06
	beq $6f50
	cmp #$07
	beq $6f50
	ldy $074e
	lda $0743
	beq $6f50
	ldy #$04
	lda $6f27,y
	sta $fb
	rts

	.byte $28
	.byte $18
	.byte $38
	.byte $28
	.byte $08
	.byte $00
	.byte $00
	.byte $20
	.byte $b0
	.byte $50
	.byte $00
	.byte $00
	.byte $b0
	.byte $b0
	.byte $f0
	.byte $00
	.byte $20
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $20
	.byte $04
	.byte $03
	.byte $02
	lda $071a
	sta $6d
	lda #$28
	sta $070a
	lda #$01
	sta $33
	sta $b5
	lda #$00
	sta $1d
	dec $0490
	ldy #$00
	sty $075b
	lda $074e
	bne $6f93
	iny
	sty $0704
	ldx $0710
	ldy $0752
	beq $6fa5
	cpy #$01
	beq $6fa5
	ldx $6f58,y
	lda $6f56,y
	sta $86
	lda $6f5c,x
	sta $ce
	lda $6f65,x
	sta $03c4
	jsr $654d
	ldy $0715
	beq $6fd7
	lda $0757
	beq $6fd7
	lda $6f6d,y
	sta $07ec
	lda #$01
	sta $07ee
	lsr
	sta $07ed
	sta $0757
	sta $079f
	ldy $0758
	beq $6ff0
	lda #$03
	sta $1d
	ldx #$00
	jsr $8945
	lda #$f0
	sta $d7
	ldx #$05
	ldy #$00
	jsr $84aa
	ldy $074e
	bne $6ff8
	jsr $8277
	lda #$07
	sta $0e
	rts

HalfwayPageNybbles:
	.byte $66, $60
	.byte $88, $60
	.byte $66, $70
	.byte $77, $60
	.byte $d6, $00
	.byte $77, $80
	.byte $70, $b0
	.byte $00, $00
	.byte $00, $00
	inc $0774
	lda #$00
	sta $0722
	lda #$80
	sta $fc
	dec $075a
	bpl $702b
	lda #$00
	sta $0772
	lda #$03
	sta $0770
	rts

	lda $075f
	asl
	tax
	lda $075c
	and #$02
	beq $7038
	inx
	ldy $6ffd,x
	lda $075c
	lsr
	tya
	bcs $7046
	lsr
	lsr
	lsr
	lsr
	and #$0f
	cmp $071a
	beq $7051
	bcc $7051
	lda #$00
	sta $075b
	jmp $709d

	lda $0772
	jsr $6c7d
	.byte $63
	.byte $70
	.byte $c5
	.byte $64
	.byte $79
	bvs $700d
	brk
	sta $073c
	sta $0722
	sta $07f8
	lda #$02
	sta $fc
	inc $0774
	inc $0772
	rts

	lda #$00
	sta $0774
	lda $075f
	cmp #$08
	beq $7088
	jmp $c1c2

EndTimerTest:
	lda ScreenTimer
	bne StillTimer
TerminateGame:
	lda #Silence
	sta EventMusicQueue
	lda #$00
	sta OperMode_Task
	sta ScreenTimer
	sta OperMode
StillTimer:
	rts

	jsr $c2a4
	lda #$01
	sta $0754
	inc $0757
	lda #$00
	sta $0747
	sta $0756
	sta $0e
	sta $0772
	lda #$01
	sta $0770
	rts

	lda #$ff
	sta $06c9
	rts

	ldy $071f
	bne $70cb
	ldy #$08
	sty $071f
	dey
	tya
	jsr $70d9
	dec $071f
	bne $70d8
	jsr $6847
	rts

	jsr $6c7d
	.byte $ec
	.byte $70
	.byte $8b
	.byte $67
	.byte $8b
	.byte $67
	.byte $0d
	.byte $72
	.byte $ec
	.byte $70
	.byte $8b
	.byte $67
	.byte $8b
	.byte $67
	.byte $0d
	.byte $72
	inc $0726
	lda $0726
	and #$0f
	bne $70fc
	sta $0726
	inc $0725
	inc $06a0
	lda $06a0
	and #$1f
	sta $06a0
	rts

	.byte $00
	.byte $30
	.byte $60
	.byte $93
	.byte $00
	.byte $00
	.byte $11
	.byte $12
	.byte $12
	.byte $13
	.byte $00
	.byte $00
	.byte $51
	.byte $52
	.byte $53
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $01
	.byte $02
	.byte $02
	.byte $03
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $91
	.byte $92
	.byte $93
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $51
	.byte $52
	.byte $53
	.byte $41
	.byte $42
	.byte $43
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $91
	.byte $92
	.byte $97
	.byte $87
	.byte $88
	.byte $89
	.byte $99
	.byte $00
	.byte $00
	.byte $00
	.byte $11
	.byte $12
	.byte $13
	.byte $a4
	.byte $a5
	.byte $a5
	.byte $a5
	.byte $a6
	.byte $97
	.byte $98
	.byte $99
	.byte $01
	.byte $02
	.byte $03
	.byte $00
	.byte $a4
	.byte $a5
	.byte $a6
	.byte $00
	.byte $11
	.byte $12
	.byte $12
	.byte $12
	.byte $13
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $01
	.byte $02
	.byte $02
	.byte $03
	.byte $00
	.byte $a4
	.byte $a5
	.byte $a5
	.byte $a6
	.byte $00
	.byte $00
	.byte $00
	.byte $11
	.byte $12
	.byte $12
	.byte $13
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $9c
	.byte $00
	.byte $8b
	.byte $aa
	.byte $aa
	.byte $aa
	.byte $aa
	.byte $11
	.byte $12
	.byte $13
	.byte $8b
	.byte $00
	.byte $9c
	.byte $9c
	.byte $00
	.byte $00
	.byte $01
	.byte $02
	.byte $03
	.byte $11
	.byte $12
	.byte $12
	.byte $13
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $aa
	.byte $aa
	.byte $9c
	.byte $aa
	.byte $00
	.byte $8b
	.byte $00
	.byte $01
	.byte $02
	.byte $03
	.byte $80
	.byte $83
	.byte $00
	.byte $81
	.byte $84
	.byte $00
	.byte $82
	.byte $85
	.byte $00
	.byte $02
	.byte $00
	.byte $00
	.byte $03
	.byte $00
	.byte $00
	.byte $04
	.byte $00
	.byte $00
	.byte $00
	.byte $05
	.byte $06
	.byte $07
	.byte $06
	.byte $0a
	.byte $00
	.byte $08
	.byte $09
	.byte $4d
	.byte $00
	.byte $00
	.byte $0d
	.byte $0f
	.byte $4e
	.byte $0e
	.byte $4e
	.byte $4e
	.byte $00
	.byte $0d
	.byte $1a
	.byte $86
	.byte $87
	.byte $87
	.byte $87
	.byte $87
	.byte $87
	.byte $87
	.byte $87
	.byte $87
	.byte $87
	.byte $87
	.byte $6a
	.byte $6a
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $45
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $86
	.byte $87
	.byte $6a
	.byte $6b
	.byte $50
	.byte $63
	.byte $00
	.byte $00
	.byte $00
	.byte $18
	.byte $01
	.byte $18
	.byte $07
	.byte $18
	.byte $0f
	.byte $18
	.byte $ff
	.byte $18
	.byte $01
	.byte $1f
	.byte $07
	.byte $1f
	.byte $0f
	.byte $1f
	.byte $81
	.byte $1f
	.byte $01
	.byte $00
	.byte $8f
	.byte $1f
	.byte $f1
	.byte $1f
	.byte $f9
	.byte $18
	.byte $f1
	.byte $18
	.byte $ff
	.byte $1f
	lda $0728
	beq $7215
	jsr $7319
	ldx #$0c
	lda #$00
	sta $06a1,x
	dex
	bpl $7219
	ldy $0742
	beq $7266
	lda $0725
	cmp #$03
	bmi $7230
	sec
	sbc #$03
	bpl $7227
	asl
	asl
	asl
	asl
	adc $7107,y
	adc $0726
	tax
	lda $710b,x
	beq $7266
	pha
	and #$0f
	sec
	sbc #$01
	sta $00
	asl
	adc $00
	tax
	pla
	lsr
	lsr
	lsr
	lsr
	tay
	lda #$03
	sta $00
	lda $719b,x
	sta $06a1,y
	inx
	iny
	cpy #$0b
	beq $7266
	dec $00
	bne $7256
	ldx $0741
	beq $727e
	ldy $71be,x
	ldx #$00
	lda $71c2,y
	beq $7278
	sta $06a1,x
	iny
	inx
	cpx #$0d
	bne $7270
	ldy $074e
	bne $728f
	lda $075f
	cmp #$07
	bne $728f
	lda #$63
	jmp $7299

	lda $71e9,y
	ldy $0743
	beq $7299
	lda #$88
	sta $07
	ldx #$00
	lda $0727
	asl
	tay
	lda $71ed,y
	sta $00
	iny
	sty $01
	lda $0743
	beq $72b9
	cpx #$00
	beq $72b9
	lda $00
	and #$08
	sta $00
	ldy #$00
	lda $929b,y
	bit $00
	beq $72c7
	lda $07
	sta $06a1,x
	inx
	cpx #$0d
	beq $72e4
	lda $074e
	cmp #$02
	bne $72db
	cpx #$0b
	bne $72db
	lda #$6b
	sta $07
	iny
	cpy #$08
	bne $72bb
	ldy $01
	bne $72a2
	jsr $7319
	lda $06a0
	jsr $7a22
	ldx #$00
	ldy #$00
	sty $00
	lda $06a1,x
	and #$c0
	asl
	rol
	rol
	tay
	lda $06a1,x
	cmp $7315,y
	bcs $7306
	lda #$00
	ldy $00
	sta ($06),y
	tya
	clc
	adc #$10
	tay
	inx
	cpx #$0d
	bcc $72f1
	rts

	.byte $10
	.byte $4f
	.byte $88
	.byte $c0
	ldx #$02
	stx $08
	lda #$00
	sta $0729
	ldy $072c
	lda ($e7),y
	cmp #$fd
	beq $7376
	lda $0730,x
	bpl $7376
	iny
	lda ($e7),y
	asl
	bcc $7341
	lda $072b
	bne $7341
	inc $072b
	inc $072a
	dey
	lda ($e7),y
	and #$0f
	cmp #$0d
	bne $7365
	iny
	lda ($e7),y
	dey
	and #$40
	bne $736e
	lda $072b
	bne $736e
	iny
	lda ($e7),y
	and #$1f
	sta $072a
	inc $072b
	jmp $737f

	cmp #$0e
	bne $736e
	lda $0728
	bne $7376
	lda $072a
	cmp $0725
	bcc $737c
	jsr $73a6
	jmp $7382

	inc $0729
	jsr $739a
	ldx $08
	lda $0730,x
	bmi $738c
	dec $0730,x
	dex
	bpl $731b
	lda $0729
	bne $7319
	lda $0728
	bne $7319
	rts

	inc $072c
	inc $072c
	lda #$00
	sta $072b
	rts

	lda $0730,x
	bmi $73ae
	ldy $072d,x
	ldx #$10
	lda ($e7),y
	cmp #$fd
	beq $7399
	and #$0f
	cmp #$0f
	beq $73c4
	ldx #$08
	cmp #$0c
	beq $73c4
	ldx #$00
	stx $07
	ldx $08
	cmp #$0e
	bne $73d4
	lda #$00
	sta $07
	lda #$36
	bne $7427
	cmp #$0d
	bne $73f3
	lda #$28
	sta $07
	iny
	lda ($e7),y
	and #$40
	beq $7446
	lda ($e7),y
	and #$7f
	cmp #$4b
	bne $73ee
	inc $0745
	and #$3f
	jmp $7427

	cmp #$0c
	bcs $741e
	iny
	lda ($e7),y
	and #$70
	bne $7409
	lda #$18
	sta $07
	lda ($e7),y
	and #$0f
	jmp $7427

	sta $00
	cmp #$70
	bne $7419
	lda ($e7),y
	and #$08
	beq $7419
	lda #$00
	sta $00
	lda $00
	jmp $7423

	iny
	lda ($e7),y
	and #$70
	lsr
	lsr
	lsr
	lsr
	sta $00
	lda $0730,x
	bpl $7470
	lda $072a
	cmp $0725
	beq $7447
	ldy $072c
	lda ($e7),y
	and #$0f
	cmp #$0e
	bne $7446
	lda $0728
	bne $7467
	rts

	lda $0728
	beq $7457
	lda #$00
	sta $0728
	sta $0729
	sta $08
	rts

	ldy $072c
	lda ($e7),y
	and #$f0
	lsr
	lsr
	lsr
	lsr
	cmp $0726
	bne $7446
	lda $072c
	sta $072d,x
	jsr $739a
	lda $00
	clc
	adc $07
	jsr $6c7d
	.byte $31
	.byte $77
	.byte $97
	.byte $75
	.byte $75
	.byte $78
	.byte $85
	.byte $78
	.byte $39
	.byte $78
	.byte $97
	.byte $78
	.byte $a0
	.byte $78
	.byte $31
	.byte $77
	.byte $8a
	.byte $79
	.byte $06
	.byte $76
	.byte $c0
	.byte $77
	.byte $c3
	.byte $77
	.byte $c6
	.byte $77
	.byte $9e
	.byte $77
	.byte $af
	.byte $77
	.byte $b2
	.byte $77
	.byte $17
	.byte $78
	.byte $1e
	.byte $78
	.byte $52
	.byte $76
	.byte $fe
	.byte $78
	.byte $f7
	.byte $76
	.byte $db
	.byte $77
	.byte $70
	.byte $c4
	.byte $75
	.byte $c4
	.byte $57
	.byte $79
	.byte $57
	.byte $79
	.byte $57
	.byte $79
	.byte $57
	.byte $79
	.byte $4a
	.byte $79
	.byte $57
	.byte $79
	.byte $57
	.byte $79
	.byte $62
	.byte $79
	.byte $62
	.byte $79
	.byte $62
	.byte $79
	.byte $62
	.byte $79
	.byte $5d
	.byte $79
	.byte $62
	.byte $79
	.byte $bb
	.byte $76
	.byte $60
	.byte $78
	.byte $1a
	.byte $79
	.byte $ce
	.byte $76
	.byte $e5
	.byte $77
	.byte $50
	.byte $78
	.byte $55
	.byte $78
	.byte $48
	.byte $78
	.byte $13
	.byte $75
	.byte $64
	.byte $75
	.byte $64
	.byte $75
	.byte $82
	.byte $75
	.byte $82
	.byte $75
	.byte $82
	.byte $75
	.byte $56
	.byte $74
	.byte $be
	.byte $c5
	.byte $c2
	.byte $c5
	.byte $e6
	.byte $74
	ldy $072d,x
	iny
	lda ($e7),y
	pha
	and #$40
	bne $7503
	pla
	pha
	and #$0f
	sta $0727
	pla
	and #$30
	lsr
	lsr
	lsr
	lsr
	sta $0742
	rts

	pla
	and #$07
	cmp #$04
	bcc $750f
	sta $0744
	lda #$00
	sta $0741
	rts

	ldx #$80
	lda $07fb
	bne $752f
	lda $075f
	bne $7537
	ldy $074e
	dey
	beq $752b
	lda $074f
	beq $752c
	inx
	inx
	jmp $7558

	lda #$87
	clc
	adc $075c
	bne $7559
	ldx #$83
	lda $075f
	cmp #$02
	beq $7558
	inx
	cmp #$04
	bne $7555
	lda $074f
	cmp #$0b
	beq $7558
	ldy $074e
	dey
	beq $7556
	jmp $7557

	inx
	inx
	inx
	txa
	sta $06d6
	jsr $675e
	lda #$0d
	jsr $756d
	lda $0723
	eor #$01
	sta $0723
	rts

	sta $00
	lda #$00
	ldx #$04
	ldy $16,x
	cpy $00
	bne $757b
	sta $0f,x
	dex
	bpl $7573
	rts

	.byte $14
	.byte $17
	.byte $18
	ldx $00
	lda $7577,x
	ldy #$05
	dey
	bmi $7593
	cmp $0016,y
	bne $7589
	lda #$00
	sta $06cd
	rts

	lda $0733
	jsr $6c7d
	.byte $a3
	.byte $75
	.byte $cf
	.byte $75
	.byte $b0
	.byte $78
	jsr $79fc
	lda $0730,x
	beq $75ca
	bpl $75be
	tya
	sta $0730,x
	lda $0725
	ora $0726
	beq $75be
	lda #$16
	jmp $75fc

	ldx $07
	lda #$17
	sta $06a1,x
	lda #$4c
	jmp $75f6

	lda #$18
	jmp $75fc

	jsr $79ed
	sty $06
	bcc $75e2
	lda $0730,x
	lsr
	sta $0736,x
	lda #$8a
	jmp $75fc

	lda #$8c
	ldy $0730,x
	beq $75fc
	lda $0736,x
	sta $06
	ldx $07
	lda #$8b
	sta $06a1,x
	rts

	inx
	ldy #$0f
	jmp $79c6

	ldx $07
	ldy #$00
	jmp $79c6

	.byte $42
	.byte $41
	.byte $43
	jsr $79ed
	ldy #$00
	bcs $7614
	iny
	lda $0730,x
	bne $7614
	iny
	lda $7603,y
	sta $06a1
	rts

	.byte $00
	.byte $45
	.byte $45
	.byte $45
	.byte $00
	.byte $00
	.byte $48
	.byte $47
	.byte $46
	.byte $00
	.byte $45
	.byte $49
	.byte $49
	.byte $49
	.byte $45
	.byte $47
	.byte $47
	.byte $4a
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $4b
	.byte $47
	.byte $47
	.byte $49
	.byte $49
	.byte $49
	.byte $49
	.byte $49
	.byte $47
	.byte $4a
	.byte $47
	.byte $4a
	.byte $47
	.byte $47
	.byte $4b
	.byte $47
	.byte $4b
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $47
	.byte $4a
	.byte $47
	.byte $4a
	.byte $47
	.byte $4a
	.byte $4b
	.byte $47
	.byte $4b
	.byte $47
	.byte $4b
	jsr $79fc
	sty $07
	ldy #$04
	jsr $79f0
	txa
	pha
	ldy $0730,x
	ldx $07
	lda #$0b
	sta $06
	lda $761b,y
	sta $06a1,x
	inx
	lda $06
	beq $7679
	iny
	iny
	iny
	iny
	iny
	dec $06
	cpx #$0b
	bne $7667
	pla
	tax
	lda $0725
	beq $76ba
	lda $0730,x
	cmp #$01
	beq $76b5
	ldy $07
	bne $7693
	cmp #$03
	beq $76b5
	cmp #$02
	bne $76ba
	jsr $7a0c
	pha
	jsr $7791
	pla
	sta $87,x
	lda $0725
	sta $6e,x
	lda #$01
	sta $b6,x
	sta $0f,x
	lda #$90
	sta $cf,x
	lda #$31
	sta $16,x
	rts

	ldy #$50
	sty $06ab
	rts

	jsr $79fc
	ldy $0730,x
	ldx $07
	lda #$6d
	sta $06a1,x
	lda #$6e
	sta $06a2,x
	rts

	ldy #$03
	jsr $79f0
	ldy #$0a
	jsr $76ff
	bcs $76ea
	ldx #$06
	lda #$00
	sta $06a1,x
	dex
	bpl $76dc
	lda $7729,y
	sta $06a8
	rts

	.byte $15
	.byte $14
	.byte $00
	.byte $00
	.byte $15
	.byte $1b
	.byte $1a
	.byte $19
	.byte $15
	.byte $1e
	.byte $1d
	.byte $1c
	ldy #$03
	jsr $79f0
	jsr $79fc
	dey
	dey
	sty $05
	ldy $0730,x
	sty $06
	ldx $05
	inx
	lda $76eb,y
	cmp #$00
	beq $771a
	ldx #$00
	ldy $05
	jsr $79c6
	clc
	ldy $06
	lda $76ef,y
	sta $06a1,x
	lda $76f3,y
	sta $06a2,x
	rts

	.byte $11
	.byte $10
	.byte $15
	.byte $14
	.byte $13
	.byte $12
	.byte $15
	.byte $14
	jsr $7761
	lda $00
	beq $773c
	iny
	iny
	iny
	iny
	tya
	pha
	ldy $0730,x
	beq $774d
	jsr $7791
	bcs $774d
	lda #$0d
	jsr $7772
	pla
	tay
	ldx $07
	lda $7729,y
	sta $06a1,x
	inx
	lda $772b,y
	ldy $06
	dey
	jmp $79c6

	ldy #$01
	jsr $79f0
	jsr $79fc
	tya
	and #$07
	sta $06
	ldy $0730,x
	rts

	sta $16,x
	jsr $7a0c
	clc
	adc #$08
	sta $87,x
	lda $0725
	adc #$00
	sta $6e,x
	lda #$01
	sta $b6,x
	sta $0f,x
	jsr $7a14
	sta $cf,x
	jmp $9398

	ldx #$00
	clc
	lda $0f,x
	beq $779d
	inx
	cpx #$05
	bne $7793
	rts

	jsr $79ed
	lda #$86
	sta $06ab
	ldx #$0b
	ldy #$01
	lda #$87
	jmp $79c6

	lda #$03
	bit $07a9
	pha
	jsr $79ed
	pla
	tax
	lda #$c0
	sta $06a1,x
	rts

	.byte $a9
	.byte $06
	.byte $2c
	lda #$07
	bit $09a9
	pha
	jsr $79ed
	pla
	tax
	lda #$0b
	sta $06a1,x
	inx
	ldy #$00
	lda #$64
	jmp $79c6

	jsr $79fc
	ldx #$02
	lda #$6f
	jmp $79c6

	lda #$21
	sta $06a1
	ldx #$01
	ldy #$08
	lda #$22
	jsr $79c6
	lda #$62
	sta $06ab
	jsr $7a0c
	sec
	sbc #$08
	sta $8c
	lda $0725
	sbc #$00
	sta $73
	lda #$30
	sta $d4
	lda #$b0
	sta $010d
	lda #$30
	sta $1b
	inc $14
	rts

	ldx #$00
	ldy #$0f
	jmp $7830

	txa
	pha
	ldx #$01
	ldy #$0f
	lda #$44
	jsr $79c6
	pla
	tax
	jsr $79fc
	ldx #$01
	lda #$40
	jmp $79c6

	.byte $c4
	.byte $c3
	.byte $c3
	.byte $c3
	ldy $074e
	lda $7835,y
	jmp $788b

	.byte $06
	.byte $07
	.byte $08
	.byte $c6
	.byte $0c
	.byte $89
	ldy #$0c
	jsr $79f0
	jmp $7855

	lda #$08
	sta $0773
	ldy $00
	ldx $7840,y
	lda $7843,y
	jmp $7867

	jsr $79fc
	ldx $07
	lda #$c5
	ldy #$00
	jmp $79c6

	.byte $6a
	.byte $62
	.byte $62
	.byte $63
	.byte $1f
	.byte $4f
	.byte $50
	.byte $50
	.byte $88
	ldy $074e
	lda $0743
	beq $787f
	ldy #$04
	lda $7870,y
	jmp $788b

	ldy $074e
	lda $786c,y
	pha
	jsr $79ed
	ldx $07
	ldy #$00
	pla
	jmp $79c6

	ldy $074e
	lda $7870,y
	jmp $78a6

	ldy $074e
	lda $786c,y
	pha
	jsr $79fc
	pla
	ldx $07
	jmp $79c6

	jsr $79fc
	ldx $07
	lda #$65
	sta $06a1,x
	inx
	dey
	bmi $78cc
	lda #$66
	sta $06a1,x
	inx
	dey
	bmi $78cc
	lda #$67
	jsr $79c6
	ldx $046a
	jsr $7a14
	sta $0477,x
	lda $0725
	sta $046b,x
	jsr $7a0c
	sta $0471,x
	inx
	cpx #$06
	bcc $78e8
	ldx #$00
	stx $046a
	rts

	.byte $07
	.byte $07
	.byte $06
	.byte $05
	.byte $04
	.byte $03
	.byte $02
	.byte $01
	.byte $00
	.byte $03
	.byte $03
	.byte $04
	.byte $05
	.byte $06
	.byte $07
	.byte $08
	.byte $09
	.byte $0a
	jsr $79ed
	bcc $7908
	lda #$09
	sta $0734
	dec $0734
	ldy $0734
	ldx $78f5,y
	lda $78ec,y
	tay
	lda #$62
	jmp $79c6

	jsr $79fc
	jsr $7791
	bcs $7949
	jsr $7a0c
	sta $87,x
	lda $0725
	sta $6e,x
	jsr $7a14
	sta $cf,x
	sta $58,x
	lda #$32
	sta $16,x
	ldy #$01
	sty $b6,x
	inc $0f,x
	ldx $07
	lda #$68
	sta $06a1,x
	lda #$69
	sta $06a2,x
	rts

	lda $075d
	beq $7985
	lda #$00
	sta $075d
	jmp $7962

	jsr $797f
	jmp $7975

	lda #$00
	sta $06bc
	jsr $797f
	sty $07
	lda #$00
	ldy $074e
	dey
	beq $7971
	lda #$06
	clc
	adc $07
	tay
	lda $89b4,y
	pha
	jsr $79fc
	jmp $788f

	lda $00
	sec
	sbc #$00
	tay
	rts

	.byte $87
	.byte $00
	.byte $00
	.byte $00
	jsr $79ed
	bcc $79bc
	lda $074e
	bne $79bc
	ldx $046a
	jsr $7a0c
	sec
	sbc #$10
	sta $0471,x
	lda $0725
	sbc #$00
	sta $046b,x
	iny
	iny
	tya
	asl
	asl
	asl
	asl
	sta $0477,x
	inx
	cpx #$05
	bcc $79b9
	ldx #$00
	stx $046a
	ldx $074e
	lda $7986,x
	ldx #$08
	ldy #$0f
	sty $0735
	ldy $06a1,x
	beq $79de
	cpy #$17
	beq $79e1
	cpy #$8b
	beq $79e1
	cpy #$c0
	beq $79de
	cpy #$c0
	bcs $79e1
	sta $06a1,x
	inx
	cpx #$0d
	bcs $79ec
	ldy $0735
	dey
	bpl $79c6
	rts

	jsr $79fc
	lda $0730,x
	clc
	bpl $79fb
	tya
	sta $0730,x
	sec
	rts

	ldy $072d,x
	lda ($e7),y
	and #$0f
	sta $07
	iny
	lda ($e7),y
	and #$0f
	tay
	rts

	lda $0726
	asl
	asl
	asl
	asl
	rts

	lda $07
	asl
	asl
	asl
	asl
	clc
	adc #$20
	rts

	.byte $00
	.byte $d0
	.byte $05
	.byte $05
	pha
	lsr
	lsr
	lsr
	lsr
	tay
	lda $7a20,y
	sta $07
	pla
	and #$0f
	clc
	adc $7a1e,y
	sta $06
	rts

	lda $0772
	jsr $6c7d
	.byte $4d
	.byte $c0
	.byte $39
	.byte $6e
	.byte $c5
	.byte $64
	.byte $b9
	.byte $6e
	.byte $47
	.byte $7a
	jsr $7ba2
	lda $0772
	cmp #$04
	bcs $7a52
	rts

	jsr $8190
	ldx #$00
	stx $08
	jsr $8c23
	jsr $6421
	inx
	cpx #$06
	bne $7a57
	jsr $be65
	jsr $be0f
	jsr $bbc4
	jsr $8aa5
	ldx #$01
	stx $08
	jsr $8a41
	dex
	stx $08
	jsr $8a41
	jsr $8761
	jsr $8587
	jsr $8321
	jsr $83be
	jsr $82bb
	jsr $68be
	lda $07f7
	beq $7a97
	jsr $c550
	lda $b5
	cmp #$02
	bpl $7aae
	lda $079f
	beq $7ac0
	cmp #$04
	bne $7aae
	lda $077f
	bne $7aae
	jsr $6f2d
	ldy $079f
	lda $09
	cpy #$08
	bcs $7ab9
	lsr
	lsr
	lsr
	jsr $7de0
	jmp $7ac3

	jsr $7df2
	lda $0a
	sta $0d
	lda #$00
	sta $0c
	lda $0773
	cmp #$06
	beq $7aee
	lda $071f
	bne $7aeb
	lda $073d
	cmp #$20
	bmi $7aee
	lda $073d
	sbc #$20
	sta $073d
	lda #$00
	sta $0340
	jsr $70c1
	rts

	lda $06ff
	clc
	adc $03a1
	sta $06ff
	lda $0723
	bne $7b53
	lda $0755
	cmp #$50
	bcc $7b53
	lda $0785
	bne $7b53
	ldy $06ff
	dey
	bmi $7b53
	iny
	cpy #$02
	bcc $7b16
	dey
	lda $0755
	cmp #$70
	bcc $7b20
	ldy $06ff
	lda $077b
	bne $7b20
	tya
	sta $0775
	clc
	adc $073d
	sta $073d
	tya
	clc
	adc $071c
	sta $071c
	sta $073f
	lda $071a
	adc #$00
	sta $071a
	and #$01
	sta $077a
	jsr $7b90
	lda #$08
	sta $0795
	jmp $7b58

	lda #$00
	sta $0775
	ldx #$00
	jsr $bedb
	sta $00
	ldy #$00
	asl
	bcs $7b6b
	iny
	lda $00
	and #$20
	beq $7b86
	lda $071c,y
	sec
	sbc $7b8c,y
	sta $86
	lda $071a,y
	sbc #$00
	sta $6d
	lda $0c
	cmp $7b8e,y
	beq $7b86
	lda #$00
	sta $57
	lda #$00
	sta $03a1
	rts

	.byte $00
	.byte $10
	.byte $01
	.byte $02
	lda $071c
	clc
	adc #$ff
	sta $071d
	lda $071a
	adc #$00
	sta $071b
	rts

	lda $0e
	jsr $6c7d
	.byte $71
	.byte $6f
	.byte $1f
	.byte $7d
	.byte $5e
	.byte $7d
	.byte $3d
	.byte $7d
	.byte $fb
	.byte $7d
	.byte $19
	.byte $7e
	.byte $0f
	.byte $70
	.byte $c1
	.byte $7b
	.byte $41
	.byte $7c
	.byte $8b
	.byte $7d
	.byte $9d
	.byte $7d
	.byte $c1
	.byte $7d
	.byte $d5
	.byte $7d
	lda $0752
	cmp #$02
	beq $7bf3
	lda #$00
	ldy $ce
	cpy #$30
	bcc $7c3e
	lda $0710
	cmp #$06
	beq $7bdb
	cmp #$07
	bne $7c2b
	lda $03c4
	bne $7be5
	lda #$01
	jmp $7c3e

	jsr $7d77
	dec $06de
	bne $7c3d
	inc $0769
	jmp $7e66

	lda $0758
	bne $7c04
	lda #$ff
	jsr $7d58
	lda $ce
	cmp #$91
	bcc $7c2b
	rts

	lda $0399
	cmp #$60
	bne $7c3d
	lda $ce
	cmp #$99
	ldy #$00
	lda #$01
	bcc $7c1f
	lda #$03
	sta $1d
	iny
	lda #$08
	sta $05b4
	sty $0716
	jsr $7c3e
	lda $86
	cmp #$48
	bcc $7c3d
	lda #$08
	sta $0e
	lda #$01
	sta $33
	lsr
	sta $0752
	sta $0716
	sta $0758
	rts

	sta $06fc
	lda $0e
	cmp #$0b
	beq $7c83
	lda $074e
	bne $7c5c
	ldy $b5
	dey
	bne $7c57
	lda $ce
	cmp #$d0
	bcc $7c5c
	lda #$00
	sta $06fc
	lda $06fc
	and #$c0
	sta $0a
	lda $06fc
	and #$03
	sta $0c
	lda $06fc
	and #$0c
	sta $0b
	and #$04
	beq $7c83
	lda $1d
	bne $7c83
	ldy $0c
	beq $7c83
	lda #$00
	sta $0c
	sta $0b
	jsr $7e90
	ldy #$01
	lda $0754
	bne $7c96
	ldy #$00
	lda $0714
	beq $7c96
	ldy #$02
	sty $0499
	lda #$01
	ldy $57
	beq $7ca4
	bpl $7ca2
	asl
	sta $45
	jsr $7aef
	jsr $be65
	jsr $be0f
	ldx #$00
	jsr $af3a
	jsr $a8d2
	lda $ce
	cmp #$40
	bcc $7cd1
	lda $0e
	cmp #$05
	beq $7cd1
	cmp #$07
	beq $7cd1
	cmp #$04
	bcc $7cd1
	lda $03c4
	and #$df
	sta $03c4
	lda $b5
	cmp #$02
	bmi $7d12
	ldx #$01
	stx $0723
	ldy #$04
	sty $07
	ldx #$00
	ldy $0759
	bne $7cec
	ldy $0743
	bne $7d02
	inx
	ldy $0e
	cpy #$0b
	beq $7d02
	ldy $0712
	bne $7cfe
	iny
	sty $fc
	sty $0712
	ldy #$06
	sty $07
	cmp $07
	bmi $7d12
	dex
	bmi $7d13
	ldy $07b1
	bne $7d12
	lda #$06
	sta $0e
	rts

	lda #$00
	sta $0758
	jsr $7d35
	inc $0752
	rts

	lda $b5
	bne $7d29
	lda $ce
	cmp #$e4
	bcc $7d35
	lda #$08
	sta $0758
	ldy #$03
	sty $1d
	jmp $7c3e

	lda #$02
	sta $0752
	jmp $7d6b

	lda #$01
	jsr $7d58
	jsr $7aef
	ldy #$00
	lda $06d6
	bne $7d63
	iny
	lda $074e
	cmp #$03
	bne $7d63
	iny
	jmp $7d63

	clc
	adc $ce
	sta $ce
	rts

	jsr $7d77
	ldy #$02
	dec $06de
	bne $7d76
	sty $0752
	inc $0774
	lda #$00
	sta $0772
	sta $0722
	rts

	lda #$08
	sta $57
	ldy #$01
	lda $86
	and #$0f
	bne $7d86
	sta $57
	tay
	tya
	jsr $7c3e
	rts

	lda $0747
	cmp #$f8
	bne $7d95
	jmp $7dad

	cmp #$c4
	bne $7d9c
	jsr $7dcb
	rts

	lda $0747
	cmp #$f0
	bcs $7dab
	cmp #$c8
	beq $7dcb
	jmp $7c41

	bne $7dc0
	ldy $070b
	bne $7dc0
	sty $070d
	inc $070b
	lda $0754
	eor #$01
	sta $0754
	rts

	lda $0747
	cmp #$f0
	bcs $7dfa
	jmp $7c41

	lda #$00
	sta $0747
	lda #$08
	sta $0e
	rts

	lda $0747
	cmp #$c0
	beq $7def
	lda $09
	lsr
	lsr
	and #$03
	sta $00
	lda $03c4
	and #$fc
	ora $00
	sta $03c4
	rts

	jsr $7dcb
	lda $03c4
	and #$fc
	sta $03c4
	rts

	lda $1b
	cmp #$30
	bne $7e16
	lda $0713
	sta $ff
	lda #$00
	sta $0713
	ldy $ce
	cpy #$9e
	bcs $7e13
	lda #$04
	jmp $7c3e

	inc $0e
	rts

	lda #$01
	jsr $7c3e
	lda $ce
	cmp #$ae
	bcc $7e35
	lda #$00
	sta $0723
	lda $07f6
	bne $7e35
	lda #$20
	sta $fc
	inc $07f6
	lda $0490
	lsr
	bcs $7e48
	lda $0746
	bne $7e43
	inc $0746
	lda #$20
	sta $03c4
	lda $0746
	cmp #$05
	bne $7e8f
	inc $075c
	lda $075c
	cmp #$03
	bne $7e66
	ldy $075f
	lda $0748
	cmp #$0a
	bcc $7e66
	inc $075d
	inc $0760
	lda $075f
	cmp #$08
	bne $7e7f
	lda $075c
	cmp #$04
	bne $7e7f
	lda #$00
	sta $075c
	sta $0760
	jsr $c2a4
	inc $0757
	jsr $7d6b
	sta $075b
	lda #$80
	sta $fc
	rts

	lda #$00
	ldy $0754
	bne $7e9f
	lda $1d
	bne $7ea2
	lda $0b
	and #$04
	sta $0714
	jsr $7fbc
	lda $070b
	bne $7ec0
	lda $1d
	cmp #$03
	beq $7eb5
	ldy #$18
	sty $0789
	jsr $6c7d
	.byte $c1
	.byte $7e
	.byte $e5
	.byte $7e
	.byte $dc
	.byte $7e
	.byte $3b
	.byte $7f
	rts

	jsr $80fb
	lda $0c
	beq $7eca
	sta $33
	jsr $8138
	jsr $8ada
	sta $06ff
	lda $07f7
	beq $7edb
	jsr $c4fe
	rts

	lda $070a
	sta $0709
	jmp $7f1b

	ldy $9f
	bpl $7efc
	lda $0a
	and #$80
	and $0d
	bne $7f02
	lda $0708
	sec
	sbc $ce
	cmp $0706
	bcc $7f02
	lda $070a
	sta $0709
	lda $0704
	beq $7f1b
	jsr $80fb
	lda $ce
	cmp #$14
	bcs $7f15
	lda #$18
	sta $0709
	lda $0c
	beq $7f1b
	sta $33
	lda $0c
	beq $7f22
	jsr $8138
	jsr $7ecd
	lda $0e
	cmp #$0b
	bne $7f30
	lda #$28
	sta $0709
	jmp $8b1e

	.byte $0e
	.byte $04
	.byte $fc
	.byte $f2
	.byte $00
	.byte $00
	.byte $ff
	.byte $ff
	lda $0416
	clc
	adc $0433
	sta $0416
	ldy #$00
	lda $9f
	bpl $7f4c
	dey
	sty $00
	adc $ce
	sta $ce
	lda $b5
	adc $00
	sta $b5
	lda $0c
	and $0490
	beq $7f8c
	ldy $0789
	bne $7f8b
	ldy #$18
	sty $0789
	ldx #$00
	ldy $33
	lsr
	bcs $7f72
	.byte $e8
	.byte $e8
	dey
	beq $7f76
	inx
	lda $86
	clc
	adc $7f33,x
	sta $86
	lda $6d
	adc $7f37,x
	sta $6d
	lda $0c
	eor #$03
	sta $33
	rts

	sta $0789
	rts

	.byte $20
	.byte $20
	.byte $1e
	.byte $28
	.byte $28
	.byte $0d
	.byte $04
	.byte $70
	.byte $70
	.byte $60
	.byte $90
	.byte $90
	.byte $0a
	.byte $09
	.byte $e4
	.byte $98
	.byte $d0
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fb
	.byte $fb
	.byte $fe
	.byte $ff
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $80
	.byte $00
	.byte $d8
	.byte $e8
	.byte $f0
	.byte $28
	.byte $18
	.byte $10
	.byte $0c
	.byte $00
	.byte $ff
	.byte $01
	.byte $00
	.byte $20
	.byte $ff
	lda $1d
	cmp #$03
	bne $7fe5
	ldy #$00
	lda $0b
	and $0490
	beq $7fd1
	iny
	and #$08
	bne $7fd1
	iny
	ldx $7fb9,y
	stx $0433
	lda #$08
	ldx $7fb6,y
	stx $9f
	bmi $7fe1
	lsr
	sta $070c
	rts

	lda $070e
	bne $7ff4
	lda $0a
	and #$80
	beq $7ff4
	and $0d
	beq $7ff7
	jmp $8088

	lda $1d
	beq $800c
	lda $0704
	beq $7ff4
	lda $0782
	bne $800c
	lda $9f
	bpl $800c
	jmp $8088

	lda #$20
	sta $0782
	ldy #$00
	sty $0416
	sty $0433
	lda $b5
	sta $0707
	lda $ce
	sta $0708
	lda #$01
	sta $1d
	lda $0700
	cmp #$09
	bcc $803e
	iny
	cmp #$10
	bcc $803e
	iny
	cmp #$19
	bcc $803e
	iny
	cmp #$1c
	bcc $803e
	iny
	lda #$01
	sta $0706
	lda $0704
	beq $8050
	ldy #$05
	lda $047d
	beq $8050
	iny
	lda $7f90,y
	sta $0709
	lda $7f97,y
	sta $070a
	lda $7fa8,y
	sta $0433
	lda $7fa1,y
	sta $9f
	lda $0704
	beq $807d
	lda #$04
	sta $ff
	lda $ce
	cmp #$14
	bcs $8088
	lda #$00
	sta $9f
	jmp $8088

	lda #$01
	ldy $0754
	beq $8086
	lda #$80
	sta $ff
	ldy #$00
	sty $00
	lda $1d
	beq $8099
	lda $0700
	cmp #$19
	bcs $80ca
	bcc $80b1
	iny
	lda $074e
	beq $80b1
	dey
	lda $0c
	cmp $45
	bne $80b1
	lda $0a
	and #$40
	bne $80c5
	lda $0783
	bne $80ca
	iny
	inc $00
	lda $0703
	bne $80c0
	lda $0700
	cmp #$21
	bcc $80ca
	inc $00
	jmp $80ca

	lda #$0a
	sta $0783
	lda $7faf,y
	sta $0450
	lda $0e
	cmp #$07
	bne $80d8
	ldy #$03
	lda $7fb2,y
	sta $0456
	ldy $00
	lda $7f9e,y
	sta $0702
	lda #$00
	sta $0701
	lda $33
	cmp $45
	beq $80f7
	asl $0702
	rol $0701
	rts

	.byte $02
	.byte $04
	.byte $07
	ldy #$00
	lda $0700
	cmp #$1c
	bcs $8119
	iny
	cmp #$0e
	bcs $810a
	iny
	lda $06fc
	and #$7f
	beq $8131
	and #$03
	cmp $45
	bne $811f
	lda #$00
	sta $0703
	jmp $8131

	lda $0700
	cmp #$0b
	bcs $8131
	lda $33
	sta $45
	lda #$00
	sta $57
	sta $0705
	lda $80f8,y
	sta $070c
	rts

	and $0490
	cmp #$00
	bne $8147
	lda $57
	beq $818c
	bpl $8168
	bmi $814a
	lsr
	bcc $8168
	lda $0705
	clc
	adc $0702
	sta $0705
	lda $57
	adc $0701
	sta $57
	cmp $0456
	bmi $8183
	lda $0456
	sta $57
	jmp $818c

	lda $0705
	sec
	sbc $0702
	sta $0705
	lda $57
	sbc $0701
	sta $57
	cmp $0450
	bpl $8183
	lda $0450
	sta $57
	cmp #$00
	bpl $818c
	eor #$ff
	clc
	adc #$01
	sta $0700
	rts

	lda $0756
	cmp #$02
	bcc $81da
	lda $0a
	and #$40
	beq $81d0
	and $0d
	bne $81d0
	lda $06ce
	and #$01
	tax
	lda $24,x
	bne $81d0
	ldy $b5
	dey
	bne $81d0
	lda $0714
	bne $81d0
	lda $1d
	cmp #$03
	beq $81d0
	lda #$20
	sta $ff
	lda #$02
	sta $24,x
	ldy $070c
	sty $0711
	dey
	sty $0781
	inc $06ce
	ldx #$00
	jsr $81f5
	ldx #$01
	jsr $81f5
	lda $074e
	bne $81f2
	ldx #$02
	stx $08
	jsr $8265
	jsr $be16
	jsr $be76
	jsr $babc
	dex
	bpl $81e1
	rts

	.byte $40
	.byte $c0
	stx $08
	lda $24,x
	asl
	bcs $825f
	ldy $24,x
	beq $825e
	dey
	beq $822a
	lda $86
	adc #$04
	sta $8d,x
	lda $6d
	adc #$00
	sta $74,x
	lda $ce
	sta $d5,x
	lda #$01
	sta $bc,x
	ldy $33
	dey
	lda $81f3,y
	sta $5e,x
	lda #$04
	sta $a6,x
	lda #$07
	sta $04a0,x
	dec $24,x
	txa
	clc
	adc #$07
	tax
	lda #$50
	sta $00
	lda #$03
	sta $02
	lda #$00
	jsr $8ba8
	jsr $8ae0
	ldx $08
	jsr $be20
	jsr $be6c
	jsr $aecb
	jsr $ae66
	lda $03d2
	and #$cc
	bne $825a
	jsr $a319
	jmp $b9b9

	lda #$00
	sta $24,x
	rts

	jsr $be20
	jmp $b9e4

	lda $07a8,x
	and #$01
	sta $07
	lda $e4,x
	cmp #$f8
	bne $829e
	lda $0792
	bne $82b6
	ldy #$00
	lda $33
	lsr
	bcc $8280
	ldy #$08
	tya
	adc $86
	sta $9c,x
	lda $6d
	adc #$00
	sta $83,x
	lda $ce
	clc
	adc #$08
	sta $e4,x
	lda #$01
	sta $cb,x
	ldy $07
	lda $82b9,y
	sta $0792
	ldy $07
	lda $042c,x
	sec
	sbc $82b7,y
	sta $042c,x
	lda $e4,x
	sbc #$00
	cmp #$20
	bcs $82b4
	lda #$f8
	sta $e4,x
	rts

	.byte $ff
	.byte $50
	.byte $40
	.byte $20
	lda $0770
	beq $830f
	lda $0e
	cmp #$08
	bcc $830f
	cmp #$0b
	beq $830f
	lda $b5
	cmp #$02
	bpl $830f
	lda $0787
	bne $830f
	lda $07ec
	ora $07ed
	ora $07ee
	beq $8306
	ldy $07ec
	dey
	bne $82f2
	lda $07ed
	ora $07ee
	bne $82f2
	lda #$40
	sta $fc
	lda #$18
	sta $0787
	ldy #$17
	lda #$ff
	sta $0139
	jsr $6dd2
	lda #$a2
	jmp $6d79

	sta $0756
	jsr $a58f
	inc $0759
	rts

	lda $0723
	beq $830f
	lda $ce
	and $b5
	bne $830f
	sta $0723
	jmp $95cd

	lda $074e
	bne $835d
	sta $047d
	lda $0747
	bne $835d
	ldy #$04
	lda $0471,y
	clc
	adc $0477,y
	sta $02
	lda $046b,y
	beq $835a
	adc #$00
	sta $01
	lda $86
	sec
	sbc $0471,y
	lda $6d
	sbc $046b,y
	bmi $835a
	lda $02
	sec
	sbc $86
	lda $01
	sbc $6d
	bpl $835e
	dey
	bpl $8330
	rts

	lda $0477,y
	lsr
	sta $00
	lda $0471,y
	clc
	adc $00
	sta $01
	lda $046b,y
	adc #$00
	sta $00
	lda $09
	lsr
	bcc $83a4
	lda $01
	sec
	sbc $86
	lda $00
	sbc $6d
	bpl $8391
	lda $86
	sec
	sbc #$01
	sta $86
	lda $6d
	sbc #$00
	jmp $83a2

	lda $0490
	lsr
	bcc $83a4
	lda $86
	clc
	adc #$01
	sta $86
	lda $6d
	adc #$00
	sta $6d
	lda #$10
	sta $00
	lda #$01
	sta $047d
	sta $02
	lsr
	tax
	jmp $8ba8

	.byte $05
	.byte $02
	.byte $08
	.byte $04
	.byte $01
	.byte $03
	.byte $03
	.byte $04
	.byte $04
	.byte $04
	ldx #$05
	stx $08
	lda $16,x
	cmp #$30
	bne $842c
	lda $0e
	cmp #$04
	bne $83ff
	lda $1d
	cmp #$03
	bne $83ff
	lda $cf,x
	cmp #$aa
	bcs $8402
	lda $ce
	cmp #$a2
	bcs $8402
	lda $0417,x
	adc #$ff
	sta $0417,x
	lda $cf,x
	adc #$01
	sta $cf,x
	lda $010e
	sec
	sbc #$ff
	sta $010e
	lda $010d
	sbc #$01
	sta $010d
	jmp $8423

	ldy $010f
	cpy #$05
	bne $8413
	inc $075a
	lda #$40
	sta $fe
	jmp $841f

	lda $83b4,y
	ldx $83b9,y
	sta $0134,x
	jsr $87e8
	lda #$05
	sta $0e
	jsr $be94
	jsr $be37
	jsr $b1f1
	rts

	.byte $08
	.byte $10
	.byte $08
	.byte $00
	jsr $be94
	lda $0747
	bne $848e
	lda $070e
	beq $848e
	tay
	dey
	tya
	and #$02
	bne $844c
	inc $ce
	inc $ce
	jmp $8450

	dec $ce
	dec $ce
	lda $58,x
	clc
	adc $842d,y
	sta $cf,x
	cpy #$01
	bcc $8480
	lda $0a
	and #$80
	beq $8480
	and $0d
	bne $8480
	tya
	pha
	lda #$f4
	ldy $075f
	cpy #$01
	beq $8479
	cpy #$02
	beq $8479
	cpy #$06
	bne $847b
	lda #$e0
	sta $06db
	pla
	tay
	cpy #$03
	bne $848e
	lda $06db
	sta $9f
	lda #$00
	sta $070e
	jsr $be37
	jsr $b52c
	jsr $a2b4
	lda $070e
	beq $84a9
	lda $0786
	bne $84a9
	lda #$04
	sta $0786
	inc $070e
	rts

	lda #$2f
	sta $16,x
	lda #$01
	sta $0f,x
	lda $0076,y
	sta $6e,x
	lda $008f,y
	sta $87,x
	lda $00d7,y
	sta $cf,x
	ldy $0398
	bne $84c9
	sta $039d
	txa
	sta $039a,y
	inc $0398
	lda #$04
	sta $fe
	rts

	.byte $30
	.byte $60
	cpx #$05
	beq $84dc
	rts

	ldy $0398
	dey
	lda $0399
	cmp $84d5,y
	beq $84f7
	lda $09
	lsr
	lsr
	bcc $84f7
	lda $d4
	sbc #$01
	sta $d4
	inc $0399
	lda $0399
	cmp #$08
	bcc $8544
	jsr $be37
	jsr $be94
	ldy #$00
	jsr $b0d9
	iny
	cpy $0398
	bne $8506
	lda $03d1
	and #$0c
	beq $8526
	dey
	ldx $039a,y
	jsr $95cd
	dey
	bpl $8517
	sta $0398
	sta $0399
	lda $0399
	cmp #$20
	bcc $8544
	ldx #$06
	lda #$01
	ldy #$1b
	jsr $b08e
	ldy $02
	cpy #$d0
	bcs $8544
	lda ($06),y
	bne $8544
	lda #$23
	sta ($06),y
	lda $8c
	sec
	sbc $071c
	tay
	lda $73
	sbc $071a
	bmi $8556
	cpy #$09
	bcs $8582
	lda #$00
	sta $14
	lda $73
	and #$01
	tay
	lda $7a1e,y
	sta $06
	lda $7a20,y
	sta $07
	lda $8c
	lsr
	lsr
	lsr
	lsr
	tay
	lda ($06),y
	cmp #$23
	bne $857a
	lda #$00
	sta ($06),y
	tya
	clc
	adc #$10
	cmp #$d0
	bcc $856f
	ldx $08
	rts

	.byte $0f
	.byte $07
	lda $074e
	beq $85fb
	ldx #$02
	stx $08
	lda $0f,x
	bne $85e5
	lda $07a8,x
	ldy $06cc
	and $8585,y
	cmp #$06
	bcs $85e5
	tay
	lda $046b,y
	beq $85e5
	lda $047d,y
	beq $85b4
	sbc #$00
	sta $047d,y
	jmp $85e5

	lda $0747
	bne $85e5
	lda #$0e
	sta $047d,y
	lda $046b,y
	sta $6e,x
	lda $0471,y
	sta $87,x
	lda $0477,y
	sec
	sbc #$08
	sta $cf,x
	lda #$01
	sta $b6,x
	sta $0f,x
	lsr
	sta $1e,x
	lda #$09
	sta $049a,x
	lda #$33
	sta $16,x
	jmp $85f8

	lda $16,x
	cmp #$33
	bne $85f8
	jsr $a2b4
	lda $0f,x
	beq $85f8
	jsr $be94
	jsr $85fe
	dex
	bpl $858e
	rts

	.byte $18
	.byte $e8
	lda $0747
	bne $8641
	lda $1e,x
	bne $8635
	lda $03d1
	and #$0c
	cmp #$0c
	beq $8650
	ldy #$01
	jsr $add9
	bmi $8618
	iny
	sty $46,x
	dey
	lda $85fc,y
	sta $58,x
	lda $00
	adc #$28
	cmp #$50
	bcc $8650
	lda #$01
	sta $1e,x
	lda #$0a
	sta $078a,x
	lda #$08
	sta $fe
	lda $1e,x
	and #$20
	beq $863e
	jsr $8b34
	jsr $8ad3
	jsr $be94
	jsr $be37
	jsr $aee1
	jsr $a4ab
	jmp $b52c

	jsr $95cd
	rts

	.byte $04
	.byte $04
	.byte $04
	.byte $05
	.byte $05
	.byte $05
	.byte $06
	.byte $06
	.byte $06
	.byte $10
	.byte $f0
	lda $07a8
	and #$07
	bne $866b
	lda $07a8
	and #$08
	tay
	lda $002a,y
	bne $868a
	ldx $8654,y
	lda $0f,x
	bne $868a
	ldx $08
	txa
	sta $06ae,y
	lda #$90
	sta $002a,y
	lda #$07
	sta $04a2,y
	sec
	rts

	ldx $08
	clc
	rts

	lda $0747
	bne $86f6
	lda $2a,x
	and #$7f
	ldy $06ae,x
	cmp #$02
	beq $86be
	bcs $86d4
	txa
	clc
	adc #$0d
	tax
	lda #$10
	sta $00
	lda #$0f
	sta $01
	lda #$04
	sta $02
	lda #$00
	jsr $8ba8
	jsr $8ae0
	ldx $08
	jmp $86f3

	lda #$fe
	sta $ac,x
	lda $001e,y
	and #$f7
	sta $001e,y
	ldx $46,y
	dex
	lda $865d,x
	ldx $08
	sta $64,x
	dec $2a,x
	lda $0087,y
	clc
	adc #$02
	sta $93,x
	lda $006e,y
	adc #$00
	sta $7a,x
	lda $00cf,y
	sec
	sbc #$0a
	sta $db,x
	lda #$01
	sta $c2,x
	bne $86f6
	jsr $a410
	jsr $be80
	jsr $be2d
	jsr $aed4
	jsr $b180
	rts

	jsr $874f
	lda $76,x
	sta $007a,y
	lda $8f,x
	ora #$05
	sta $0093,y
	lda $d7,x
	sbc #$10
	sta $00db,y
	jmp $8737

	jsr $874f
	lda $03ea,x
	sta $007a,y
	lda $06
	asl
	asl
	asl
	asl
	ora #$05
	sta $0093,y
	lda $02
	adc #$20
	sta $00db,y
	lda #$fb
	sta $00ac,y
	lda #$01
	sta $00c2,y
	sta $002a,y
	sta $fe
	stx $08
	jsr $87c3
	inc $0748
	rts

	ldy #$08
	lda $002a,y
	beq $875d
	dey
	cpy #$05
	bne $8751
	ldy #$08
	sty $06b7
	rts

	ldx #$08
	stx $08
	lda $2a,x
	beq $87bf
	asl
	bcc $8772
	jsr $868e
	jmp $87bf

	ldy $2a,x
	dey
	beq $8794
	inc $2a,x
	lda $93,x
	clc
	adc $0775
	sta $93,x
	lda $7a,x
	adc #$00
	sta $7a,x
	lda $2a,x
	cmp #$30
	bne $87b3
	lda #$00
	sta $2a,x
	jmp $87bf

	txa
	clc
	adc #$0d
	tax
	lda #$50
	sta $00
	lda #$06
	sta $02
	lsr
	sta $01
	lda #$00
	jsr $8ba8
	ldx $08
	lda $ac,x
	cmp #$05
	bne $87b3
	inc $2a,x
	jsr $be2d
	jsr $be80
	jsr $aed4
	jsr $b32c
	dex
	bpl $8763
	rts

	lda #$01
	sta $0139
	ldy #$11
	jsr $6dd2
	inc $075e
	lda $075e
	cmp #$64
	bne $87e3
	lda #$00
	sta $075e
	inc $075a
	lda #$40
	sta $fe
	lda #$02
	sta $0138
	ldy #$0b
	jsr $6dd2
	lda #$01
	jsr $6d79
	ldy $0300
	lda $02fb,y
	bne $87ff
	lda #$24
	sta $02fb,y
	ldx $08
	rts

	lda #$2e
	sta $1b
	lda $76,x
	sta $73
	lda $8f,x
	sta $8c
	lda #$01
	sta $bb
	lda $d7,x
	sec
	sbc #$08
	sta $d4
	lda #$01
	sta $23
	sta $14
	lda #$03
	sta $049f
	lda $39
	cmp #$02
	bcs $8834
	lda $0756
	cmp #$02
	bcc $8832
	.byte $4a
	sta $39
	lda #$20
	sta $03ca
	lda #$02
	sta $fe
	rts

	ldx #$05
	stx $08
	lda $23
	beq $88ab
	asl
	bcc $8874
	lda $0747
	bne $8899
	lda $39
	beq $886b
	cmp #$03
	beq $886b
	cmp #$04
	beq $886b
	cmp #$05
	beq $886b
	cmp #$02
	bne $8899
	jsr $972e
	jsr $adf9
	jmp $8899

	jsr $96ac
	jsr $ac4a
	jmp $8899

	lda $09
	and #$03
	bne $8893
	dec $d4
	lda $23
	inc $23
	cmp #$11
	bcc $8893
	lda #$10
	sta $58,x
	lda #$80
	sta $23
	asl
	sta $03ca
	rol
	sta $46,x
	lda $23
	cmp #$06
	bcc $88ab
	jsr $be37
	jsr $be94
	jsr $aee1
	jsr $b37d
	jsr $a4ab
	jsr $a2b4
	rts

	.byte $04
	.byte $12
	pha
	lda #$11
	ldx $03ee
	ldy $0754
	bne $88bb
	lda #$12
	sta $26,x
	jsr $6948
	ldx $03ee
	lda $02
	sta $03e4,x
	tay
	lda $06
	sta $03e6,x
	lda ($06),y
	jsr $89c7
	sta $00
	ldy $0754
	bne $88db
	tya
	bcc $8902
	ldy #$11
	sty $26,x
	lda #$c5
	ldy $00
	cpy #$56
	beq $88ed
	cpy #$5c
	bne $8902
	lda $06bc
	bne $88fa
	lda #$0b
	sta $079d
	inc $06bc
	lda $079d
	bne $8901
	ldy #$c5
	tya
	sta $03e8,x
	jsr $8945
	ldy $02
	lda #$20
	sta ($06),y
	lda #$10
	sta $0784
	pla
	sta $05
	ldy #$00
	lda $0714
	bne $8922
	lda $0754
	beq $8923
	iny
	lda $ce
	clc
	adc $88ac,y
	and #$f0
	sta $d7,x
	ldy $26,x
	cpy #$11
	beq $8939
	jsr $89d3
	jmp $893c

	jsr $895c
	lda $03ee
	eor #$01
	sta $03ee
	rts

	lda $86
	clc
	adc #$08
	and #$f0
	sta $8f,x
	lda $6d
	adc #$00
	sta $76,x
	sta $03ea,x
	lda $b5
	sta $be,x
	rts

	jsr $89f0
	lda #$02
	sta $ff
	lda #$00
	sta $60,x
	sta $043c,x
	sta $9f
	lda #$fe
	sta $a8,x
	lda $05
	jsr $89c7
	bcc $89b3
	tya
	cmp #$0d
	bcc $897e
	sbc #$06
	jsr $6c7d
	.byte $9b
	.byte $89
	.byte $a1
	.byte $89
	.byte $03
	.byte $87
	.byte $03
	.byte $87
	.byte $a4
	.byte $89
	.byte $a1
	.byte $89
	.byte $9b
	.byte $89
	.byte $9b
	.byte $89
	.byte $a1
	.byte $89
	.byte $ab
	.byte $89
	.byte $9e
	.byte $89
	.byte $03
	.byte $87
	.byte $a4
	.byte $89
	lda #$00
	bit $02a9
	bit $04a9
	bit $03a9
	sta $39
	jmp $8802

	ldx #$05
	ldy $03ee
	jsr $84aa
	rts

	.byte $c1
	.byte $c2
	.byte $c0
	.byte $5e
	.byte $5f
	.byte $60
	.byte $61
	.byte $52
	.byte $53
	.byte $54
	.byte $55
	.byte $56
	.byte $57
	.byte $58
	.byte $59
	.byte $5a
	.byte $5b
	.byte $5c
	.byte $5d
	ldy #$12
	cmp $89b4,y
	beq $89d2
	dey
	bpl $89c9
	clc
	rts

	jsr $89f0
	lda #$01
	sta $03ec,x
	sta $fd
	jsr $8a12
	lda #$fe
	sta $9f
	lda #$05
	sta $0139
	jsr $87e8
	ldx $03ee
	rts

	ldx $03ee
	ldy $02
	beq $8a11
	tya
	sec
	sbc #$10
	sta $02
	tay
	lda ($06),y
	cmp #$c3
	bne $8a11
	lda #$00
	sta ($06),y
	jsr $692a
	ldx $03ee
	jsr $871c
	rts

	lda $8f,x
	sta $03f1,x
	lda #$f0
	sta $60,x
	sta $62,x
	lda #$fa
	sta $a8,x
	lda #$fc
	sta $aa,x
	lda #$00
	sta $043c,x
	sta $043e,x
	lda $76,x
	sta $78,x
	lda $8f,x
	sta $91,x
	lda $d7,x
	clc
	adc #$08
	sta $d9,x
	lda #$fa
	sta $a8,x
	rts

	lda $26,x
	beq $8aa2
	and #$0f
	pha
	tay
	txa
	clc
	adc #$09
	tax
	dey
	beq $8a84
	jsr $8b75
	jsr $8ae0
	txa
	clc
	adc #$02
	tax
	jsr $8b75
	jsr $8ae0
	ldx $08
	jsr $be3e
	jsr $be9b
	jsr $b92e
	pla
	ldy $be,x
	beq $8aa2
	pha
	lda #$f0
	cmp $d9,x
	bcs $8a7b
	sta $d9,x
	lda $d7,x
	cmp #$f0
	pla
	bcc $8aa2
	bcs $8aa0
	jsr $8b75
	ldx $08
	jsr $be3e
	jsr $be9b
	jsr $b8ac
	lda $d7,x
	and #$0f
	cmp #$05
	pla
	bcs $8aa2
	lda #$01
	sta $03ec,x
	lda #$00
	sta $26,x
	rts

	ldx #$01
	stx $08
	lda $0301
	bne $8acf
	lda $03ec,x
	beq $8acf
	lda $03e6,x
	sta $06
	lda #$05
	sta $07
	lda $03e4,x
	sta $02
	tay
	lda $03e8,x
	sta ($06),y
	jsr $693e
	lda #$00
	sta $03ec,x
	dex
	bpl $8aa7
	rts

	inx
	jsr $8ae0
	ldx $08
	rts

	lda $070e
	bne $8b1d
	tax
	lda $57,x
	asl
	asl
	asl
	asl
	sta $01
	lda $57,x
	lsr
	lsr
	lsr
	lsr
	cmp #$08
	bcc $8af4
	ora #$f0
	sta $00
	ldy #$00
	cmp #$00
	bpl $8afd
	dey
	sty $02
	lda $0400,x
	clc
	adc $01
	sta $0400,x
	lda #$00
	rol
	pha
	ror
	lda $86,x
	adc $00
	sta $86,x
	lda $6d,x
	adc $02
	sta $6d,x
	pla
	clc
	adc $00
	rts

	ldx #$00
	lda $0747
	bne $8b2a
	lda $070e
	bne $8b1d
	lda $0709
	sta $00
	lda #$04
	jmp $8b7e

	ldy #$3d
	lda $1e,x
	cmp #$05
	bne $8b3e
	ldy #$20
	jmp $8b65

	ldy #$00
	jmp $8b48

	ldy #$01
	inx
	lda #$03
	sta $00
	lda #$06
	sta $01
	lda #$02
	sta $02
	tya
	jmp $8ba2

	ldy #$7f
	bne $8b5f
	ldy #$0f
	lda #$02
	bne $8b67
	ldy #$1c
	lda #$03
	sty $00
	inx
	jsr $8b7e
	ldx $08
	rts

	.byte $06
	.byte $08
	.byte $a0
	.byte $00
	.byte $2c
	ldy #$01
	lda #$50
	sta $00
	lda $8b70,y
	sta $02
	lda #$00
	jmp $8ba8

	lda #$00
	bit $01a9
	pha
	ldy $16,x
	inx
	lda #$05
	cpy #$29
	bne $8b96
	lda #$09
	sta $00
	lda #$0a
	sta $01
	lda #$03
	sta $02
	pla
	tay
	jsr $8ba8
	ldx $08
	rts

	pha
	lda $0416,x
	clc
	adc $0433,x
	sta $0416,x
	ldy #$00
	lda $9f,x
	bpl $8bba
	dey
	sty $07
	adc $ce,x
	sta $ce,x
	lda $b5,x
	adc $07
	sta $b5,x
	lda $0433,x
	clc
	adc $00
	sta $0433,x
	lda $9f,x
	adc #$00
	sta $9f,x
	cmp $02
	bmi $8be9
	lda $0433,x
	cmp #$80
	bcc $8be9
	lda $02
	sta $9f,x
	lda #$00
	sta $0433,x
	pla
	beq $8c17
	lda $02
	eor #$ff
	tay
	iny
	sty $07
	lda $0433,x
	sec
	sbc $01
	sta $0433,x
	lda $9f,x
	sbc #$00
	sta $9f,x
	cmp $07
	bpl $8c17
	lda $0433,x
	cmp #$80
	bcs $8c17
	lda $07
	sta $9f,x
	lda #$ff
	sta $0433,x
	rts

	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	lda $0f,x
	pha
	asl
	bcs $8c3b
	pla
	beq $8c2f
	jmp $94b7

	lda $071f
	and #$07
	cmp #$07
	beq $8c46
	jmp $8cb7

	pla
	and #$0f
	tay
	lda $000f,y
	bne $8c46
	sta $0f,x
	rts

	.byte $02
	.byte $02
	.byte $02
	.byte $02
	.byte $05
	.byte $05
	.byte $05
	.byte $05
	.byte $06
	.byte $07
	.byte $07
	.byte $04
	.byte $03
	.byte $05
	.byte $08
	.byte $09
	.byte $03
	.byte $06
	.byte $07
	.byte $0a
	.byte $05
	.byte $05
	.byte $0b
	.byte $05
	.byte $b0
	.byte $b0
	.byte $40
	.byte $30
	.byte $b0
	.byte $30
	.byte $b0
	.byte $b0
	.byte $f0
	.byte $f0
	.byte $b0
	.byte $f0
	.byte $02
	.byte $02
	.byte $02
	.byte $02
	.byte $02
	.byte $02
	.byte $02
	.byte $02
	.byte $01
	.byte $01
	.byte $01
	.byte $01
	lda $6d
	sec
	sbc #$04
	sta $6d
	lda $0725
	sec
	sbc #$04
	sta $0725
	lda $071a
	sec
	sbc #$04
	sta $071a
	lda $071b
	sec
	sbc #$04
	sta $071b
	lda $072a
	sec
	sbc #$04
	sta $072a
	lda #$00
	sta $073b
	sta $072b
	sta $0739
	sta $073a
	lda $c388,y
	sta $072c
	rts

	lda $0745
	beq $8d0c
	lda $0726
	bne $8d0c
	ldy #$0c
	dey
	bmi $8d0c
	lda $075f
	cmp $8c47,y
	bne $8cc3
	lda $0725
	cmp $8c53,y
	bne $8cc3
	lda $ce
	cmp $8c5f,y
	bne $8ce6
	lda $1d
	cmp #$00
	bne $8ce6
	inc $06d9
	inc $06da
	lda $06da
	cmp $8c6b,y
	bne $8d07
	lda $06d9
	cmp $8c6b,y
	beq $8cff
	jsr $8c77
	jsr $9ca6
	lda #$00
	sta $06da
	sta $06d9
	lda #$00
	sta $0745
	lda $06cd
	beq $8d21
	sta $16,x
	lda #$01
	sta $0f,x
	lda #$00
	sta $1e,x
	sta $06cd
	jmp $8e03

	ldy $0739
	lda ($e9),y
	cmp #$ff
	bne $8d2d
	jmp $8df3

	and #$0f
	cmp #$0e
	beq $8d41
	cpx #$05
	bcc $8d41
	iny
	lda ($e9),y
	and #$3f
	cmp #$2e
	beq $8d41
	rts

	lda $071d
	clc
	adc #$30
	and #$f0
	sta $07
	lda $071b
	adc #$00
	sta $06
	ldy $0739
	iny
	lda ($e9),y
	asl
	bcc $8d66
	lda $073b
	bne $8d66
	inc $073b
	inc $073a
	dey
	lda ($e9),y
	and #$0f
	cmp #$0f
	bne $8d88
	lda $073b
	bne $8d88
	iny
	lda ($e9),y
	and #$3f
	sta $073a
	inc $0739
	inc $0739
	inc $073b
	jmp $8cb7

	lda $073a
	sta $6e,x
	lda ($e9),y
	and #$f0
	sta $87,x
	cmp $071d
	lda $6e,x
	sbc $071b
	bcs $8da8
	lda ($e9),y
	and #$0f
	cmp #$0e
	beq $8e0e
	jmp $8e34

	lda $07
	cmp $87,x
	lda $06
	sbc $6e,x
	bcc $8df3
	lda #$01
	sta $b6,x
	lda ($e9),y
	asl
	asl
	asl
	asl
	sta $cf,x
	cmp #$e0
	beq $8e0e
	iny
	lda ($e9),y
	and #$40
	beq $8dce
	lda $06cc
	beq $8e42
	lda ($e9),y
	and #$3f
	cmp #$37
	bcc $8dda
	cmp #$3f
	bcc $8e0b
	cmp #$06
	bne $8de5
	ldy $076a
	beq $8de5
	lda #$02
	sta $16,x
	lda #$01
	sta $0f,x
	jsr $8e03
	lda $0f,x
	bne $8e42
	rts

	lda $06cb
	bne $8e01
	lda $0398
	cmp #$01
	bne $8e0a
	lda #$2f
	sta $16,x
	lda #$00
	sta $1e,x
	jsr $8e50
	rts

	jmp $932c

	iny
	iny
	lda $075f
	cmp #$08
	beq $8e23
	lda ($e9),y
	lsr
	lsr
	lsr
	lsr
	lsr
	cmp $075f
	bne $8e31
	dey
	lda ($e9),y
	sta $0750
	iny
	lda ($e9),y
	and #$1f
	sta $0751
	jmp $8e3f

	ldy $0739
	lda ($e9),y
	and #$0f
	cmp #$0e
	bne $8e42
	inc $0739
	inc $0739
	inc $0739
	lda #$00
	sta $073b
	ldx $08
	rts

	lda $16,x
	cmp #$15
	bcs $8e63
	tay
	lda $cf,x
	adc #$08
	sta $cf,x
	lda #$01
	sta $03d8,x
	tya
	jsr $6c7d
	.byte $f2
	.byte $8e
	.byte $f2
	.byte $8e
	.byte $f2
	.byte $8e
	.byte $02
	.byte $8f
	.byte $98
	.byte $93
	.byte $0c
	.byte $8f
	.byte $d5
	.byte $8e
	.byte $2d
	.byte $8f
	.byte $56
	.byte $8f
	.byte $d4
	.byte $8e
	.byte $60
	.byte $8f
	.byte $60
	.byte $8f
	.byte $db
	.byte $8e
	.byte $98
	.byte $93
	.byte $06
	.byte $94
	.byte $35
	.byte $8f
	.byte $28
	.byte $8f
	.byte $70
	.byte $8f
	.byte $d5
	.byte $93
	.byte $d4
	.byte $8e
	.byte $d5
	.byte $93
	.byte $d5
	.byte $93
	.byte $d5
	.byte $93
	.byte $d5
	.byte $93
	.byte $ed
	.byte $93
	.byte $d4
	.byte $8e
	.byte $d4
	.byte $8e
	.byte $55
	.byte $90
	.byte $55
	.byte $90
	.byte $55
	.byte $90
	.byte $55
	.byte $90
	.byte $52
	.byte $90
	.byte $d4
	.byte $8e
	.byte $d4
	.byte $8e
	.byte $d4
	.byte $8e
	.byte $d4
	.byte $8e
	.byte $14
	.byte $94
	.byte $47
	.byte $94
	.byte $74
	.byte $94
	.byte $7a
	.byte $94
	.byte $40
	.byte $94
	.byte $38
	.byte $94
	.byte $40
	.byte $94
	.byte $80
	.byte $94
	.byte $8c
	.byte $94
	.byte $42
	.byte $91
	.byte $19
	.byte $88
	.byte $aa
	.byte $84
	.byte $d4
	.byte $8e
	.byte $d4
	.byte $8e
	.byte $d4
	.byte $8e
	.byte $d4
	.byte $8e
	.byte $d4
	.byte $8e
	.byte $eb
	.byte $8e
	.byte $b6
	.byte $94
	rts

	jsr $8ef2
	jmp $8f31

	lda #$02
	sta $b6,x
	sta $cf,x
	lsr
	sta $0796,x
	lsr
	sta $1e,x
	jmp $8f31

	lda #$b8
	sta $cf,x
	rts

	.byte $f8
	.byte $f4
	ldy #$01
	lda $076a
	bne $8efa
	dey
	lda $8ef0,y
	sta $58,x
	jmp $8f45

	jsr $8ef2
	lda #$01
	sta $1e,x
	rts

	.byte $80
	.byte $50
	lda #$00
	sta $03a2,x
	sta $58,x
	lda $075f
	cmp #$06
	bcs $8f23
	ldy $06cc
	lda $8f0a,y
	sta $0796,x
	lda #$0b
	jmp $8f47

	lda #$00
	jmp $8efd

	lda #$00
	sta $58,x
	lda #$09
	bne $8f47
	ldy #$30
	lda $cf,x
	sta $0401,x
	bpl $8f40
	ldy #$e0
	tya
	adc $cf,x
	sta $58,x
	lda #$03
	sta $049a,x
	lda #$02
	sta $46,x
	lda #$00
	sta $a0,x
	sta $0434,x
	rts

	lda #$02
	sta $46,x
	lda #$09
	sta $049a,x
	rts

	jsr $8f31
	lda $07a7,x
	and #$10
	sta $58,x
	lda $cf,x
	sta $0434,x
	rts

	lda $06cb
	bne $8f80
	lda #$00
	sta $06d1
	jsr $8f28
	jmp $940e

	jmp $95cd

	.byte $26
	.byte $2c
	.byte $32
	.byte $38
	.byte $20
	.byte $22
	.byte $24
	.byte $26
	.byte $13
	.byte $14
	.byte $15
	.byte $16
	lda $078f
	bne $8fde
	cpx #$05
	bcs $8fde
	lda #$80
	sta $078f
	ldy #$04
	lda $0016,y
	cmp #$11
	beq $8fdf
	dey
	bpl $8f9f
	inc $06d1
	lda $06d1
	cmp #$03
	bcc $8fde
	ldx #$04
	lda $0f,x
	beq $8fbe
	dex
	bpl $8fb5
	bmi $8fdc
	lda #$00
	sta $1e,x
	lda #$11
	sta $16,x
	jsr $8f75
	lda #$20
	ldy $07fb
	bne $8fd7
	ldy $075f
	cpy #$06
	bcc $8fd9
	lda #$60
	jsr $91e9
	ldx $08
	rts

	lda $ce
	cmp #$2c
	bcc $8fde
	lda $001e,y
	bne $8fde
	lda $006e,y
	sta $6e,x
	lda $0087,y
	sta $87,x
	lda #$01
	sta $b6,x
	lda $00cf,y
	sec
	sbc #$08
	sta $cf,x
	lda $07a7,x
	and #$03
	tay
	ldx #$02
	lda $8f83,y
	sta $01,x
	iny
	iny
	iny
	iny
	dex
	bpl $9008
	ldx $08
	jsr $9ba1
	ldy $57
	cpy #$08
	bcs $902d
	tay
	lda $07a8,x
	and #$03
	beq $902c
	tya
	eor #$ff
	tay
	iny
	tya
	jsr $8f31
	ldy #$02
	sta $58,x
	cmp #$00
	bmi $9039
	dey
	sty $46,x
	lda #$fd
	sta $a0,x
	lda #$01
	sta $0f,x
	lda #$05
	sta $1e,x
	rts

	.byte $28
	.byte $38
	.byte $28
	.byte $38
	.byte $28
	.byte $00
	.byte $00
	.byte $10
	.byte $10
	.byte $00
	jsr $9186
	lda #$00
	sta $58,x
	lda $16,x
	sec
	sbc #$1b
	tay
	lda $9048,y
	sta $0388,x
	lda $904d,y
	sta $34,x
	lda $cf,x
	clc
	adc #$04
	sta $cf,x
	lda $87,x
	clc
	adc #$04
	sta $87,x
	lda $6e,x
	adc #$00
	sta $6e,x
	jmp $940e

	.byte $80
	.byte $30
	.byte $40
	.byte $80
	.byte $30
	.byte $50
	.byte $50
	.byte $70
	.byte $20
	.byte $40
	.byte $80
	.byte $a0
	.byte $70
	.byte $40
	.byte $90
	.byte $68
	.byte $0e
	.byte $05
	.byte $06
	.byte $0e
	.byte $1c
	.byte $20
	.byte $10
	.byte $0c
	.byte $1e
	.byte $22
	.byte $18
	.byte $14
	.byte $10
	.byte $60
	.byte $20
	.byte $48
	lda $078f
	bne $9047
	jsr $8f31
	lda $07a8,x
	and #$03
	tay
	lda $909d,y
	sta $078f
	ldy #$03
	lda $06cc
	beq $90bd
	.byte $c8
	sty $00
	cpx $00
	bcs $9047
	lda $07a7,x
	and #$03
	sta $00
	sta $01
	lda #$fb
	sta $a0,x
	lda #$00
	ldy $57
	beq $90dd
	lda #$04
	cpy #$19
	bcc $90dd
	asl
	pha
	clc
	adc $00
	sta $00
	lda $07a8,x
	and #$03
	beq $90f1
	lda $07a9,x
	and #$0f
	sta $00
	pla
	clc
	adc $01
	tay
	lda $9091,y
	sta $58,x
	lda #$01
	sta $46,x
	lda $57
	bne $9115
	ldy $00
	tya
	and #$02
	beq $9115
	lda $58,x
	eor #$ff
	clc
	adc #$01
	sta $58,x
	inc $46,x
	tya
	and #$02
	beq $9129
	lda $86
	clc
	adc $9081,y
	sta $87,x
	lda $6d
	adc #$00
	jmp $9135

	lda $86
	sec
	sbc $9081,y
	sta $87,x
	lda $6d
	sbc #$00
	sta $6e,x
	lda #$01
	sta $0f,x
	sta $b6,x
	lda #$f8
	sta $cf,x
	rts

	ldy #$04
	cpy $08
	beq $9157
	lda $0016,y
	cmp #$2d
	bne $9157
	lda #$00
	sta $0016,y
	sta $000f,y
	dey
	bpl $9144
	jsr $9186
	stx $0368
	lda #$00
	sta $0363
	sta $0369
	lda $87,x
	sta $0366
	lda #$df
	sta $0790
	sta $46,x
	lda #$20
	sta $0364
	sta $078a,x
	lda #$05
	sta $0483
	lsr
	sta $0365
	rts

	ldy #$ff
	iny
	lda $000f,y
	bne $9188
	sty $06cf
	txa
	ora #$80
	sta $000f,y
	lda $6e,x
	sta $006e,y
	lda $87,x
	sta $0087,y
	lda #$01
	sta $0f,x
	sta $00b6,y
	lda $cf,x
	sta $00cf,y
	rts

	.byte $90
	.byte $80
	.byte $70
	.byte $90
	.byte $ff
	.byte $01
	lda $078f
	bne $91ad
	sta $0434,x
	lda $fd
	ora #$02
	sta $fd
	ldy $0368
	lda $0016,y
	cmp #$2d
	beq $91fd
	jsr $9e0e
	clc
	adc #$20
	ldy $06cc
	beq $91da
	sec
	sbc #$10
	sta $078f
	lda $07a7,x
	and #$03
	sta $0417,x
	tay
	lda $91ae,y
	sta $cf,x
	lda $071d
	clc
	adc #$20
	sta $87,x
	lda $071b
	adc #$00
	sta $6e,x
	jmp $9230

	lda $0087,y
	sec
	sbc #$0e
	sta $87,x
	lda $006e,y
	sta $6e,x
	lda $00cf,y
	clc
	adc #$08
	sta $cf,x
	lda $07a7,x
	and #$03
	sta $0417,x
	tay
	lda $91ae,y
	ldy #$00
	cmp $cf,x
	bcc $9225
	iny
	lda $91b2,y
	sta $0434,x
	lda #$00
	sta $06cb
	lda #$08
	sta $049a,x
	lda #$01
	sta $b6,x
	sta $0f,x
	lsr
	sta $0401,x
	sta $1e,x
	rts

	.byte $00
	.byte $30
	.byte $60
	.byte $60
	.byte $00
	.byte $20
	.byte $60
	.byte $40
	.byte $70
	.byte $40
	.byte $60
	.byte $30
	lda $078f
	bne $929a
	lda #$20
	sta $078f
	dec $06d7
	ldy #$06
	dey
	lda $0016,y
	cmp #$31
	bne $925d
	lda $0087,y
	sec
	sbc #$30
	pha
	lda $006e,y
	sbc #$00
	sta $00
	lda $06d7
	clc
	adc $001e,y
	tay
	pla
	clc
	adc $9242,y
	sta $87,x
	lda $00
	adc #$00
	sta $6e,x
	lda $9248,y
	sta $cf,x
	lda #$01
	sta $b6,x
	sta $0f,x
	lsr
	sta $58,x
	lda #$08
	sta $a0,x
	rts

	.byte $01
	.byte $02
	.byte $04
	.byte $08
	.byte $10
	.byte $20
	.byte $40
	.byte $80
	.byte $40
	.byte $30
	.byte $90
	.byte $50
	.byte $20
	.byte $60
	.byte $a0
	.byte $70
	.byte $0a
	.byte $0b
	lda $078f
	bne $9321
	lda $074e
	bne $930e
	cpx #$03
	bcs $9321
	ldy #$00
	lda $07a7,x
	cmp #$aa
	bcc $92c5
	iny
	lda $075f
	cmp #$01
	beq $92cd
	iny
	tya
	and #$01
	tay
	lda $92ab,y
	sta $16,x
	lda $06dd
	cmp #$ff
	bne $92e2
	lda #$00
	sta $06dd
	lda $07a7,x
	and #$07
	tay
	lda $929b,y
	bit $06dd
	beq $92f7
	iny
	tya
	and #$07
	jmp $92e7

	ora $06dd
	sta $06dd
	lda $92a3,y
	jsr $91e9
	sta $0417,x
	lda #$20
	sta $078f
	jmp $8e50

	ldy #$ff
	iny
	cpy #$05
	bcs $9322
	lda $000f,y
	beq $9310
	lda $0016,y
	cmp #$08
	bne $9310
	rts

	lda $fe
	ora #$08
	sta $fe
	lda #$08
	bne $92d4
	ldy #$00
	sec
	sbc #$37
	pha
	cmp #$04
	bcs $9341
	pha
	ldy #$06
	lda $076a
	beq $9340
	ldy #$02
	pla
	sty $01
	ldy #$b0
	and #$02
	beq $934b
	ldy #$70
	sty $00
	lda $071b
	sta $02
	lda $071d
	sta $03
	ldy #$02
	pla
	lsr
	bcc $935e
	iny
	sty $06d3
	ldx #$ff
	inx
	cpx #$05
	bcs $9395
	lda $0f,x
	bne $9363
	lda $01
	sta $16,x
	lda $02
	sta $6e,x
	lda $03
	sta $87,x
	clc
	adc #$18
	sta $03
	lda $02
	adc #$00
	sta $02
	lda $00
	sta $cf,x
	lda #$01
	sta $b6,x
	sta $0f,x
	jsr $8e50
	dec $06d3
	bne $9361
	jmp $8e42

	lda #$22
	sta $00
	lda #$13
	sta $01
	lda $07fb
	bne $93b2
	lda $075f
	cmp #$03
	bcs $93b2
	dec $00
	lda #$21
	sta $01
	lda $00
	sta $b517
	lda $01
	sta $9ffe
	lda #$01
	sta $58,x
	lsr
	sta $1e,x
	sta $a0,x
	lda $cf,x
	sta $0434,x
	sec
	sbc #$18
	sta $0417,x
	lda #$09
	jmp $9410

	lda $16,x
	sta $06cb
	sec
	sbc #$12
	jsr $6c7d
	.byte $8f
	.byte $8f
	.byte $ec
	.byte $93
	.byte $a1
	.byte $90
	.byte $b4
	.byte $91
	.byte $4e
	.byte $92
	.byte $ad
	.byte $92
	rts

	ldy #$05
	lda $0016,y
	cmp #$11
	bne $93fb
	lda #$01
	sta $001e,y
	dey
	bpl $93ef
	lda #$00
	sta $06cb
	sta $0f,x
	rts

	lda #$02
	sta $46,x
	lda #$f4
	sta $58,x
	lda #$03
	sta $049a,x
	rts

	dec $cf,x
	dec $cf,x
	ldy $06cc
	bne $9422
	ldy #$02
	jsr $94a6
	ldy #$ff
	lda $03a0
	sta $1e,x
	bpl $942d
	txa
	tay
	sty $03a0
	lda #$00
	sta $46,x
	tay
	jsr $94a6
	lda #$ff
	sta $03a2,x
	jmp $945d

	lda #$00
	sta $58,x
	jmp $945d

	ldy #$40
	lda $cf,x
	bpl $9454
	eor #$ff
	clc
	adc #$01
	ldy #$c0
	sta $0401,x
	tya
	clc
	adc $cf,x
	sta $58,x
	jsr $8f4e
	lda #$05
	ldy $074e
	cpy #$03
	beq $9470
	ldy $06cc
	bne $9470
	lda #$06
	sta $049a,x
	rts

	jsr $9480
	jmp $947d

	jsr $948c
	jmp $9460

	lda #$10
	sta $0434,x
	lda #$ff
	sta $a0,x
	jmp $9495

	lda #$f0
	sta $0434,x
	lda #$00
	sta $a0,x
	ldy #$01
	jsr $94a6
	lda #$04
	sta $049a,x
	rts

	.byte $08
	.byte $0c
	.byte $f8
	.byte $00
	.byte $00
	.byte $ff
	lda $87,x
	clc
	adc $94a0,y
	sta $87,x
	lda $6e,x
	adc $94a3,y
	sta $6e,x
	rts

	rts

	ldx $08
	lda #$00
	ldy $16,x
	cpy #$15
	bcc $94c4
	tya
	sbc #$14
	jsr $6c7d
	.byte $15
	.byte $95
	.byte $6a
	.byte $95
	.byte $ca
	.byte $9e
	.byte $0b
	.byte $95
	.byte $0b
	.byte $95
	.byte $0b
	.byte $95
	.byte $0b
	.byte $95
	.byte $7c
	.byte $95
	.byte $7c
	.byte $95
	.byte $7c
	.byte $95
	.byte $7c
	.byte $95
	.byte $7c
	.byte $95
	.byte $7c
	.byte $95
	.byte $7c
	.byte $95
	.byte $7c
	.byte $95
	.byte $0b
	.byte $95
	.byte $9a
	.byte $95
	.byte $9a
	.byte $95
	.byte $9a
	.byte $95
	.byte $9a
	.byte $95
	.byte $9a
	.byte $95
	.byte $9a
	.byte $95
	.byte $9a
	.byte $95
	.byte $82
	.byte $95
	.byte $82
	.byte $95
	.byte $9a
	.byte $9c
	.byte $3e
	.byte $88
	.byte $d7
	.byte $84
	.byte $0b
	.byte $95
	.byte $0e
	.byte $9f
	.byte $31
	.byte $84
	.byte $0b
	.byte $95
	.byte $10
	.byte $83
	.byte $0c
	.byte $95
	rts

	jsr $be94
	jsr $be37
	jmp $b52c

	lda #$00
	sta $03c5,x
	jsr $be94
	jsr $be37
	jsr $b52c
	jsr $aee1
	jsr $ac4a
	jsr $a69c
	jsr $a4ab
	ldy $0747
	bne $9537
	jsr $953a
	jmp $a2b4

	lda $16,x
	jsr $6c7d
	.byte $ac
	.byte $96
	.byte $ac
	.byte $96
	.byte $ac
	.byte $96
	.byte $ac
	.byte $96
	.byte $c0
	.byte $c4
	.byte $0d
	.byte $96
	.byte $ac
	.byte $96
	.byte $be
	.byte $97
	.byte $6b
	.byte $98
	.byte $69
	.byte $95
	.byte $7f
	.byte $98
	.byte $7f
	.byte $98
	.byte $e5
	.byte $95
	.byte $dc
	.byte $9f
	.byte $2e
	.byte $97
	.byte $34
	.byte $97
	.byte $5a
	.byte $97
	.byte $5d
	.byte $9b
	.byte $ac
	.byte $96
	.byte $69
	.byte $95
	.byte $14
	.byte $9b
	.byte $60
	jsr $9e20
	jsr $be94
	jsr $be37
	jsr $aee1
	jsr $a4ab
	jmp $a2b4

	jsr $9971
	jmp $a2b4

	jsr $be94
	jsr $be37
	jsr $aeea
	jsr $a7f0
	jsr $be37
	jsr $ba41
	jsr $a28f
	jmp $a2b4

	jsr $be94
	jsr $be37
	jsr $af11
	jsr $a7ba
	lda $0747
	bne $95ae
	jsr $95b7
	jsr $be37
	jsr $b26e
	jmp $a2b4

	lda $16,x
	sec
	sbc #$24
	jsr $6c7d
	.byte $65
	.byte $a0
	.byte $0d
	.byte $a2
	.byte $89
	.byte $a2
	.byte $89
	.byte $a2
	.byte $41
	.byte $a2
	.byte $6b
	.byte $a2
	.byte $77
	.byte $a2
	lda #$00
	sta $0f,x
	sta $16,x
	sta $1e,x
	sta $0110,x
	sta $0796,x
	sta $0125,x
	sta $03c5,x
	sta $078a,x
	rts

	lda $0796,x
	bne $9600
	jsr $8edb
	lda $07a8,x
	ora #$80
	sta $0434,x
	and #$0f
	ora #$06
	sta $0796,x
	lda #$f9
	sta $a0,x
	jmp $8b63

	.byte $30
	.byte $1c
	.byte $00
	.byte $e8
	.byte $00
	.byte $18
	.byte $08
	.byte $f8
	.byte $0c
	.byte $f4
	lda $1e,x
	and #$20
	beq $9616
	jmp $971a

	lda $3c,x
	beq $9647
	dec $3c,x
	lda $03d1
	and #$0c
	bne $968d
	lda $03a2,x
	bne $963f
	ldy $06cc
	lda $9603,y
	sta $03a2,x
	jsr $865f
	bcc $963f
	lda $1e,x
	ora #$08
	sta $1e,x
	jmp $968d

	dec $03a2,x
	jmp $968d

	.byte $20
	.byte $37
	lda $1e,x
	and #$07
	cmp #$01
	beq $968d
	lda #$00
	sta $00
	ldy #$fa
	lda $cf,x
	bmi $966c
	ldy #$fd
	cmp #$70
	inc $00
	bcc $966c
	dec $00
	lda $07a8,x
	and #$01
	bne $966c
	ldy #$fa
	sty $a0,x
	lda $1e,x
	ora #$01
	sta $1e,x
	lda $00
	and $07a9,x
	tay
	lda $06cc
	bne $9680
	tay
	lda $9645,y
	sta $078a,x
	lda $07a8,x
	ora #$c0
	sta $3c,x
	ldy #$fc
	lda $09
	and #$40
	bne $9697
	ldy #$04
	sty $58,x
	ldy #$01
	jsr $add9
	bmi $96aa
	iny
	lda $0796,x
	bne $96aa
	lda #$f8
	sta $58,x
	sty $46,x
	ldy #$00
	lda $1e,x
	and #$40
	bne $96cd
	lda $1e,x
	asl
	bcs $96e9
	lda $1e,x
	and #$20
	bne $971a
	lda $1e,x
	and #$07
	beq $96e9
	cmp #$05
	beq $96cd
	cmp #$03
	bcs $96fd
	jsr $8b34
	ldy #$00
	lda $1e,x
	cmp #$02
	beq $96e4
	and #$40
	beq $96e9
	lda $16,x
	cmp #$2e
	beq $96e9
	bne $96e7
	jmp $8ad3

	ldy #$01
	lda $58,x
	pha
	bpl $96f0
	iny
	iny
	clc
	adc $9605,y
	sta $58,x
	jsr $8ad3
	pla
	sta $58,x
	rts

	lda $0796,x
	bne $9720
	sta $1e,x
	lda $09
	and #$01
	tay
	iny
	sty $46,x
	dey
	lda $076a
	beq $9714
	iny
	iny
	lda $9609,y
	sta $58,x
	rts

	jsr $8b34
	jmp $8ad3

	cmp #$0e
	bne $972d
	lda $16,x
	cmp #$06
	bne $972d
	jsr $95cd
	rts

	jsr $8b63
	jmp $8ad3

	lda $a0,x
	ora $0434,x
	bne $974e
	sta $0417,x
	lda $cf,x
	cmp $0401,x
	bcs $974e
	lda $09
	and #$07
	bne $974d
	inc $cf,x
	rts

	lda $cf,x
	cmp $58,x
	bcc $9757
	jmp $8b46

	jmp $8b41

	jsr $977a
	jsr $979b
	ldy #$01
	lda $09
	and #$03
	bne $9779
	lda $09
	and #$40
	bne $9770
	ldy #$ff
	sty $00
	lda $cf,x
	clc
	adc $00
	sta $cf,x
	rts

	lda #$13
	sta $01
	lda $09
	and #$03
	bne $9791
	ldy $58,x
	lda $a0,x
	lsr
	bcs $9795
	cpy $01
	beq $9792
	inc $58,x
	rts

	inc $a0,x
	rts

	tya
	beq $9792
	dec $58,x
	rts

	lda $58,x
	pha
	ldy #$01
	lda $a0,x
	and #$02
	bne $97b1
	lda $58,x
	eor #$ff
	clc
	adc #$01
	sta $58,x
	ldy #$02
	sty $46,x
	jsr $8ad3
	sta $00
	pla
	sta $58,x
	rts

	.byte $3f
	.byte $03
	lda $1e,x
	and #$20
	bne $9811
	ldy $06cc
	lda $07a8,x
	and $97bc,y
	bne $97e1
	txa
	lsr
	bcc $97d7
	ldy $45
	bcs $97df
	ldy #$02
	jsr $add9
	bpl $97df
	dey
	sty $46,x
	jsr $9814
	lda $cf,x
	sec
	sbc $0434,x
	cmp #$20
	bcc $97f0
	sta $cf,x
	ldy $46,x
	dey
	bne $9803
	lda $87,x
	clc
	adc $58,x
	sta $87,x
	lda $6e,x
	adc #$00
	sta $6e,x
	rts

	lda $87,x
	sec
	sbc $58,x
	sta $87,x
	lda $6e,x
	sbc #$00
	sta $6e,x
	rts

	jmp $8b5d

	lda $a0,x
	and #$02
	bne $9851
	lda $09
	and #$07
	pha
	lda $a0,x
	lsr
	bcs $9839
	pla
	bne $9838
	lda $0434,x
	clc
	adc #$01
	sta $0434,x
	sta $58,x
	cmp #$02
	bne $9838
	inc $a0,x
	rts

	pla
	bne $9850
	lda $0434,x
	sec
	sbc #$01
	sta $0434,x
	sta $58,x
	bne $9850
	inc $a0,x
	lda #$02
	sta $0796,x
	rts

	lda $0796,x
	beq $985e
	lda $09
	lsr
	bcs $985d
	inc $cf,x
	rts

	lda $cf,x
	adc #$10
	cmp $ce
	bcc $9856
	lda #$00
	sta $a0,x
	rts

	lda $1e,x
	and #$20
	beq $9874
	.byte $4c
	.byte $63
	.byte $8b
	lda #$e8
	sta $58,x
	jmp $8ad3

	.byte $40
	.byte $80
	.byte $04
	.byte $04
	lda $1e,x
	and #$20
	beq $9888
	jmp $8b5d

	sta $03
	lda $16,x
	sec
	sbc #$0a
	tay
	lda $987b,y
	sta $02
	lda $0401,x
	sec
	sbc $02
	sta $0401,x
	lda $87,x
	sbc #$00
	sta $87,x
	lda $6e,x
	sbc #$00
	sta $6e,x
	lda #$40
	sta $02
	cpx #$02
	bcc $98fb
	lda $58,x
	cmp #$10
	bcc $98ce
	lda $0417,x
	clc
	adc $02
	sta $0417,x
	lda $cf,x
	adc $03
	sta $cf,x
	lda $b6,x
	adc #$00
	jmp $98e1

	lda $0417,x
	sec
	sbc $02
	sta $0417,x
	lda $cf,x
	sbc $03
	sta $cf,x
	lda $b6,x
	sbc #$00
	sta $b6,x
	ldy #$00
	lda $cf,x
	sec
	sbc $0434,x
	bpl $98f4
	ldy #$10
	eor #$ff
	clc
	adc #$01
	cmp #$0f
	bcc $98fb
	tya
	sta $58,x
	rts

	.byte $00
	.byte $01
	.byte $03
	.byte $04
	.byte $05
	.byte $06
	.byte $07
	.byte $07
	.byte $08
	.byte $00
	.byte $03
	.byte $06
	.byte $09
	.byte $0b
	.byte $0d
	.byte $0e
	.byte $0f
	.byte $10
	.byte $00
	.byte $04
	.byte $09
	.byte $0d
	.byte $10
	.byte $13
	.byte $16
	.byte $17
	.byte $18
	.byte $00
	.byte $06
	.byte $0c
	.byte $12
	.byte $16
	.byte $1a
	.byte $1d
	.byte $1f
	.byte $20
	.byte $00
	.byte $07
	.byte $0f
	.byte $16
	.byte $1c
	.byte $21
	.byte $25
	.byte $27
	.byte $28
	.byte $00
	.byte $09
	.byte $12
	.byte $1b
	.byte $21
	.byte $27
	.byte $2c
	.byte $2f
	.byte $30
	.byte $00
	.byte $0b
	.byte $15
	.byte $1f
	.byte $27
	.byte $2e
	.byte $33
	.byte $37
	.byte $38
	.byte $00
	.byte $0c
	.byte $18
	.byte $24
	.byte $2d
	.byte $35
	.byte $3b
	.byte $3e
	.byte $40
	.byte $00
	.byte $0e
	.byte $1b
	.byte $28
	.byte $32
	.byte $3b
	.byte $42
	.byte $46
	.byte $48
	.byte $00
	.byte $0f
	.byte $1f
	.byte $2d
	.byte $38
	.byte $42
	.byte $4a
	.byte $4e
	.byte $50
	.byte $00
	.byte $11
	.byte $22
	.byte $31
	.byte $3e
	.byte $49
	.byte $51
	.byte $56
	.byte $58
	.byte $01
	.byte $03
	.byte $02
	.byte $00
	.byte $00
	.byte $09
	.byte $12
	.byte $1b
	.byte $24
	.byte $2d
	.byte $36
	.byte $3f
	.byte $48
	.byte $51
	.byte $5a
	.byte $63
	.byte $0c
	.byte $18
	jsr $be94
	lda $03d1
	and #$08
	bne $99ef
	lda $0747
	bne $998a
	lda $0388,x
	jsr $a043
	and #$1f
	sta $a0,x
	lda $a0,x
	ldy $16,x
	cpy #$1f
	bcc $999f
	cmp #$08
	beq $999a
	cmp #$18
	bne $999f
	clc
	adc #$01
	sta $a0,x
	sta $ef
	jsr $be37
	jsr $9ac3
	ldy $06e5,x
	lda $03b9
	sta $0200,y
	sta $07
	lda $03ae
	sta $0203,y
	sta $06
	lda #$01
	sta $00
	jsr $9a3d
	ldy #$05
	lda $16,x
	cmp #$1f
	bcc $99cb
	ldy #$0b
	sty $ed
	lda #$00
	sta $00
	lda $ef
	jsr $9ac3
	jsr $99f0
	lda $00
	cmp #$04
	bne $99e7
	ldy $06cf
	lda $06e5,y
	sta $06
	inc $00
	lda $00
	cmp $ed
	bcc $99d1
	rts

	lda $03
	sta $05
	ldy $06
	lda $01
	lsr $05
	bcs $9a00
	eor #$ff
	adc #$01
	clc
	adc $03ae
	sta $0203,y
	sta $06
	cmp $03ae
	bcs $9a17
	lda $03ae
	sec
	sbc $06
	jmp $9a1b

	sec
	sbc $03ae
	cmp #$59
	bcc $9a23
	lda #$f8
	bne $9a38
	lda $03b9
	cmp #$f8
	beq $9a38
	lda $02
	lsr $05
	bcs $9a34
	eor #$ff
	adc #$01
	clc
	adc $03b9
	sta $0200,y
	sta $07
	jsr $b9c8
	tya
	pha
	lda $079f
	ora $0747
	bne $9aba
	sta $05
	ldy $b5
	dey
	bne $9aba
	ldy $ce
	lda $0754
	bne $9a5d
	lda $0714
	beq $9a66
	inc $05
	inc $05
	tya
	clc
	adc #$18
	tay
	tya
	sec
	sbc $07
	bpl $9a71
	eor #$ff
	clc
	adc #$01
	cmp #$08
	bcs $9a91
	lda $06
	cmp #$f0
	bcs $9a91
	lda $0207
	clc
	adc #$04
	sta $04
	sec
	sbc $06
	bpl $9a8d
	eor #$ff
	clc
	adc #$01
	cmp #$08
	bcc $9aa4
	lda $05
	cmp #$02
	beq $9aba
	ldy $05
	lda $ce
	clc
	adc $996f,y
	inc $05
	jmp $9a67

	ldx #$01
	lda $04
	cmp $06
	bcs $9aad
	inx
	stx $46
	ldx #$00
	lda $00
	pha
	jsr $a587
	pla
	sta $00
	pla
	clc
	adc #$04
	sta $06
	ldx $08
	rts

	pha
	and #$0f
	cmp #$09
	bcc $9acf
	eor #$0f
	clc
	adc #$01
	sta $01
	ldy $00
	lda $9963,y
	clc
	adc $01
	tay
	lda $98fc,y
	sta $01
	pla
	pha
	clc
	adc #$08
	and #$0f
	cmp #$09
	bcc $9aef
	eor #$0f
	clc
	adc #$01
	sta $02
	ldy $00
	lda $9963,y
	clc
	adc $02
	tay
	lda $98fc,y
	sta $02
	pla
	lsr
	lsr
	lsr
	tay
	lda $995f,y
	sta $03
	rts

	.byte $f8
	.byte $a0
	.byte $70
	.byte $bd
	.byte $00
	.byte $20
	.byte $20
	.byte $20
	.byte $00
	.byte $00
	lda $1e,x
	and #$20
	beq $9b22
	lda #$00
	sta $03c5,x
	jmp $8b63

	jsr $8ad3
	ldy #$0d
	lda #$05
	jsr $8b67
	lda $0434,x
	lsr
	lsr
	lsr
	lsr
	tay
	lda $cf,x
	sec
	sbc $9b0a,y
	bpl $9b41
	eor #$ff
	clc
	adc #$01
	cmp #$08
	bcs $9b53
	lda $0434,x
	clc
	adc #$10
	sta $0434,x
	lsr
	lsr
	lsr
	lsr
	tay
	lda $9b0f,y
	sta $03c5,x
	rts

	.byte $15
	.byte $30
	.byte $40
	lda $1e,x
	and #$20
	beq $9b66
	jmp $8b34

	lda $1e,x
	beq $9b75
	lda #$00
	sta $a0,x
	sta $06cb
	lda #$10
	bne $9b88
	lda #$12
	sta $06cb
	ldy #$02
	lda $9b5a,y
	sta $0001,y
	dey
	bpl $9b7c
	jsr $9ba1
	sta $58,x
	ldy #$01
	lda $a0,x
	and #$01
	bne $9b9c
	lda $58,x
	eor #$ff
	clc
	adc #$01
	sta $58,x
	iny
	sty $46,x
	jmp $8ad3

	ldy #$00
	jsr $add9
	bpl $9bb2
	iny
	lda $00
	eor #$ff
	clc
	adc #$01
	sta $00
	lda $00
	cmp #$3c
	bcc $9bd4
	lda #$3c
	sta $00
	lda $16,x
	cmp #$11
	bne $9bd4
	tya
	cmp $a0,x
	beq $9bd4
	lda $a0,x
	beq $9bd1
	dec $58,x
	lda $58,x
	bne $9c11
	tya
	sta $a0,x
	lda $00
	and #$3c
	lsr
	lsr
	sta $00
	ldy #$00
	lda $57
	beq $9c06
	lda $0775
	beq $9c06
	iny
	lda $57
	cmp #$19
	bcc $9bf6
	lda $0775
	cmp #$02
	bcc $9bf6
	iny
	lda $16,x
	cmp #$12
	bne $9c00
	lda $57
	bne $9c06
	lda $a0,x
	bne $9c06
	ldy #$00
	lda $0001,y
	ldy $00
	sec
	sbc #$01
	dey
	bpl $9c0b
	rts

	.byte $1a
	.byte $58
	.byte $98
	.byte $96
	.byte $94
	.byte $92
	.byte $90
	.byte $8e
	.byte $8c
	.byte $8a
	.byte $88
	.byte $86
	.byte $84
	.byte $82
	.byte $80
	ldx $0368
	lda $16,x
	cmp #$2d
	bne $9c3a
	stx $08
	lda $1e,x
	beq $9c4a
	and #$40
	beq $9c3a
	lda $cf,x
	cmp #$e0
	bcc $9c44
	lda #$80
	sta $fc
	inc $0772
	jmp $9ca6

	jsr $8b5d
	jmp $9db0

	dec $0364
	bne $9c93
	lda #$04
	sta $0364
	lda $0363
	eor #$01
	sta $0363
	lda #$22
	sta $05
	ldy $0369
	lda $9c12,y
	sta $04
	ldy $0300
	iny
	ldx #$0c
	jsr $69aa
	ldx $08
	jsr $696c
	lda #$08
	sta $fe
	lda #$01
	sta $fd
	inc $0369
	lda $0369
	cmp #$0f
	bne $9c93
	jsr $8f4e
	lda #$40
	sta $1e,x
	lda #$80
	sta $fe
	jmp $9db0

	.byte $21
	.byte $41
	.byte $11
	.byte $31
	lda $1e,x
	and #$20
	beq $9cb4
	lda $cf,x
	cmp #$e0
	bcc $9c44
	ldx #$04
	jsr $95cd
	dex
	bpl $9ca8
	sta $06cb
	ldx $08
	rts

	lda #$00
	sta $06cb
	lda $0747
	beq $9cc1
	jmp $9d6e

	lda $0363
	bpl $9cc9
	jmp $9d44

	dec $0364
	bne $9cdb
	lda #$20
	sta $0364
	lda $0363
	eor #$01
	sta $0363
	lda $09
	and #$0f
	bne $9ce5
	lda #$02
	sta $46,x
	lda $078a,x
	beq $9d06
	jsr $add9
	bpl $9d06
	lda #$01
	sta $46,x
	lda #$02
	sta $0365
	lda #$20
	sta $078a,x
	sta $0790
	lda $87,x
	cmp #$c8
	bcs $9d44
	lda $09
	and #$03
	bne $9d44
	lda $87,x
	cmp $0366
	bne $9d1f
	lda $07a7,x
	and #$03
	tay
	lda $9c96,y
	sta $06dc
	lda $87,x
	clc
	adc $0365
	sta $87,x
	ldy $46,x
	cpy #$01
	beq $9d44
	ldy #$ff
	sec
	sbc $0366
	bpl $9d3c
	eor #$ff
	clc
	adc #$01
	ldy #$01
	cmp $06dc
	bcc $9d44
	sty $0365
	lda $078a,x
	bne $9d71
	jsr $8b5d
	lda $075f
	cmp #$05
	bcc $9d5c
	lda $09
	and #$03
	bne $9d5c
	jsr $865f
	lda $cf,x
	cmp #$80
	bcc $9d7e
	lda $07a7,x
	and #$03
	tay
	lda $9c96,y
	sta $078a,x
	jmp $9d7e

	cmp #$01
	bne $9d7e
	dec $cf,x
	jsr $8f4e
	lda #$fe
	sta $a0,x
	lda $075f
	cmp #$07
	beq $9d89
	cmp #$05
	bcs $9db0
	lda $0790
	bne $9db0
	lda #$20
	sta $0790
	lda $0363
	eor #$80
	sta $0363
	bmi $9d7e
	jsr $9e0e
	ldy $06cc
	beq $9da8
	sec
	sbc #$10
	sta $0790
	lda #$15
	sta $06cb
	jsr $9df1
	ldy #$10
	lda $46,x
	lsr
	bcc $9dbc
	ldy #$f0
	tya
	clc
	adc $87,x
	ldy $06cf
	sta $0087,y
	lda $cf,x
	clc
	adc #$08
	sta $00cf,y
	lda $1e,x
	sta $001e,y
	lda $46,x
	sta $0046,y
	lda $08
	pha
	ldx $06cf
	stx $08
	lda #$2d
	sta $16,x
	jsr $9df1
	pla
	sta $08
	tax
	lda #$00
	sta $036a
	rts

	inc $036a
	jsr $950c
	lda $1e,x
	bne $9df0
	lda #$0a
	sta $049a,x
	jsr $aee1
	jmp $a4ab

	.byte $bf
	.byte $40
	.byte $bf
	.byte $bf
	.byte $bf
	.byte $40
	.byte $40
	.byte $bf
	ldy $0367
	inc $0367
	lda $0367
	and #$07
	sta $0367
	lda $9e06,y
	rts

	lda $0747
	bne $9e55
	lda #$40
	ldy $06cc
	beq $9e2e
	lda #$60
	sta $00
	lda $0401,x
	sec
	sbc $00
	sta $0401,x
	lda $87,x
	sbc #$01
	sta $87,x
	lda $6e,x
	sbc #$00
	sta $6e,x
	ldy $0417,x
	lda $cf,x
	cmp $91ae,y
	beq $9e55
	clc
	adc $0434,x
	sta $cf,x
	jsr $be37
	lda $1e,x
	bne $9e1f
	lda #$51
	sta $00
	ldy #$02
	lda $09
	and #$02
	beq $9e6a
	ldy #$82
	sty $01
	ldy $06e5,x
	ldx #$00
	lda $03b9
	sta $0200,y
	lda $00
	sta $0201,y
	inc $00
	lda $01
	sta $0202,y
	lda $03ae
	sta $0203,y
	clc
	adc #$08
	sta $03ae
	iny
	iny
	iny
	iny
	inx
	cpx #$03
	bcc $9e71
	ldx $08
	jsr $be94
	ldy $06e5,x
	lda $03d1
	lsr
	pha
	bcc $9eac
	lda #$f8
	sta $020c,y
	pla
	lsr
	pha
	bcc $9eb6
	lda #$f8
	sta $0208,y
	pla
	lsr
	pha
	bcc $9ec0
	lda #$f8
	sta $0204,y
	pla
	lsr
	bcc $9ec9
	lda #$f8
	sta $0200,y
	rts

	dec $a0,x
	bne $9eda
	lda #$08
	sta $a0,x
	inc $58,x
	lda $58,x
	cmp #$03
	bcs $9ef2
	jsr $be37
	lda $03b9
	sta $03ba
	lda $03ae
	sta $03af
	ldy $06e5,x
	lda $58,x
	jsr $b9f2
	rts

	lda #$00
	sta $0f,x
	lda #$08
	sta $fe
	lda #$05
	sta $0138
	jmp $9f70

	.byte $00
	.byte $00
	.byte $08
	.byte $08
	.byte $00
	.byte $08
	.byte $00
	.byte $08
	.byte $54
	.byte $55
	.byte $56
	.byte $57
	lda #$00
	sta $06cb
	lda $0746
	cmp #$05
	bcs $9f4b
	jsr $6c7d
	.byte $4b
	.byte $9f
	.byte $27
	.byte $9f
	.byte $4c
	.byte $9f
	.byte $7a
	.byte $9f
	.byte $ce
	.byte $9f
	lda $07ee
	cmp $07e8
	bne $9f3f
	and #$01
	beq $9f39
	ldy #$03
	lda #$03
	bne $9f43
	ldy #$00
	lda #$06
	bne $9f43
	ldy #$00
	lda #$ff
	sta $06d7
	sty $1e,x
	inc $0746
	rts

	lda $07ec
	ora $07ed
	ora $07ee
	beq $9f48
	lda $09
	and #$04
	beq $9f61
	lda #$10
	sta $fe
	ldy #$17
	lda #$ff
	sta $0139
	jsr $6dd2
	lda #$05
	sta $0139
	ldy #$0b
	jsr $6dd2
	lda #$02
	jmp $87ef

	lda $cf,x
	cmp #$72
	bcc $9f85
	dec $cf,x
	jmp $9f91

	lda $06d7
	beq $9fc2
	bmi $9fc2
	lda #$16
	sta $06cb
	jsr $be37
	ldy $06e5,x
	ldx #$03
	lda $03b9
	clc
	adc $9f02,x
	sta $0200,y
	lda $9f0a,x
	sta $0201,y
	lda #$22
	sta $0202,y
	lda $03ae
	clc
	adc $9f06,x
	sta $0203,y
	iny
	iny
	iny
	iny
	dex
	bpl $9f99
	ldx $08
	rts

	jsr $9f91
	lda #$06
	sta $0796,x
	inc $0746
	rts

	jsr $9f91
	lda $0796,x
	bne $9fdb
	lda $07b1
	beq $9fca
	rts

	lda $1e,x
	bne $a03d
	lda $078a,x
	bne $a03d
	lda $a0,x
	bne $a00c
	lda $58,x
	bmi $a001
	jsr $add9
	bpl $9ffb
	lda $00
	eor #$ff
	clc
	adc #$01
	sta $00
	lda $00
	cmp #$21
	bcc $a03d
	lda $58,x
	eor #$ff
	clc
	adc #$01
	sta $58,x
	inc $a0,x
	lda $0434,x
	ldy $58,x
	bpl $a016
	lda $0417,x
	sta $00
	lda $b517
	cmp #$22
	beq $a024
	lda $09
	lsr
	bcc $a03d
	lda $0747
	bne $a03d
	lda $cf,x
	clc
	adc $58,x
	sta $cf,x
	cmp $00
	bne $a03d
	lda #$00
	sta $a0,x
	lda #$40
	sta $078a,x
	lda #$20
	sta $03c5,x
	rts

	sta $07
	lda $34,x
	bne $a057
	ldy #$18
	lda $58,x
	clc
	adc $07
	sta $58,x
	lda $a0,x
	adc #$00
	rts

	ldy #$08
	lda $58,x
	sec
	sbc $07
	sta $58,x
	lda $a0,x
	sbc #$00
	rts

	lda $b6,x
	cmp #$03
	bne $a06e
	jmp $95cd

	lda $1e,x
	bpl $a073
	rts

	tay
	lda $0016,y
	cmp #$24
	bne $a072
	lda $03a2,x
	sta $00
	lda $46,x
	beq $a087
	jmp $a1f5

	lda #$2d
	cmp $cf,x
	bcc $a09c
	cpy $00
	beq $a099
	clc
	adc #$02
	sta $cf,x
	jmp $a1eb

	jmp $a1d2

	cmp $00cf,y
	bcc $a0ae
	cpx $00
	beq $a099
	clc
	adc #$02
	sta $00cf,y
	jmp $a1eb

	lda $cf,x
	pha
	lda $03a2,x
	bpl $a0ce
	lda $0434,x
	clc
	adc #$05
	sta $00
	lda $a0,x
	adc #$00
	bmi $a0de
	bne $a0d2
	lda $00
	cmp #$0b
	bcc $a0d8
	bcs $a0d2
	cmp $08
	beq $a0de
	jsr $8b88
	jmp $a0e1

	jsr $a1eb
	jmp $a0e1

	jsr $8b85
	ldy $1e,x
	pla
	sec
	sbc $cf,x
	clc
	adc $00cf,y
	sta $00cf,y
	lda $03a2,x
	bmi $a0f7
	tax
	jsr $a896
	ldy $08
	lda $00a0,y
	ora $0434,y
	beq $a178
	ldx $0300
	cpx #$20
	bcs $a178
	lda $00a0,y
	pha
	pha
	jsr $a17b
	lda $01
	sta $0301,x
	lda $00
	sta $0302,x
	lda #$02
	sta $0303,x
	lda $00a0,y
	bmi $a131
	lda #$a2
	sta $0304,x
	lda #$a3
	sta $0305,x
	jmp $a139

	lda #$24
	sta $0304,x
	sta $0305,x
	lda $001e,y
	tay
	pla
	eor #$ff
	jsr $a17b
	lda $01
	sta $0306,x
	lda $00
	sta $0307,x
	lda #$02
	sta $0308,x
	pla
	bpl $a162
	lda #$a2
	sta $0309,x
	lda #$a3
	sta $030a,x
	jmp $a16a

	lda #$24
	sta $0309,x
	sta $030a,x
	lda #$00
	sta $030b,x
	lda $0300
	clc
	adc #$0a
	sta $0300
	ldx $08
	rts

	pha
	lda $0087,y
	clc
	adc #$08
	ldx $06cc
	bne $a18a
	clc
	adc #$10
	pha
	lda $006e,y
	adc #$00
	sta $02
	pla
	and #$f0
	lsr
	lsr
	lsr
	sta $00
	ldx $cf,y
	pla
	bpl $a1a4
	txa
	clc
	adc #$08
	tax
	txa
	ldx $0300
	asl
	rol
	pha
	rol
	and #$03
	ora #$20
	sta $01
	lda $02
	and #$01
	asl
	asl
	ora $01
	sta $01
	pla
	and #$e0
	clc
	adc $00
	sta $00
	lda $00cf,y
	cmp #$e8
	bcc $a1d1
	lda $00
	and #$bf
	sta $00
	rts

	tya
	tax
	jsr $be94
	lda #$06
	jsr $a67a
	lda $03ad
	sta $0117,x
	lda $ce
	sta $011e,x
	lda #$01
	sta $46,x
	jsr $8f4e
	sta $00a0,y
	sta $0434,y
	rts

	tya
	pha
	jsr $8b3c
	pla
	tax
	jsr $8b3c
	ldx $08
	lda $03a2,x
	bmi $a20a
	tax
	jsr $a896
	ldx $08
	rts

	lda $a0,x
	ora $0434,x
	bne $a229
	sta $0417,x
	lda $cf,x
	cmp $0401,x
	bcs $a229
	lda $09
	and #$07
	bne $a226
	inc $cf,x
	jmp $a238

	lda $cf,x
	cmp $58,x
	bcc $a235
	jsr $8b88
	jmp $a238

	jsr $8b85
	lda $03a2,x
	bmi $a240
	jsr $a896
	rts

	lda #$0e
	jsr $977c
	jsr $979b
	lda $03a2,x
	bmi $a26a
	lda $86
	clc
	adc $00
	sta $86
	lda $6d
	ldy $00
	bmi $a260
	adc #$00
	jmp $a262

	sbc #$00
	sta $6d
	sty $03a1
	jsr $a896
	rts

	lda $03a2,x
	bmi $a276
	jsr $8b59
	jsr $a896
	rts

	jsr $8ad3
	sta $00
	lda $03a2,x
	bmi $a288
	lda #$10
	sta $58,x
	jsr $a24e
	rts

	jsr $a295
	jmp $a238

	jsr $a295
	jmp $a2ab

	lda $0747
	bne $a2b3
	lda $0417,x
	clc
	adc $0434,x
	sta $0417,x
	lda $cf,x
	adc $a0,x
	sta $cf,x
	rts

	lda $03a2,x
	beq $a2b3
	jsr $a88e
	rts

	lda $16,x
	cmp #$14
	beq $a317
	lda $071c
	ldy $16,x
	cpy #$05
	beq $a2cb
	cpy #$04
	beq $a2cb
	cpy #$0d
	bne $a2cd
	adc #$38
	sbc #$48
	sta $01
	lda $071a
	sbc #$00
	sta $00
	lda $071d
	adc #$48
	sta $03
	lda $071b
	adc #$00
	sta $02
	lda $87,x
	cmp $01
	lda $6e,x
	sbc $00
	bmi $a314
	lda $87,x
	cmp $03
	lda $6e,x
	sbc $02
	bmi $a317
	lda $1e,x
	cmp #$05
	beq $a317
	cpy #$0d
	beq $a317
	cpy #$04
	beq $a317
	cpy #$30
	beq $a317
	cpy #$31
	beq $a317
	cpy #$32
	beq $a317
	jsr $95cd
	rts

	.byte $ff
	lda $24,x
	beq $a373
	asl
	bcs $a373
	lda $09
	lsr
	bcs $a373
	txa
	asl
	asl
	clc
	adc #$1c
	tay
	ldx #$04
	stx $01
	tya
	pha
	lda $1e,x
	and #$20
	bne $a36c
	lda $0f,x
	beq $a36c
	lda $16,x
	cmp #$24
	bcc $a346
	cmp #$2b
	bcc $a36c
	cmp #$06
	bne $a350
	lda $1e,x
	cmp #$02
	bcs $a36c
	lda $03d8,x
	bne $a36c
	txa
	asl
	asl
	clc
	adc #$04
	tax
	jsr $afc5
	ldx $08
	bcc $a36c
	lda #$80
	sta $24,x
	ldx $01
	jsr $a37f
	pla
	tay
	ldx $01
	dex
	bpl $a32e
	ldx $08
	rts

	.byte $06
	.byte $00
	.byte $02
	.byte $12
	.byte $11
	.byte $07
	.byte $05
	.byte $2d
	.byte $2d
	jsr $be37
	ldx $01
	lda $0f,x
	bpl $a393
	and #$0f
	tax
	lda $16,x
	cmp #$2d
	beq $a39d
	ldx $01
	lda $16,x
	cmp #$02
	beq $a40f
	cmp #$2d
	bne $a3ca
	dec $0483
	bne $a40f
	jsr $8f4e
	sta $58,x
	sta $06cb
	lda #$fe
	sta $a0,x
	ldy $075f
	lda $a376,y
	sta $16,x
	lda #$20
	cpy #$03
	bcs $a3be
	ora #$03
	sta $1e,x
	lda #$80
	sta $fe
	ldx $01
	lda #$09
	bne $a408
	cmp #$08
	beq $a40f
	cmp #$0c
	beq $a40f
	cmp #$15
	bcs $a40f
	lda $16,x
	cmp #$04
	beq $a3e0
	cmp #$0d
	bne $a3ed
	tay
	lda $cf,x
	adc #$18
	cpy #$04
	bne $a3eb
	sbc #$31
	sta $cf,x
	jsr $aca3
	lda $1e,x
	and #$1f
	ora #$20
	sta $1e,x
	lda #$02
	ldy $16,x
	cpy #$05
	bne $a402
	lda #$06
	cpy #$06
	bne $a408
	lda #$01
	jsr $a67a
	lda #$08
	sta $ff
	rts

	lda $09
	lsr
	bcc $a44e
	lda $03d0
	ora $0747
	ora $03d6
	bne $a44e
	txa
	asl
	asl
	clc
	adc #$24
	tay
	jsr $afc3
	ldx $08
	bcc $a449
	lda $06be,x
	bne $a44e
	lda #$01
	sta $06be,x
	lda $64,x
	eor #$ff
	clc
	adc #$01
	sta $64,x
	lda $079f
	bne $a44e
	jmp $a587

	lda #$00
	sta $06be,x
	rts

	jsr $95cd
	lda $39
	cmp #$04
	bne $a45b
	jmp $a587

	lda #$06
	jsr $a67a
	lda #$20
	sta $fe
	lda $39
	cmp #$02
	bcc $a478
	cmp #$03
	beq $a492
	lda #$23
	sta $079f
	lda #$40
	sta $fb
	rts

	lda $0756
	beq $a498
	cmp #$01
	bne $a4a4
	ldx $08
	lda #$02
	sta $0756
	jsr $654d
	ldx $08
	lda #$0c
	jmp $a49f

	lda #$0b
	sta $0110,x
	rts

	lda #$01
	sta $0756
	lda #$09
	ldy #$00
	jsr $a5a6
	rts

	.byte $18
	.byte $e8
	.byte $30
	.byte $d0
	.byte $08
	.byte $f8
	lda $09
	lsr
	bcs $a4a4
	jsr $a8b6
	bcs $a4d8
	lda $03d8,x
	bne $a4d8
	lda $0e
	cmp #$08
	bne $a4d8
	lda $1e,x
	and #$20
	bne $a4d8
	jsr $a8c0
	jsr $afc3
	ldx $08
	bcs $a4d9
	lda $0491,x
	and #$fe
	sta $0491,x
	rts

	ldy $16,x
	cpy #$2e
	bne $a4e2
	jmp $a44f

	lda $079f
	beq $a4ed
	jmp $a3d6

	.byte $0a
	.byte $06
	.byte $04
	lda $0491,x
	and #$01
	ora $03d8,x
	bne $a554
	lda #$01
	ora $0491,x
	sta $0491,x
	cpy #$12
	beq $a555
	cpy #$33
	beq $a555
	cpy #$0d
	beq $a587
	cpy #$04
	beq $a587
	cpy #$0c
	beq $a587
	cpy #$15
	bcs $a587
	lda $074e
	beq $a587
	lda $1e,x
	asl
	bcs $a555
	lda $1e,x
	and #$07
	cmp #$02
	bcc $a555
	lda $16,x
	cmp #$06
	beq $a554
	lda #$08
	sta $ff
	lda $1e,x
	ora #$80
	sta $1e,x
	jsr $a66e
	lda $a4a7,y
	sta $58,x
	lda #$03
	clc
	adc $0484
	ldy $0796,x
	cpy #$03
	bcs $a551
	lda $a4ea,y
	jsr $a67a
	rts

	ldy $9f
	dey
	bpl $a5c7
	lda $16,x
	cmp #$07
	bcc $a569
	lda $ce
	clc
	adc #$0c
	cmp $cf,x
	bcc $a5c7
	lda $0791
	bne $a5c7
	lda $079e
	bne $a5b3
	lda $03ad
	cmp $03ae
	bcc $a57e
	jmp $a65f

	lda $46,x
	cmp #$01
	bne $a587
	jmp $a668

	lda $079e
	ora $079f
	bne $a5b3
	ldx $0756
	beq $a5b6
	sta $0756
	lda #$08
	sta $079e
	asl
	sta $ff
	jsr $654d
	lda #$0a
	ldy #$01
	sta $0e
	sty $1d
	ldy #$ff
	sty $0747
	iny
	sty $0775
	ldx $08
	rts

	stx $57
	inx
	stx $fc
	lda #$fc
	sta $9f
	lda #$0b
	bne $a5a4
	.byte $02
	.byte $06
	.byte $05
	.byte $06
	lda $16,x
	cmp #$12
	beq $a587
	lda #$04
	sta $ff
	lda $16,x
	ldy #$00
	cmp #$14
	beq $a5f4
	cmp #$08
	beq $a5f4
	cmp #$33
	beq $a5f4
	cmp #$0c
	beq $a5f4
	iny
	cmp #$05
	beq $a5f4
	iny
	cmp #$11
	beq $a5f4
	iny
	cmp #$07
	bne $a60f
	lda $a5c3,y
	jsr $a67a
	lda $46,x
	pha
	jsr $acc1
	pla
	sta $46,x
	lda #$20
	sta $1e,x
	jsr $8f4e
	sta $58,x
	jmp $a64e

	cmp #$09
	bcc $a631
	jsr $a64e
	and #$01
	sta $16,x
	lda #$00
	sta $1e,x
	lda #$03
	jsr $a67a
	jsr $8f4e
	jsr $a66e
	lda $a4a9,y
	sta $58,x
	rts

	.byte $10
	.byte $0b
	lda #$04
	sta $1e,x
	inc $0484
	lda $0484
	clc
	adc $0791
	jsr $a67a
	inc $0791
	ldy $076a
	lda $a62f,y
	sta $0796,x
	ldy #$fa
	lda $16,x
	cmp #$0f
	beq $a65a
	cmp #$10
	bne $a65c
	ldy #$f8
	sty $9f
	rts

	lda $46,x
	cmp #$01
	bne $a668
	jmp $a587

	jsr $a78d
	jmp $a587

	ldy #$01
	jsr $add9
	bpl $a676
	iny
	sty $46,x
	dey
	rts

	sta $0110,x
	lda #$30
	sta $012c,x
	lda $cf,x
	sta $011e,x
	lda $03ae
	sta $0117,x
	rts

	.byte $80
	.byte $40
	.byte $20
	.byte $10
	.byte $08
	.byte $04
	.byte $02
	.byte $7f
	.byte $bf
	.byte $df
	.byte $ef
	.byte $f7
	.byte $fb
	.byte $fd
	lda $09
	lsr
	bcc $a68d
	lda $074e
	beq $a68d
	lda $16,x
	cmp #$15
	bcs $a722
	cmp #$11
	beq $a722
	cmp #$0d
	beq $a722
	cmp #$04
	beq $a722
	lda $03d8,x
	bne $a722
	jsr $a8c0
	dex
	bmi $a722
	stx $01
	tya
	pha
	lda $0f,x
	beq $a71b
	lda $16,x
	cmp #$15
	bcs $a71b
	cmp #$11
	beq $a71b
	cmp #$0d
	beq $a71b
	cmp #$04
	beq $a71b
	lda $03d8,x
	bne $a71b
	txa
	asl
	asl
	clc
	adc #$04
	tax
	jsr $afc5
	ldx $08
	ldy $01
	bcc $a712
	lda $1e,x
	ora $001e,y
	and #$80
	bne $a70c
	lda $0491,y
	and $a68e,x
	bne $a71b
	lda $0491,y
	ora $a68e,x
	sta $0491,y
	jsr $a725
	jmp $a71b

	lda $0491,y
	and $a695,x
	sta $0491,y
	pla
	tay
	ldx $01
	dex
	bpl $a6c3
	ldx $08
	rts

	lda $001e,y
	ora $1e,x
	and #$20
	bne $a761
	lda $1e,x
	cmp #$06
	bcc $a762
	lda $16,x
	cmp #$05
	beq $a761
	lda $001e,y
	asl
	bcc $a74a
	lda #$06
	jsr $a67a
	jsr $a3d6
	ldy $01
	tya
	tax
	jsr $a3d6
	ldx $08
	lda $0125,x
	clc
	adc #$04
	ldx $01
	jsr $a67a
	ldx $08
	inc $0125,x
	rts

	lda $001e,y
	cmp #$06
	bcc $a786
	lda $0016,y
	cmp #$05
	beq $a761
	jsr $a3d6
	ldy $01
	lda $0125,y
	clc
	adc #$04
	ldx $08
	jsr $a67a
	ldx $01
	inc $0125,x
	rts

	tya
	tax
	jsr $a78d
	ldx $08
	lda $16,x
	cmp #$0d
	beq $a7b9
	cmp #$04
	beq $a7b9
	cmp #$11
	beq $a7b9
	cmp #$05
	beq $a7b9
	cmp #$12
	beq $a7ab
	cmp #$0e
	beq $a7ab
	cmp #$07
	bcs $a7b9
	lda $58,x
	eor #$ff
	tay
	iny
	sty $58,x
	lda $46,x
	eor #$03
	sta $46,x
	rts

	lda #$ff
	sta $03a2,x
	lda $0747
	bne $a7ed
	lda $1e,x
	bmi $a7ed
	lda $16,x
	cmp #$24
	bne $a7d4
	lda $1e,x
	tax
	jsr $a7d4
	jsr $a8b6
	bcs $a7ed
	txa
	jsr $a8c2
	lda $cf,x
	sta $00
	txa
	pha
	jsr $afc3
	pla
	tax
	bcc $a7ed
	jsr $a831
	ldx $08
	rts

	lda $0747
	bne $a82c
	sta $03a2,x
	jsr $a8b6
	bcs $a82c
	lda #$02
	sta $00
	ldx $08
	jsr $a8c0
	and #$02
	bne $a82c
	lda $04ad,y
	cmp #$20
	bcc $a816
	jsr $afc3
	bcs $a82f
	lda $04ad,y
	clc
	adc #$80
	sta $04ad,y
	lda $04af,y
	clc
	adc #$80
	sta $04af,y
	dec $00
	bne $a801
	ldx $08
	rts

	ldx $08
	lda $04af,y
	sec
	sbc $04ad
	cmp #$04
	bcs $a844
	lda $9f
	bpl $a844
	lda #$01
	sta $9f
	lda $04af
	sec
	sbc $04ad,y
	cmp #$06
	bcs $a86a
	lda $9f
	bmi $a86a
	lda $00
	ldy $16,x
	cpy #$2b
	beq $a860
	cpy #$2c
	beq $a860
	txa
	ldx $08
	sta $03a2,x
	lda #$00
	sta $1d
	rts

	lda #$01
	sta $00
	lda $04ae
	sec
	sbc $04ac,y
	cmp #$08
	bcc $a886
	inc $00
	lda $04ae,y
	clc
	sbc $04ac
	cmp #$09
	bcs $a889
	jsr $abd4
	ldx $08
	rts

	.byte $80
	.byte $00
	.byte $a8
	.byte $b5
	.byte $cf
	.byte $18
	.byte $79
	.byte $8b
	.byte $a8
	.byte $2c
	lda $cf,x
	ldy $0e
	cpy #$0b
	beq $a8b5
	ldy $b6,x
	cpy #$01
	bne $a8b5
	sec
	sbc #$20
	sta $ce
	tya
	sbc #$00
	sta $b5
	lda #$00
	sta $9f
	sta $0433
	rts

	lda $03d0
	and #$f0
	clc
	beq $a8bf
	sec
	rts

	lda $08
	asl
	asl
	clc
	adc #$04
	tay
	lda $03d1
	and #$0f
	cmp #$0f
	rts

	.byte $20
	.byte $10
	lda $0716
	bne $a905
	lda $0e
	cmp #$0b
	beq $a905
	cmp #$04
	bcc $a905
	lda #$01
	ldy $0704
	bne $a8f2
	lda $1d
	beq $a8f0
	cmp #$03
	bne $a8f4
	lda #$02
	sta $1d
	lda $b5
	cmp #$01
	bne $a905
	lda #$ff
	sta $0490
	lda $ce
	cmp #$cf
	bcc $a906
	rts

	ldy #$02
	lda $0714
	bne $a919
	lda $0754
	bne $a919
	dey
	lda $0704
	bne $a919
	dey
	lda $b04b,y
	sta $eb
	tay
	ldx $0754
	lda $0714
	beq $a928
	inx
	lda $ce
	cmp $a8d0,x
	bcc $a964
	jsr $b087
	beq $a964
	jsr $ac2a
	bcs $a988
	ldy $9f
	bpl $a964
	ldy $04
	cpy #$04
	bcc $a964
	jsr $ac18
	bcs $a958
	ldy $074e
	beq $a960
	ldy $0784
	bne $a960
	jsr $88ae
	jmp $a964

	cmp #$23
	beq $a960
	lda #$02
	sta $ff
	lda #$01
	sta $9f
	ldy $eb
	lda $ce
	cmp #$cf
	bcs $a9cc
	jsr $b086
	jsr $ac2a
	bcs $a988
	pha
	jsr $b086
	sta $00
	pla
	sta $01
	bne $a98b
	lda $00
	beq $a9cc
	jsr $ac2a
	bcc $a98b
	jmp $aa73

	jsr $ac23
	bcs $a9cc
	ldy $9f
	bmi $a9cc
	cmp #$c6
	bne $a99b
	jmp $aa7c

	jsr $ab40
	beq $a9cc
	ldy $070e
	bne $a9c8
	ldy $04
	cpy #$05
	bcc $a9b2
	lda $45
	sta $00
	jmp $abd4

	jsr $ab4f
	lda #$f0
	and $ce
	sta $ce
	jsr $ab76
	lda #$00
	sta $9f
	sta $0433
	sta $0484
	lda #$00
	sta $1d
	ldy $eb
	iny
	iny
	lda #$02
	sta $00
	iny
	sty $eb
	lda $ce
	cmp #$20
	bcc $a9f3
	cmp #$e4
	bcs $aa09
	jsr $b08a
	beq $a9f3
	cmp #$19
	beq $a9f3
	cmp #$6d
	beq $a9f3
	jsr $ac23
	bcc $aa0a
	ldy $eb
	iny
	lda $ce
	cmp #$08
	bcc $aa09
	cmp #$d0
	bcs $aa09
	jsr $b08a
	bne $aa0a
	dec $00
	bne $a9d4
	rts

	jsr $ab40
	beq $aa70
	jsr $ac23
	bcc $aa17
	jmp $aa9f

	jsr $ac2a
	bcs $aa73
	jsr $ab6b
	bcc $aa29
	lda $070e
	bne $aa70
	jmp $aa6d

	ldy $1d
	cpy #$00
	bne $aa6d
	ldy $33
	dey
	bne $aa6d
	cmp #$6e
	beq $aa3c
	cmp #$1c
	bne $aa6d
	lda $03c4
	bne $aa45
	ldy #$10
	sty $ff
	ora #$20
	sta $03c4
	lda $86
	and #$0f
	beq $aa5e
	ldy #$00
	lda $071a
	beq $aa58
	iny
	lda $aa71,y
	sta $06de
	lda $0e
	cmp #$07
	beq $aa70
	cmp #$08
	bne $aa70
	lda #$02
	sta $0e
	rts

	jsr $abd4
	rts

	.byte $a0
	.byte $34
	jsr $aa8d
	inc $0748
	jmp $87c3

	lda #$00
	sta $0772
	lda #$02
	sta $0770
	jsr $c24a
	lda #$18
	sta $57
	ldy $02
	lda #$00
	sta ($06),y
	jmp $692a

	.byte $f9
	.byte $07
	.byte $ff
	.byte $00
	.byte $18
	.byte $22
	.byte $50
	.byte $68
	.byte $90
	ldy $04
	cpy #$06
	bcc $aaa9
	cpy #$0a
	bcc $aaaa
	rts

	cmp #$21
	beq $aab2
	cmp #$22
	bne $aafd
	lda $0e
	cmp #$05
	beq $ab0b
	lda #$01
	sta $33
	inc $0723
	lda $0e
	cmp #$04
	beq $aaf6
	lda #$33
	jsr $756d
	lda #$80
	sta $fc
	lsr
	sta $0713
	ldx #$04
	lda $ce
	sta $070f
	cmp $aa9a,x
	bcs $aae1
	dex
	bne $aad9
	stx $010f
	lda $07e7
	cmp $07e8
	bne $aaf6
	cmp $07ee
	bne $aaf6
	lda #$05
	sta $010f
	lda #$04
	sta $0e
	jmp $ab0b

	cmp #$23
	bne $ab0b
	lda $ce
	cmp #$20
	bcs $ab0b
	lda #$01
	sta $0e
	lda #$03
	sta $1d
	lda #$00
	sta $57
	sta $0705
	lda $86
	sec
	sbc $071c
	cmp #$10
	bcs $ab24
	lda #$02
	sta $33
	ldy $33
	lda $06
	asl
	asl
	asl
	asl
	clc
	adc $aa95,y
	sta $86
	lda $06
	bne $ab3f
	lda $071b
	clc
	adc $aa97,y
	sta $6d
	rts

	cmp #$5e
	beq $ab4e
	cmp #$5f
	beq $ab4e
	cmp #$60
	beq $ab4e
	cmp #$61
	rts

	jsr $ab6b
	bcc $ab6a
	lda #$70
	sta $0709
	sta $070a
	lda #$f9
	sta $06db
	lda #$03
	sta $0786
	lsr
	sta $070e
	rts

	cmp #$68
	beq $ab74
	cmp #$69
	clc
	bne $ab75
	sec
	rts

	lda $0b
	and #$04
	beq $abd3
	lda $00
	cmp #$11
	bne $abd3
	lda $01
	cmp #$10
	bne $abd3
	lda #$30
	sta $06de
	lda #$03
	sta $0e
	lda #$10
	sta $ff
	lda #$20
	sta $03c4
	lda $06d6
	beq $abd3
	and #$0f
	tax
	lda $670c,x
	ldy $07fb
	beq $abad
	sec
	sbc #$09
	tay
	dey
	sty $075f
	ldx $c357,y
	lda $c360,x
	sta $0750
	lda #$80
	sta $fc
	lda #$00
	sta $0751
	sta $0760
	sta $075c
	sta $0752
	inc $075d
	inc $0757
	rts

	lda #$00
	ldy $57
	ldx $00
	dex
	bne $abe7
	inx
	cpy #$00
	bmi $ac0a
	lda #$ff
	jmp $abef

	ldx #$02
	cpy #$01
	bpl $ac0a
	lda #$01
	ldy #$10
	sty $0785
	ldy #$00
	sty $57
	cmp #$00
	bpl $abfd
	dey
	sty $00
	clc
	adc $86
	sta $86
	lda $6d
	adc $00
	sta $6d
	txa
	eor #$ff
	and $0490
	sta $0490
	rts

	.byte $10
	.byte $62
	.byte $88
	.byte $c5
	jsr $ac39
	cmp $ac14,x
	rts

	.byte $21
	.byte $6f
	.byte $8d
	.byte $c7
	jsr $ac39
	cmp $ac1f,x
	rts

	cmp #$c3
	beq $ac34
	cmp #$c4
	beq $ac34
	clc
	rts

	lda #$01
	sta $fe
	rts

	tay
	and #$c0
	asl
	rol
	rol
	tax
	tya
	rts

	.byte $01
	.byte $01
	.byte $02
	.byte $02
	.byte $02
	.byte $05
	.byte $10
	.byte $f0
	lda $1e,x
	and #$20
	bne $ac41
	jsr $adf1
	bcc $ac41
	ldy $16,x
	cpy #$12
	bne $ac61
	lda $cf,x
	cmp #$25
	bcc $ac41
	cpy #$0e
	bne $ac68
	jmp $adf9

	cpy #$05
	bne $ac70
	jmp $ae1b

	rts

	cpy #$12
	beq $ac80
	cpy #$2e
	beq $ac80
	cpy #$04
	beq $ac6f
	cpy #$07
	bcs $ac6f
	jsr $ae44
	bne $ac88
	jmp $ad78

	jsr $ae4b
	beq $ac85
	cmp #$20
	bne $acfd
	lda $16,x
	cmp #$15
	bcs $aca3
	cmp #$06
	bne $ac9e
	jsr $ae24
	lda #$01
	jsr $a67a
	lda $16,x
	cmp #$09
	bcc $acc1
	cmp #$11
	bcs $acc1
	cmp #$0d
	beq $acc1
	cmp #$04
	beq $acc1
	cmp #$0a
	bcc $acbd
	cmp #$0d
	bcc $acc1
	and #$01
	sta $16,x
	cmp #$2e
	beq $accd
	cmp #$06
	beq $accd
	lda #$02
	sta $1e,x
	dec $cf,x
	dec $cf,x
	lda $16,x
	cmp #$07
	beq $acde
	lda #$fd
	ldy $074e
	bne $ace0
	lda #$ff
	sta $a0,x
	ldy #$01
	jsr $add9
	bpl $acea
	iny
	lda $16,x
	cmp #$33
	beq $acf6
	cmp #$08
	beq $acf6
	sty $46,x
	dey
	lda $ac48,y
	sta $58,x
	rts

	lda $04
	sec
	sbc #$08
	cmp #$05
	bcs $ad78
	lda $1e,x
	and #$40
	bne $ad63
	lda $1e,x
	asl
	bcc $ad14
	jmp $ad94

	lda $1e,x
	beq $ad11
	cmp #$05
	beq $ad3b
	cmp #$03
	bcs $ad3a
	lda $1e,x
	cmp #$02
	bne $ad3b
	lda #$10
	ldy $16,x
	cpy #$12
	bne $ad30
	lda #$00
	sta $0796,x
	lda #$03
	sta $1e,x
	jsr $ade5
	rts

	lda $16,x
	cmp #$06
	beq $ad63
	cmp #$12
	bne $ad53
	lda #$01
	sta $46,x
	lda #$08
	sta $58,x
	lda $09
	and #$07
	beq $ad63
	ldy #$01
	jsr $add9
	bpl $ad5b
	iny
	tya
	cmp $46,x
	bne $ad63
	jsr $adba
	jsr $ade5
	lda $1e,x
	and #$80
	bne $ad71
	lda #$00
	sta $1e,x
	rts

	lda $1e,x
	and #$bf
	sta $1e,x
	rts

	lda $16,x
	cmp #$03
	bne $ad82
	lda $1e,x
	beq $adba
	lda $1e,x
	tay
	asl
	bcc $ad8f
	lda $1e,x
	ora #$40
	jmp $ad92

	lda $ac42,y
	sta $1e,x
	lda $cf,x
	cmp #$20
	bcc $adb9
	ldy #$16
	lda #$02
	sta $eb
	lda $eb
	cmp $46,x
	bne $adb2
	lda #$01
	jsr $b026
	beq $adb2
	jsr $ae4b
	bne $adba
	dec $eb
	iny
	cpy #$18
	bcc $ada0
	rts

	cpx #$05
	beq $adc7
	lda $1e,x
	asl
	bcc $adc7
	lda #$02
	sta $ff
	lda $16,x
	cmp #$05
	bne $add6
	lda #$00
	sta $00
	ldy #$fa
	jmp $966c

	jmp $a7ab

	lda $87,x
	sec
	sbc $86
	sta $00
	lda $6e,x
	sbc $6d
	rts

	jsr $8f4e
	lda $cf,x
	and #$f0
	ora #$08
	sta $cf,x
	rts

	lda $cf,x
	clc
	adc #$3e
	cmp #$44
	rts

	jsr $adf1
	bcc $ae18
	lda $a0,x
	clc
	adc #$02
	cmp #$03
	bcc $ae18
	jsr $ae44
	beq $ae18
	jsr $ae4b
	beq $ae18
	jsr $ade5
	lda #$fd
	sta $a0,x
	jmp $ad94

	jsr $ae44
	beq $ae3d
	cmp #$20
	bne $ae2c
	jsr $a3d6
	lda #$fc
	sta $a0,x
	rts

	lda $078a,x
	bne $ae3d
	lda $1e,x
	and #$88
	sta $1e,x
	jsr $ade5
	jmp $ad94

	lda $1e,x
	ora #$01
	sta $1e,x
	rts

	lda #$00
	ldy #$15
	jmp $b026

	cmp #$23
	beq $ae65
	cmp #$c3
	beq $ae65
	cmp #$c4
	beq $ae65
	cmp #$5e
	beq $ae65
	cmp #$5f
	beq $ae65
	cmp #$60
	beq $ae65
	cmp #$61
	rts

	lda $d5,x
	cmp #$18
	bcc $ae8d
	jsr $b03a
	beq $ae8d
	jsr $ae4b
	beq $ae8d
	lda $a6,x
	bmi $ae92
	lda $3a,x
	bne $ae92
	lda #$fd
	sta $a6,x
	lda #$01
	sta $3a,x
	lda $d5,x
	and #$f8
	sta $d5,x
	rts

	lda #$00
	sta $3a,x
	rts

	lda #$80
	sta $24,x
	lda #$02
	sta $ff
	rts

	.byte $02
	.byte $08
	.byte $0e
	.byte $20
	.byte $03
	.byte $14
	.byte $0d
	.byte $20
	.byte $02
	.byte $14
	.byte $0e
	.byte $20
	.byte $02
	.byte $09
	.byte $0e
	.byte $15
	.byte $00
	.byte $00
	.byte $18
	.byte $06
	.byte $00
	.byte $00
	.byte $20
	.byte $0d
	.byte $00
	.byte $00
	.byte $30
	.byte $0d
	.byte $00
	.byte $00
	.byte $08
	.byte $08
	.byte $06
	.byte $04
	.byte $0a
	.byte $08
	.byte $03
	.byte $0e
	.byte $0d
	.byte $16
	.byte $00
	.byte $02
	.byte $10
	.byte $15
	.byte $04
	.byte $04
	.byte $0c
	.byte $1c
	txa
	clc
	adc #$07
	tax
	ldy #$02
	bne $aedb
	txa
	clc
	adc #$09
	tax
	ldy #$06
	jsr $af3a
	jmp $af7c

	ldy #$48
	sty $00
	ldy #$44
	jmp $aef0

	ldy #$08
	sty $00
	ldy #$04
	lda $87,x
	sec
	sbc $071c
	sta $01
	lda $6e,x
	sbc $071a
	bmi $af05
	ora $01
	beq $af05
	ldy $00
	tya
	and $03d1
	sta $03d8,x
	bne $af27
	jmp $af1a

	inx
	jsr $bedb
	dex
	cmp #$fe
	bcs $af27
	txa
	clc
	adc #$01
	tax
	ldy #$01
	jsr $af3a
	jmp $af7c

	txa
	asl
	asl
	tay
	lda #$ff
	sta $04b0,y
	sta $04b1,y
	sta $04b2,y
	sta $04b3,y
	rts

	stx $00
	lda $03b8,y
	sta $02
	lda $03ad,y
	sta $01
	txa
	asl
	asl
	pha
	tay
	lda $0499,x
	asl
	asl
	tax
	lda $01
	clc
	adc $ae9b,x
	sta $04ac,y
	lda $01
	clc
	adc $ae9d,x
	sta $04ae,y
	inx
	iny
	lda $02
	clc
	adc $ae9b,x
	sta $04ac,y
	lda $02
	clc
	adc $ae9d,x
	sta $04ae,y
	pla
	tay
	ldx $00
	rts

	lda $071c
	clc
	adc #$80
	sta $02
	lda $071a
	adc #$00
	sta $01
	lda $86,x
	cmp $02
	lda $6d,x
	sbc $01
	bcc $afaa
	lda $04ae,y
	bmi $afa7
	lda #$ff
	ldx $04ac,y
	bmi $afa4
	sta $04ac,y
	sta $04ae,y
	ldx $08
	rts

	lda $04ac,y
	bpl $afc0
	cmp #$a0
	bcc $afc0
	lda #$00
	ldx $04ae,y
	bpl $afbd
	sta $04ae,y
	sta $04ac,y
	ldx $08
	rts

	ldx #$00
	sty $06
	lda #$01
	sta $07
	lda $04ac,y
	cmp $04ac,x
	bcs $affd
	cmp $04ae,x
	bcc $afea
	beq $b01c
	lda $04ae,y
	cmp $04ac,y
	bcc $b01c
	cmp $04ac,x
	bcs $b01c
	ldy $06
	rts

	lda $04ae,x
	cmp $04ac,x
	bcc $b01c
	lda $04ae,y
	cmp $04ac,x
	bcs $b01c
	ldy $06
	rts

	cmp $04ac,x
	beq $b01c
	cmp $04ae,x
	bcc $b01c
	beq $b01c
	cmp $04ae,y
	bcc $b018
	beq $b018
	lda $04ae,y
	cmp $04ac,x
	bcs $b01c
	clc
	ldy $06
	rts

	inx
	iny
	dec $07
	bpl $afcb
	sec
	ldy $06
	rts

	pha
	txa
	clc
	adc #$01
	tax
	pla
	jmp $b043

	txa
	clc
	adc #$0d
	tax
	ldy #$1b
	jmp $b041

	ldy #$1a
	txa
	clc
	adc #$07
	tax
	lda #$00
	jsr $b08e
	ldx $08
	cmp #$00
	rts

	.byte $00
	.byte $07
	.byte $0e
	.byte $08
	.byte $03
	.byte $0c
	.byte $02
	.byte $02
	.byte $0d
	.byte $0d
	.byte $08
	.byte $03
	.byte $0c
	.byte $02
	.byte $02
	.byte $0d
	.byte $0d
	.byte $08
	.byte $03
	.byte $0c
	.byte $02
	.byte $02
	.byte $0d
	.byte $0d
	.byte $08
	.byte $00
	.byte $10
	.byte $04
	.byte $14
	.byte $04
	.byte $04
	.byte $04
	.byte $20
	.byte $20
	.byte $08
	.byte $18
	.byte $08
	.byte $18
	.byte $02
	.byte $20
	.byte $20
	.byte $08
	.byte $18
	.byte $08
	.byte $18
	.byte $12
	.byte $20
	.byte $20
	.byte $18
	.byte $18
	.byte $18
	.byte $18
	.byte $18
	.byte $14
	.byte $14
	.byte $06
	.byte $06
	.byte $08
	.byte $10
	iny
	lda #$00
	bit $01a9
	ldx #$00
	pha
	sty $04
	lda $b04e,y
	clc
	adc $86,x
	sta $05
	lda $6d,x
	adc #$00
	and #$01
	lsr
	ora $05
	ror
	lsr
	lsr
	lsr
	jsr $7a22
	ldy $04
	lda $ce,x
	clc
	adc $b06a,y
	and #$f0
	sec
	sbc #$20
	sta $02
	tay
	lda ($06),y
	sta $03
	ldy $04
	pla
	bne $b0c7
	lda $ce,x
	jmp $b0c9

	lda $86,x
	and #$0f
	sta $04
	lda $03
	rts

	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $ff
	.byte $00
	.byte $30
	sty $00
	lda $03b9
	clc
	adc $b0d7,y
	ldx $039a,y
	ldy $06e5,x
	sty $02
	jsr $b152
	lda $03ae
	sta $0203,y
	sta $020b,y
	sta $0213,y
	clc
	adc #$06
	sta $0207,y
	sta $020f,y
	sta $0217,y
	lda #$21
	sta $0202,y
	sta $020a,y
	sta $0212,y
	ora #$40
	sta $0206,y
	sta $020e,y
	sta $0216,y
	ldx #$05
	lda #$e1
	sta $0201,y
	iny
	iny
	iny
	iny
	dex
	bpl $b11d
	ldy $02
	lda $00
	bne $b134
	lda #$e0
	sta $0201,y
	ldx #$00
	lda $039d
	sec
	sbc $0200,y
	cmp #$64
	bcc $b146
	lda #$f8
	sta $0200,y
	iny
	iny
	iny
	iny
	inx
	cpx #$06
	bne $b136
	ldy $00
	rts

	ldx #$06
	sta $0200,y
	clc
	adc #$08
	iny
	iny
	iny
	iny
	dex
	bne $b154
	ldy $02
	rts

	.byte $04
	.byte $00
	.byte $04
	.byte $00
	.byte $00
	.byte $04
	.byte $00
	.byte $04
	.byte $00
	.byte $08
	.byte $00
	.byte $08
	.byte $08
	.byte $00
	.byte $08
	.byte $00
	.byte $80
	.byte $82
	.byte $81
	.byte $83
	.byte $81
	.byte $83
	.byte $80
	.byte $82
	.byte $03
	.byte $03
	.byte $c3
	.byte $c3
	ldy $06f3,x
	lda $0747
	bne $b190
	lda $2a,x
	and #$7f
	cmp #$01
	beq $b194
	ldx #$00
	beq $b19b
	lda $09
	lsr
	lsr
	and #$03
	tax
	lda $03be
	clc
	adc $b168,x
	sta $0200,y
	clc
	adc $b170,x
	sta $0204,y
	lda $03b3
	clc
	adc $b164,x
	sta $0203,y
	clc
	adc $b16c,x
	sta $0207,y
	lda $b174,x
	sta $0201,y
	lda $b178,x
	sta $0205,y
	lda $b17c,x
	sta $0202,y
	sta $0206,y
	ldx $08
	lda $03d6
	and #$fc
	beq $b1e4
	lda #$00
	sta $2a,x
	lda #$f8
	jsr $b267
	rts

	.byte $f9
	.byte $50
	.byte $f7
	.byte $50
	.byte $fa
	.byte $fb
	.byte $f8
	.byte $fb
	.byte $f6
	.byte $fb
	.byte $fd
	.byte $fe
	ldy $06e5,x
	lda $03ae
	sta $0203,y
	clc
	adc #$08
	sta $0207,y
	sta $020b,y
	clc
	adc #$0c
	sta $05
	lda $cf,x
	jsr $b267
	adc #$08
	sta $0208,y
	lda $010d
	sta $02
	lda #$01
	sta $03
	sta $04
	sta $0202,y
	sta $0206,y
	sta $020a,y
	lda #$7e
	sta $0201,y
	sta $0209,y
	lda #$7f
	sta $0205,y
	lda $070f
	beq $b24d
	tya
	clc
	adc #$0c
	tay
	lda $010f
	asl
	tax
	lda $b1e5,x
	sta $00
	lda $b1e6,x
	jsr $b88d
	ldx $08
	ldy $06e5,x
	lda $03d1
	and #$0e
	beq $b26d
	lda #$f8
	sta $0214,y
	sta $0210,y
	sta $020c,y
	sta $0208,y
	sta $0204,y
	sta $0200,y
	rts

	ldy $06e5,x
	sty $02
	iny
	iny
	iny
	lda $03ae
	jsr $b152
	ldx $08
	lda $cf,x
	jsr $b261
	ldy $074e
	cpy #$03
	beq $b28f
	ldy $06cc
	beq $b291
	lda #$f8
	ldy $06e5,x
	sta $0210,y
	sta $0214,y
	lda #$5b
	ldx $0743
	beq $b2a3
	lda #$75
	ldx $08
	iny
	jsr $b25b
	lda #$02
	iny
	jsr $b25b
	inx
	jsr $bedb
	dex
	ldy $06e5,x
	asl
	pha
	bcc $b2c0
	lda #$f8
	sta $0200,y
	pla
	asl
	pha
	bcc $b2ca
	lda #$f8
	sta $0204,y
	pla
	asl
	pha
	bcc $b2d4
	lda #$f8
	sta $0208,y
	pla
	asl
	pha
	bcc $b2de
	lda #$f8
	sta $020c,y
	pla
	asl
	pha
	bcc $b2e8
	lda #$f8
	sta $0210,y
	pla
	asl
	bcc $b2f1
	lda #$f8
	sta $0214,y
	lda $03d1
	asl
	bcc $b2fa
	jsr $b259
	rts

	lda $09
	lsr
	bcs $b302
	dec $db,x
	lda $db,x
	jsr $b267
	lda $03b3
	sta $0203,y
	clc
	adc #$08
	sta $0207,y
	lda #$02
	sta $0202,y
	sta $0206,y
	lda #$f7
	sta $0201,y
	lda #$fb
	sta $0205,y
	jmp $b363

	.byte $60
	.byte $61
	.byte $62
	.byte $63
	ldy $06f3,x
	lda $2a,x
	cmp #$02
	bcs $b2fb
	lda $db,x
	sta $0200,y
	clc
	adc #$08
	sta $0204,y
	lda $03b3
	sta $0203,y
	sta $0207,y
	lda $09
	lsr
	and #$03
	tax
	lda $b328,x
	iny
	jsr $b267
	dey
	lda #$02
	sta $0202,y
	lda #$82
	sta $0206,y
	ldx $08
	rts

	.byte $d8
	.byte $da
	.byte $db
	.byte $ff
	.byte $d6
	.byte $d6
	.byte $d9
	.byte $d9
	.byte $8d
	.byte $8d
	.byte $e4
	.byte $e4
	.byte $d8
	.byte $da
	.byte $db
	.byte $ff
	.byte $d8
	.byte $da
	.byte $db
	.byte $ff
	.byte $02
	.byte $01
	.byte $02
	.byte $01
	.byte $03
	ldy $06ea
	lda $03b9
	clc
	adc #$08
	sta $02
	lda $03ae
	sta $05
	ldx $39
	lda $b378,x
	ora $03ca
	sta $04
	txa
	pha
	asl
	asl
	tax
	lda #$01
	sta $07
	sta $03
	lda $b364,x
	sta $00
	lda $b365,x
	jsr $b88d
	dec $07
	bpl $b3a2
	ldy $06ea
	pla
	beq $b3ea
	cmp #$03
	beq $b3ea
	cmp #$04
	beq $b3ea
	sta $00
	lda $09
	lsr
	and #$03
	ora $03ca
	sta $0202,y
	sta $0206,y
	ldx $00
	dex
	beq $b3da
	sta $020a,y
	sta $020e,y
	lda $0206,y
	ora #$40
	sta $0206,y
	lda $020e,y
	ora #$40
	sta $020e,y
	jmp $b83f

	.byte $fc
	.byte $fc
	.byte $aa
	.byte $ab
	.byte $ac
	.byte $ad
	.byte $fc
	.byte $fc
	.byte $ae
	.byte $af
	.byte $b0
	.byte $b1
	.byte $fc
	.byte $a5
	.byte $a6
	.byte $a7
	.byte $a8
	.byte $a9
	.byte $fc
	.byte $a0
	.byte $a1
	.byte $a2
	.byte $a3
	.byte $a4
	.byte $69
	.byte $a5
	.byte $6a
	.byte $a7
	.byte $a8
	.byte $a9
	.byte $6b
	.byte $a0
	.byte $6c
	.byte $a2
	.byte $a3
	.byte $a4
	.byte $fc
	.byte $fc
	.byte $96
	.byte $97
	.byte $98
	.byte $99
	.byte $fc
	.byte $fc
	.byte $9a
	.byte $9b
	.byte $9c
	.byte $9d
	.byte $fc
	.byte $fc
	.byte $8f
	.byte $8e
	.byte $8e
	.byte $8f
	.byte $fc
	.byte $fc
	.byte $95
	.byte $94
	.byte $94
	.byte $95
	.byte $fc
	.byte $fc
	.byte $dc
	.byte $dc
	.byte $df
	.byte $df
	.byte $dc
	.byte $dc
	.byte $dd
	.byte $dd
	.byte $de
	.byte $de
	.byte $fc
	.byte $fc
	.byte $b2
	.byte $b3
	.byte $b4
	.byte $b5
	.byte $fc
	.byte $fc
	.byte $b6
	.byte $b3
	.byte $b7
	.byte $b5
	.byte $fc
	.byte $fc
	.byte $70
	.byte $71
	.byte $72
	.byte $73
	.byte $fc
	.byte $fc
	.byte $6e
	.byte $6e
	.byte $6f
	.byte $6f
	.byte $fc
	.byte $fc
	.byte $6d
	.byte $6d
	.byte $6f
	.byte $6f
	.byte $fc
	.byte $fc
	.byte $6f
	.byte $6f
	.byte $6e
	.byte $6e
	.byte $fc
	.byte $fc
	.byte $6f
	.byte $6f
	.byte $6d
	.byte $6d
	.byte $fc
	.byte $fc
	.byte $f4
	.byte $f4
	.byte $f5
	.byte $f5
	.byte $fc
	.byte $fc
	.byte $f4
	.byte $f4
	.byte $f5
	.byte $f5
	.byte $fc
	.byte $fc
	.byte $f5
	.byte $f5
	.byte $f4
	.byte $f4
	.byte $fc
	.byte $fc
	.byte $f5
	.byte $f5
	.byte $f4
	.byte $f4
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $ef
	.byte $ef
	.byte $b9
	.byte $b8
	.byte $bb
	.byte $ba
	.byte $bc
	.byte $bc
	.byte $fc
	.byte $fc
	.byte $bd
	.byte $bd
	.byte $bc
	.byte $bc
	.byte $76
	.byte $79
	.byte $77
	.byte $77
	.byte $78
	.byte $78
	.byte $cd
	.byte $cd
	.byte $ce
	.byte $ce
	.byte $cf
	.byte $cf
	.byte $7d
	.byte $7c
	.byte $d1
	.byte $8c
	.byte $d3
	.byte $d2
	.byte $7d
	.byte $7c
	.byte $89
	.byte $88
	.byte $8b
	.byte $8a
	.byte $d5
	.byte $d4
	.byte $e3
	.byte $e2
	.byte $d3
	.byte $d2
	.byte $d5
	.byte $d4
	.byte $e3
	.byte $e2
	.byte $8b
	.byte $8a
	.byte $e5
	.byte $e5
	.byte $e6
	.byte $e6
	.byte $eb
	.byte $eb
	.byte $ec
	.byte $ec
	.byte $ed
	.byte $ed
	.byte $ee
	.byte $ee
	.byte $fc
	.byte $fc
	.byte $d0
	.byte $d0
	.byte $d7
	.byte $d7
	.byte $bf
	.byte $be
	.byte $c1
	.byte $c0
	.byte $c2
	.byte $fc
	.byte $c4
	.byte $c3
	.byte $c6
	.byte $c5
	.byte $c8
	.byte $c7
	.byte $bf
	.byte $be
	.byte $ca
	.byte $c9
	.byte $c2
	.byte $fc
	.byte $c4
	.byte $c3
	.byte $c6
	.byte $c5
	.byte $cc
	.byte $cb
	.byte $fc
	.byte $fc
	.byte $e8
	.byte $e7
	.byte $ea
	.byte $e9
	.byte $f2
	.byte $f2
	.byte $f3
	.byte $f3
	.byte $f2
	.byte $f2
	.byte $f1
	.byte $f1
	.byte $f1
	.byte $f1
	.byte $fc
	.byte $fc
	.byte $f0
	.byte $f0
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $0c
	.byte $0c
	.byte $00
	.byte $0c
	.byte $c0
	.byte $a8
	.byte $54
	.byte $3c
	.byte $ea
	.byte $18
	.byte $48
	.byte $48
	.byte $cc
	.byte $c0
	.byte $18
	.byte $18
	.byte $18
	.byte $90
	.byte $24
	.byte $ff
	.byte $48
	.byte $9c
	.byte $d2
	.byte $d8
	.byte $f0
	.byte $f6
	.byte $fc
	.byte $01
	.byte $02
	.byte $03
	.byte $02
	.byte $22
	.byte $01
	.byte $03
	.byte $03
	.byte $03
	.byte $01
	.byte $01
	.byte $02
	.byte $02
	.byte $21
	.byte $01
	.byte $02
	.byte $01
	.byte $01
	.byte $02
	.byte $ff
	.byte $02
	.byte $02
	.byte $01
	.byte $01
	.byte $02
	.byte $02
	.byte $02
	.byte $08
	.byte $18
	.byte $18
	.byte $19
	.byte $1a
	.byte $19
	.byte $18
	lda #$02
	ldy $075f
	cpy #$01
	beq $b53d
	cpy #$02
	beq $b53d
	cpy #$06
	bne $b53e
	lsr
	sta $b522
	sta $b523
	sta $b524
	lda $cf,x
	sta $02
	lda $03ae
	sta $05
	ldy $06e5,x
	sty $eb
	lda #$00
	sta $0109
	lda $46,x
	sta $03
	lda $03c5,x
	sta $04
	lda $16,x
	cmp #$0d
	bne $b573
	ldy $58,x
	bmi $b573
	ldy $078a,x
	beq $b573
	rts

	lda $1e,x
	sta $ed
	and #$1f
	tay
	lda $16,x
	cmp #$35
	bne $b588
	ldy #$00
	lda #$01
	sta $03
	lda #$15
	cmp #$33
	bne $b59f
	dec $02
	lda #$03
	ldy $078a,x
	beq $b597
	ora #$20
	sta $04
	ldy #$00
	sty $ed
	lda #$08
	cmp #$32
	bne $b5ab
	ldy #$03
	ldx $070e
	lda $b527,x
	sta $ef
	sty $ec
	ldx $08
	cmp #$0c
	bne $b5bc
	lda $a0,x
	bmi $b5bc
	inc $0109
	lda $036a
	beq $b5ca
	ldy #$16
	cmp #$01
	beq $b5c8
	iny
	sty $ef
	ldy $ef
	cpy #$06
	bne $b5ed
	lda $1e,x
	cmp #$02
	bcc $b5da
	ldx #$04
	stx $ec
	and #$20
	ora $0747
	bne $b5ed
	lda $09
	and #$08
	bne $b5ed
	lda $03
	eor #$03
	sta $03
	lda $b50a,y
	ora $04
	sta $04
	lda $b4ef,y
	tax
	ldy $ec
	lda $036a
	beq $b62f
	cmp #$01
	bne $b616
	lda $0363
	bpl $b60a
	ldx #$de
	lda $ed
	and #$20
	beq $b613
	stx $0109
	jmp $b71b

	lda $0363
	and #$01
	beq $b61f
	ldx #$e4
	lda $ed
	and #$20
	beq $b613
	lda $02
	sec
	sbc #$10
	sta $02
	jmp $b610

	cpx #$24
	bne $b644
	cpy #$05
	bne $b641
	ldx #$30
	lda #$02
	sta $03
	lda #$05
	sta $ec
	jmp $b694

	cpx #$90
	bne $b65a
	lda $ed
	and #$20
	bne $b657
	lda $078f
	cmp #$10
	bcs $b657
	ldx #$96
	jmp $b701

	lda $ef
	cmp #$04
	bcs $b670
	cpy #$02
	bcc $b670
	ldx #$5a
	ldy $ef
	cpy #$02
	bne $b670
	ldx #$7e
	inc $02
	lda $ec
	cmp #$04
	bne $b694
	ldx #$72
	inc $02
	ldy $ef
	cpy #$02
	beq $b684
	ldx #$66
	inc $02
	cpy #$06
	bne $b694
	ldx #$54
	lda $ed
	and #$20
	bne $b694
	ldx #$8a
	dec $02
	ldy $08
	lda $ef
	cmp #$05
	bne $b6a8
	lda $ed
	beq $b6c4
	and #$08
	beq $b701
	ldx #$b4
	bne $b6c4
	cpx #$48
	beq $b6c4
	lda $0796,y
	cmp #$05
	bcs $b701
	cpx #$3c
	bne $b6c4
	cmp #$01
	beq $b701
	inc $02
	inc $02
	inc $02
	jmp $b6f3

	lda $ef
	cmp #$06
	beq $b701
	cmp #$08
	beq $b701
	cmp #$0c
	beq $b701
	cmp #$18
	bcs $b701
	ldy #$00
	cmp #$15
	bne $b6ec
	iny
	lda #$03
	sta $ec
	lda $075f
	cmp #$07
	bcs $b701
	ldx #$a2
	bne $b701
	lda $09
	and $b525,y
	bne $b701
	lda $ed
	and #$a0
	ora $0747
	bne $b701
	txa
	clc
	adc #$06
	tax
	lda $ef
	cmp #$04
	beq $b713
	lda $ed
	and #$20
	beq $b71b
	lda $ef
	cmp #$04
	bcc $b71b
	ldy #$01
	sty $0109
	dey
	sty $ec
	ldy $eb
	jsr $b885
	jsr $b885
	jsr $b885
	ldx $08
	ldy $06e5,x
	lda $ef
	cmp #$08
	bne $b734
	jmp $b83f

	lda $0109
	beq $b77a
	lda $0202,y
	ora #$80
	iny
	iny
	jsr $b25b
	dey
	dey
	tya
	tax
	lda $ef
	cmp #$05
	beq $b75e
	cmp #$04
	beq $b75e
	cmp #$11
	beq $b75e
	cmp #$15
	bcs $b75e
	txa
	clc
	adc #$08
	tax
	lda $0201,x
	pha
	lda $0205,x
	pha
	lda $0211,y
	sta $0201,x
	lda $0215,y
	sta $0205,x
	pla
	sta $0215,y
	pla
	sta $0211,y
	lda $036a
	bne $b731
	lda $ef
	ldx $ec
	cmp #$05
	bne $b78a
	jmp $b83f

	cmp #$07
	beq $b7af
	cmp #$0d
	beq $b7af
	cmp #$04
	beq $b7af
	cmp #$0c
	beq $b7af
	cmp #$12
	bne $b7a2
	cpx #$05
	bne $b7ea
	cmp #$15
	bne $b7ab
	lda #$42
	sta $0216,y
	cpx #$02
	bcc $b7ea
	lda $036a
	bne $b7ea
	lda $0202,y
	and #$a3
	sta $0202,y
	sta $020a,y
	sta $0212,y
	ora #$40
	cpx #$05
	bne $b7ca
	ora #$80
	sta $0206,y
	sta $020e,y
	sta $0216,y
	cpx #$04
	bne $b7ea
	lda $020a,y
	ora #$80
	sta $020a,y
	sta $0212,y
	ora #$40
	sta $020e,y
	sta $0216,y
	lda $ef
	cmp #$11
	bne $b826
	lda $0109
	bne $b816
	lda $0212,y
	and #$81
	sta $0212,y
	lda $0216,y
	ora #$41
	sta $0216,y
	ldx $078f
	cpx #$10
	bcs $b83f
	sta $020e,y
	and #$81
	sta $020a,y
	bcc $b83f
	lda $0202,y
	and #$81
	sta $0202,y
	lda $0206,y
	ora #$41
	sta $0206,y
	lda $ef
	cmp #$18
	bcc $b83f
	lda #$80
	ora $b522
	sta $020a,y
	sta $0212,y
	ora #$40
	sta $020e,y
	sta $0216,y
	ldx $08
	lda $03d1
	lsr
	lsr
	lsr
	pha
	bcc $b84f
	lda #$04
	jsr $b89c
	pla
	lsr
	pha
	bcc $b859
	lda #$00
	jsr $b89c
	pla
	lsr
	lsr
	pha
	bcc $b864
	lda #$10
	jsr $b892
	pla
	lsr
	pha
	bcc $b86e
	lda #$08
	jsr $b892
	pla
	lsr
	bcc $b884
	jsr $b892
	lda $16,x
	cmp #$0c
	beq $b884
	lda $b6,x
	cmp #$02
	bne $b884
	jsr $95cd
	rts

	lda $b3ed,x
	sta $00
	lda $b3ee,x
	sta $01
	jmp $bf67

	clc
	adc $06e5,x
	tay
	lda #$f8
	jmp $b267

	clc
	adc $06e5,x
	tay
	jsr $b925
	sta $0210,y
	rts

	.byte $85
	.byte $85
	.byte $86
	.byte $86
	lda $03bc
	sta $02
	lda $03b1
	sta $05
	lda #$03
	sta $04
	lsr
	sta $03
	ldy $06ec,x
	ldx #$00
	lda $b8a8,x
	sta $00
	lda $b8a9,x
	jsr $b88d
	cpx #$04
	bne $b8c2
	ldx $08
	ldy $06ec,x
	lda $074e
	cmp #$01
	beq $b8e5
	lda #$86
	sta $0201,y
	sta $0205,y
	lda $03e8,x
	cmp #$c5
	bne $b910
	lda #$87
	iny
	jsr $b261
	dey
	lda #$03
	ldx $074e
	dex
	beq $b8fc
	lsr
	ldx $08
	sta $0202,y
	ora #$40
	sta $0206,y
	ora #$80
	sta $020e,y
	and #$83
	sta $020a,y
	lda $03d4
	pha
	and #$04
	beq $b920
	lda #$f8
	sta $0204,y
	sta $020c,y
	pla
	and #$08
	beq $b92d
	lda #$f8
	sta $0200,y
	sta $0208,y
	rts

	lda #$02
	sta $00
	lda #$75
	ldy $0e
	cpy #$05
	beq $b940
	lda #$03
	sta $00
	lda #$84
	ldy $06ec,x
	iny
	jsr $b261
	lda $09
	asl
	asl
	asl
	asl
	and #$c0
	ora $00
	iny
	jsr $b261
	dey
	dey
	lda $03bc
	jsr $b267
	lda $03b1
	sta $0203,y
	lda $03f1,x
	sec
	sbc $071c
	sta $00
	sec
	sbc $03b1
	adc $00
	adc #$06
	sta $0207,y
	lda $03bd
	sta $0208,y
	sta $020c,y
	lda $03b2
	sta $020b,y
	lda $00
	sec
	sbc $03b2
	adc $00
	adc #$06
	sta $020f,y
	lda $03d4
	jsr $b921
	lda $03d4
	asl
	bcc $b9a4
	lda #$f8
	jsr $b267
	lda $00
	bpl $b9b8
	lda $0203,y
	cmp $0207,y
	bcc $b9b8
	lda #$f8
	sta $0204,y
	sta $020c,y
	rts

	ldy $06f1,x
	lda $03ba
	sta $0200,y
	lda $03af
	sta $0203,y
	lda $09
	lsr
	lsr
	pha
	and #$01
	eor #$64
	sta $0201,y
	pla
	lsr
	lsr
	lda #$02
	bcc $b9dd
	ora #$c0
	sta $0202,y
	rts

	.byte $68
	.byte $67
	.byte $66
	ldy $06ec,x
	lda $24,x
	inc $24,x
	lsr
	and #$07
	cmp #$03
	bcs $ba3c
	tax
	lda $b9e1,x
	iny
	jsr $b261
	dey
	ldx $08
	lda $03ba
	sec
	sbc #$04
	sta $0200,y
	sta $0208,y
	clc
	adc #$08
	sta $0204,y
	sta $020c,y
	lda $03af
	sec
	sbc #$04
	sta $0203,y
	sta $0207,y
	clc
	adc #$08
	sta $020b,y
	sta $020f,y
	lda #$02
	sta $0202,y
	lda #$82
	sta $0206,y
	lda #$42
	sta $020a,y
	lda #$c2
	sta $020e,y
	rts

	lda #$00
	sta $24,x
	rts

	ldy $06e5,x
	lda #$5b
	iny
	jsr $b25b
	iny
	lda #$02
	jsr $b25b
	dey
	dey
	lda $03ae
	sta $0203,y
	sta $020f,y
	clc
	adc #$08
	sta $0207,y
	sta $0213,y
	clc
	adc #$08
	sta $020b,y
	sta $0217,y
	lda $cf,x
	tax
	pha
	cpx #$20
	bcs $ba77
	lda #$f8
	jsr $b264
	pla
	clc
	adc #$80
	tax
	cpx #$20
	bcs $ba85
	lda #$f8
	sta $020c,y
	sta $0210,y
	sta $0214,y
	lda $03d1
	pha
	and #$08
	beq $ba9e
	lda #$f8
	sta $0200,y
	sta $020c,y
	pla
	pha
	and #$04
	beq $baac
	lda #$f8
	sta $0204,y
	sta $0210,y
	pla
	and #$02
	beq $bab9
	lda #$f8
	sta $0208,y
	sta $0214,y
	ldx $08
	rts

	ldy $b5
	dey
	bne $bae1
	lda $03d3
	and #$08
	bne $bae1
	ldy $06ee,x
	lda $03b0
	sta $0203,y
	lda $03bb
	sta $0200,y
	lda #$74
	sta $0201,y
	lda #$02
	sta $0202,y
	rts

	.byte $20
	.byte $28
	.byte $c8
	.byte $18
	.byte $00
	.byte $40
	.byte $50
	.byte $58
	.byte $80
	.byte $88
	.byte $b8
	.byte $78
	.byte $60
	.byte $a0
	.byte $b0
	.byte $b8
	.byte $00
	.byte $01
	.byte $02
	.byte $03
	.byte $04
	.byte $05
	.byte $06
	.byte $07
	.byte $08
	.byte $09
	.byte $0a
	.byte $0b
	.byte $0c
	.byte $0d
	.byte $0e
	.byte $0f
	.byte $10
	.byte $11
	.byte $12
	.byte $13
	.byte $14
	.byte $15
	.byte $16
	.byte $17
	.byte $18
	.byte $19
	.byte $1a
	.byte $1b
	.byte $1c
	.byte $1d
	.byte $1e
	.byte $1f
	.byte $20
	.byte $21
	.byte $22
	.byte $23
	.byte $24
	.byte $25
	.byte $26
	.byte $27
	.byte $08
	.byte $09
	.byte $28
	.byte $29
	.byte $2a
	.byte $2b
	.byte $2c
	.byte $2d
	.byte $08
	.byte $09
	.byte $0a
	.byte $0b
	.byte $0c
	.byte $30
	.byte $2c
	.byte $2d
	.byte $08
	.byte $09
	.byte $0a
	.byte $0b
	.byte $2e
	.byte $2f
	.byte $2c
	.byte $2d
	.byte $08
	.byte $09
	.byte $28
	.byte $29
	.byte $2a
	.byte $2b
	.byte $5c
	.byte $5d
	.byte $08
	.byte $09
	.byte $0a
	.byte $0b
	.byte $0c
	.byte $0d
	.byte $5e
	.byte $5f
	.byte $fc
	.byte $fc
	.byte $08
	.byte $09
	.byte $58
	.byte $59
	.byte $5a
	.byte $5a
	.byte $08
	.byte $09
	.byte $28
	.byte $29
	.byte $2a
	.byte $2b
	.byte $0e
	.byte $0f
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $32
	.byte $33
	.byte $34
	.byte $35
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $36
	.byte $37
	.byte $38
	.byte $39
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $3a
	.byte $37
	.byte $3b
	.byte $3c
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $3d
	.byte $3e
	.byte $3f
	.byte $40
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $32
	.byte $41
	.byte $42
	.byte $43
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $32
	.byte $33
	.byte $44
	.byte $45
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $32
	.byte $33
	.byte $44
	.byte $47
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $32
	.byte $33
	.byte $48
	.byte $49
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $32
	.byte $33
	.byte $90
	.byte $91
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $3a
	.byte $37
	.byte $92
	.byte $93
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $9e
	.byte $9e
	.byte $9f
	.byte $9f
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $fc
	.byte $3a
	.byte $37
	.byte $4f
	.byte $4f
	.byte $fc
	.byte $fc
	.byte $00
	.byte $01
	.byte $4c
	.byte $4d
	.byte $4e
	.byte $4e
	.byte $00
	.byte $01
	.byte $4c
	.byte $4d
	.byte $4a
	.byte $4a
	.byte $4b
	.byte $4b
	.byte $31
	.byte $46
	lda $079e
	beq $bbce
	lda $09
	lsr
	bcs $bc0e
	lda $0e
	cmp #$0b
	beq $bc1b
	lda $070b
	bne $bc15
	ldy $0704
	beq $bc0f
	lda $1d
	cmp #$00
	beq $bc0f
	jsr $bc0f
	lda $09
	and #$04
	bne $bc0e
	tax
	ldy $06e4
	lda $33
	lsr
	bcs $bbfa
	iny
	iny
	iny
	iny
	lda $0754
	beq $bc08
	lda $0219,y
	cmp $bb90
	beq $bc0e
	inx
	lda $bbc2,x
	sta $0219,y
	rts

	jsr $bcc7
	jmp $bc20

	jsr $bd95
	jmp $bc20

	ldy #$0e
	lda $bae2,y
	sta $06d5
	lda #$04
	jsr $bc99
	jsr $bdce
	lda $0711
	beq $bc55
	ldy #$00
	lda $0781
	cmp $0711
	sty $0711
	bcs $bc55
	sta $0711
	ldy #$07
	lda $bae2,y
	sta $06d5
	ldy #$04
	lda $57
	ora $0c
	beq $bc51
	dey
	tya
	jsr $bc99
	lda $03d0
	lsr
	lsr
	lsr
	lsr
	sta $00
	ldx #$03
	lda $06e4
	clc
	adc #$18
	tay
	lda #$f8
	lsr $00
	bcc $bc70
	jsr $b267
	tya
	sec
	sbc #$08
	tay
	dex
	bpl $bc67
	rts

	.byte $58
	.byte $01
	.byte $00
	.byte $60
	.byte $ff
	.byte $04
	ldx #$05
	lda $bc79,x
	sta $02,x
	dex
	bpl $bc81
	ldx #$b8
	ldy #$04
	jsr $bcb7
	lda $0226
	ora #$40
	sta $0222
	rts

	sta $07
	lda $03ad
	sta $0755
	sta $05
	lda $03b8
	sta $02
	lda $33
	sta $03
	lda $03c4
	sta $04
	ldx $06d5
	ldy $06e4
	lda $baf2,x
	sta $00
	lda $baf3,x
	jsr $b88d
	dec $07
	bne $bcb7
	rts

	lda $1d
	cmp #$03
	beq $bd29
	cmp #$02
	beq $bd19
	cmp #$01
	bne $bce6
	lda $0704
	bne $bd35
	ldy #$06
	lda $0714
	bne $bd0d
	ldy #$00
	jmp $bd0d

	ldy #$06
	lda $0714
	bne $bd0d
	ldy #$02
	lda $57
	ora $0c
	beq $bd0d
	lda $0700
	cmp #$09
	bcc $bd21
	lda $45
	and $33
	bne $bd21
	lda $0e
	cmp #$09
	bcs $bd0c
	lda #$80
	sta $fd
	iny
	jsr $bd76
	lda #$00
	sta $070d
	lda $bae2,y
	rts

	ldy #$04
	jsr $bd76
	jmp $bd47

	ldy #$04
	jsr $bd76
	jmp $bd4d

	ldy #$05
	lda $9f
	beq $bd0d
	jsr $bd76
	jmp $bd52

	ldy #$01
	jsr $bd76
	lda $0782
	ora $070d
	bne $bd4d
	lda $0a
	asl
	bcs $bd4d
	lda $070d
	jmp $bdb5

	lda #$03
	jmp $bd54

	lda #$02
	sta $00
	jsr $bd47
	pha
	lda $0781
	bne $bd74
	lda $070c
	sta $0781
	lda $070d
	clc
	adc #$01
	cmp $00
	bcc $bd71
	lda #$00
	sta $070d
	pla
	rts

	lda $0754
	beq $bd80
	tya
	clc
	adc #$08
	tay
	rts

	.byte $00
	.byte $01
	.byte $00
	.byte $01
	.byte $00
	.byte $01
	.byte $02
	.byte $00
	.byte $01
	.byte $02
	.byte $02
	.byte $00
	.byte $02
	.byte $00
	.byte $02
	.byte $00
	.byte $02
	.byte $00
	.byte $02
	.byte $00
	ldy $070d
	lda $09
	and #$03
	bne $bdab
	iny
	cpy #$0a
	bcc $bda8
	ldy #$00
	sty $070b
	sty $070d
	lda $0754
	bne $bdbc
	lda $bd81,y
	ldy #$0f
	asl
	asl
	asl
	adc $bae2,y
	rts

	tya
	clc
	adc #$0a
	tax
	ldy #$09
	lda $bd81,x
	bne $bdca
	ldy #$01
	lda $bae2,y
	rts

	ldy $06e4
	lda $0e
	cmp #$0b
	beq $bdea
	lda $06d5
	cmp #$50
	beq $bdfc
	cmp #$b8
	beq $bdfc
	cmp #$c0
	beq $bdfc
	cmp #$c8
	bne $be0e
	lda $0212,y
	and #$3f
	sta $0212,y
	lda $0216,y
	and #$3f
	ora #$40
	sta $0216,y
	lda $021a,y
	and #$3f
	sta $021a,y
	lda $021e,y
	and #$3f
	ora #$40
	sta $021e,y
	rts

	ldx #$00
	ldy #$00
	jmp $be27

	ldy #$01
	jsr $be8d
	ldy #$03
	jmp $be27

	ldy #$00
	jsr $be8d
	ldy #$02
	jsr $be56
	ldx $08
	rts

	ldy #$02
	jsr $be8d
	ldy #$06
	jmp $be27

	lda #$01
	ldy #$01
	jmp $be4a

	lda #$09
	ldy #$04
	jsr $be4a
	inx
	inx
	lda #$09
	iny
	stx $00
	clc
	adc $00
	tax
	jsr $be56
	ldx $08
	rts

	lda $ce,x
	sta $03b8,y
	lda $86,x
	sec
	sbc $071c
	sta $03ad,y
	rts

	ldx #$00
	ldy #$00
	jmp $bea5

	ldy #$00
	jsr $be8d
	ldy #$02
	jmp $bea5

	ldy #$01
	jsr $be8d
	ldy #$03
	jmp $bea5

	ldy #$02
	jsr $be8d
	ldy #$06
	jmp $bea5

	.byte $07
	.byte $16
	.byte $0d
	txa
	clc
	adc $be8a,y
	tax
	rts

	lda #$01
	ldy #$01
	jmp $be9f

	lda #$09
	ldy #$04
	stx $00
	clc
	adc $00
	tax
	tya
	pha
	jsr $bebc
	asl
	asl
	asl
	asl
	ora $00
	sta $00
	pla
	tay
	lda $00
	sta $03d0,y
	ldx $08
	rts

	jsr $bedb
	lsr
	lsr
	lsr
	lsr
	sta $00
	jmp $bf1e

	.byte $7f
	.byte $3f
	.byte $1f
	.byte $0f
	.byte $07
	.byte $03
	.byte $01
	.byte $00
	.byte $80
	.byte $c0
	.byte $e0
	.byte $f0
	.byte $f8
	.byte $fc
	.byte $fe
	.byte $ff
	.byte $07
	.byte $0f
	.byte $07
	stx $04
	ldy #$01
	lda $071c,y
	sec
	sbc $86,x
	sta $07
	lda $071a,y
	sbc $6d,x
	ldx $bed8,y
	cmp #$00
	bmi $bf03
	ldx $bed9,y
	cmp #$01
	bpl $bf03
	lda #$38
	sta $06
	lda #$08
	jsr $bf52
	lda $bec8,x
	ldx $04
	cmp #$00
	bne $bf0f
	dey
	bpl $bedf
	rts

	.byte $0f
	.byte $07
	.byte $03
	.byte $01
	.byte $00
	.byte $08
	.byte $0c
	.byte $0e
	.byte $00
	.byte $04
	.byte $00
	.byte $04
	.byte $00
	.byte $ff
	stx $04
	ldy #$01
	lda $bf1c,y
	sec
	sbc $ce,x
	sta $07
	lda #$01
	sbc $b5,x
	ldx $bf19,y
	cmp #$00
	bmi $bf45
	ldx $bf1a,y
	cmp #$01
	bpl $bf45
	lda #$20
	sta $06
	lda #$04
	jsr $bf52
	lda $bf10,x
	ldx $04
	cmp #$00
	bne $bf51
	dey
	bpl $bf22
	rts

	sta $05
	lda $07
	cmp $06
	bcs $bf66
	lsr
	lsr
	lsr
	and #$07
	cpy #$01
	bcs $bf65
	adc $05
	tax
	rts

	lda $03
	lsr
	lsr
	lda $00
	bcc $bf7b
	sta $0205,y
	lda $01
	sta $0201,y
	lda #$40
	bne $bf85
	sta $0201,y
	lda $01
	sta $0205,y
	lda #$00
	ora $04
	sta $0202,y
	sta $0206,y
	lda $02
	sta $0200,y
	sta $0204,y
	lda $05
	sta $0203,y
	clc
	adc #$08
	sta $0207,y
	lda $02
	clc
	adc #$08
	sta $02
	tya
	clc
	adc #$08
	tay
	inx
	inx
	rts

	.align $10

TitleScreenMode:
	lda $0772
	jsr $6c7d
	.word $c006, $c592, $64c5
	.word $c5db, $c46e, $bfc2
	lda $07fc
	jsr $6c7d
	.word $c113, $bfd2, $c126
	.word $c138, $c140
	lda $07fb
	beq $bfe6
	lda #$03
	sta $07f7
	jsr $c0f4
	bne $c038
	jsr $c0ca
	bne $c036
	jsr $c2a4
	lda $07fb
	beq $bff1
	jsr $c3d6
	inc $075d
	inc $0757
	inc $0770
	lda #$00
	sta $07fc
	sta $0772
	sta $07a2
	rts

	lda $07fc
	jsr $6c7d
	.word $c113, $c016, $c126
	.word $c138, $c140
	lda $07fd
	beq $c03e
	lda $07fb
	bne $c027
	lda $075f
	cmp #$04
	bcc $c03e
	lda #$00
	sta $07f7
	jsr $c0f4
	bne $c038
	jsr $c0ca
	beq $c03e
	lda #$40
	inc $07fc
	jmp $c182

	lda #$01
	sta $07fd
	lsr
	sta $075f
	sta $07fb
	jmp $c078

	lda $07fc
	jsr $6c7d
	.word $c113, $c05d, $c126
	.word $c138, $c140
	lda $075f
	cmp #$04
	bcc $c078
	lda $07f7
	bne $c078
	lda #$01
	sta $07f7
	jsr $c0f4
	bne $c038
	jsr $c0ca
	bne $c036
	lda #$00
	sta $07fc
	inc $0772
	rts

	lda #$10
	sta $07a1
	bne $c07d
	lda $07a1
	beq $c07d
	rts

	lda $07fc
	jsr $6c7d
	.word $c113, $c09e, $c126
	.word $c138, $c140
	lda #$02
	sta $07f7
	jsr $c0f4
	bne $c038
	jsr $c0ca
	beq $c0b2
	lda #$00
	sta $d29f
	lda $d29f
	clc
	adc #$01
	cmp #$19
	bcc $c0be
	lda #$18
	sta $d29f
	jsr $6c92
	jsr $c078
	jmp $c858

	tya
	ldy $07f7
	cmp $c0f0,y
	rts

	.byte $01
	.byte $53
	.byte $4d
	.byte $42
	.byte $20
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $00
	.byte $e4
	.byte $e8
	.byte $ea
	.byte $ee
	.byte $c0
	.byte $c0
	.byte $c0
	.byte $c0
	.byte $01
	.byte $05
	.byte $0f
	.byte $ff
	.byte $20
	.byte $ff
	.byte $10
	.byte $30
	.byte $0f
	.byte $ff
	.byte $40
	.byte $ff
	.byte $03
	.byte $01
	.byte $03
	.byte $01
	ldx $07f7
	lda $c0dc,x
	sta $c108
	lda $c0e0,x
	sta $c109
	jsr $e1f8
	.byte $d2
	.byte $c0
	.byte $e4
	.byte $c0
	rts

	.byte $3f
	.byte $00
	.byte $04
	.byte $0f
	.byte $30
	.byte $30
	.byte $0f
	.byte $00
	lda #$00
	sta $0779
	sta $2001
	inc $0774
	lda #$1a
	sta $0773
	jmp $c134

	lda #$00
	sta $077a
	sta $0774
	lda $4032
	lsr
	bcc $c137
	inc $07fc
	rts

	lda $4032
	lsr
	bcc $c134
	bcs $c137
	lda #$00
	sta $07fc
	sta $07f7
	rts

	.byte $21
	.byte $e6
	.byte $08
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $21
	.byte $f4
	.byte $06
	.byte $0e
	.byte $1b
	.byte $1b
	.byte $24
	.byte $00
	.byte $01
	.byte $00
	.byte $07
	.byte $0f
	.byte $17
	.byte $1f
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $24
	.byte $0d
	.byte $12
	.byte $1c
	.byte $14
	.byte $24
	.byte $1c
	.byte $0e
	.byte $1d
	.byte $0b
	.byte $0a
	.byte $1d
	.byte $1d
	.byte $0e
	.byte $1b
	.byte $22
	.byte $24
	.byte $0a
	.byte $24
	.byte $0b
	.byte $24
	.byte $1c
	.byte $12
	.byte $0d
	.byte $0e
	pha
	and #$0f
	sta $c15c
	pla
	pha
	lsr
	lsr
	lsr
	lsr
	sta $c15b
	ldy #$03
	pla
	cmp #$07
	beq $c1a3
	dey
	cmp #$02
	beq $c1a3
	dey
	cmp #$01
	beq $c1a3
	dey
	ldx $c15e,y
	ldy #$07
	lda $c162,x
	sta $c14c,y
	dex
	dey
	bpl $c1a8
	lda #$19
	sta $0773
	jsr $6287
	jmp $6c92

	.byte $5b
	.byte $02
	.byte $48
	.byte $77
	.byte $8f
	lda $06fc
	and #$10
	bne $c1f6
	lda $06fc
	and #$20
	beq $c1e1
	ldx $0780
	bne $c1e1
	lsr
	sta $0780
	lda $07f8
	eor #$01
	sta $07f8
	ldy #$02
	lda $c1bd,y
	sta $0201,y
	dey
	bpl $c1e3
	ldy $07f8
	lda $c1c0,y
	sta $0200
	rts

	lda $07f8
	beq $c203
	lda #$00
	sta $07fa
	jmp $708d

	ldy #$02
	sty $075a
	sta $075c
	sta $0760
	sta $075e
	ldy #$0b
	sta $07dd,y
	dey
	bpl $c213
	inc $075d
	jmp $709d

	.byte $20
	.byte $20
	.byte $1e
	.byte $28
	.byte $28
	.byte $0d
	.byte $04
	.byte $70
	.byte $70
	.byte $60
	.byte $90
	.byte $90
	.byte $0a
	.byte $09
	.byte $e4
	.byte $98
	.byte $d0
	.byte $18
	.byte $18
	.byte $18
	.byte $22
	.byte $22
	.byte $0d
	.byte $04
	.byte $42
	.byte $42
	.byte $3e
	.byte $5d
	.byte $5d
	.byte $0a
	.byte $09
	.byte $b4
	.byte $68
	.byte $a0
	ldx #$60
	ldy #$21
	lda $0753
	bne $c24e
	ldx #$0e
	ldy #$10
	stx $80f1
	ldx #$10
	lda $c21f,y
	sta $7f90,x
	dey
	dex
	bpl $c253
	rts

    .align $10

E_OverworldExit:
E_PipeIntro:    
E_OverworldWarp:
	.byte $ff
	
L_OverworldExit:	
    .byte $90, $31, $39, $f1, $bf, $37, $33, $e7, $a3, $03
    .byte $a7, $03, $cd, $41, $0f, $a6, $ed, $47, $fd
	
L_PipeIntro:	
	.byte $38, $11, $0f, $26, $ad, $40, $3d, $c7, $fd
	
L_OverworldWarp:
    .byte $10, $00, $0b, $13, $5b, $14, $6a, $42, $c7, $12
    .byte $c6, $42, $1b, $94, $2a, $42, $53, $13, $62, $41
    .byte $97, $17, $a6, $45, $6e, $81, $8f, $37, $02, $e8
    .byte $12, $3a, $68, $7a, $de, $0f, $6d, $c5, $fd

;-------------------------------------------------------------------------------------

LoadAreaPointer:
	jsr FindAreaPointer
	sta AreaPointer
GetAreaType:
	and #%01100000
	asl
	rol
	rol
	rol
	sta AreaType
	rts

FindAreaPointer:
	ldy WorldNumber
	lda WorldAddrOffsets,y
	clc
	adc AreaNumber
	tay
	lda AreaAddrOffsets,y
	rts

GetAreaDataAddrs:
	lda AreaPointer
	jsr GetAreaType
	tay
	lda AreaPointer
	and #%0011111
	sta AreaAddrsOffset
	lda EnemyDataHOffsets,y
	clc
	adc AreaAddrsOffset
	asl
	tay
	lda EnemyDataOffsets+1,y
	sta EnemyDataHigh
	lda EnemyDataOffsets,y
	sta EnemyDataLow
	ldy AreaType
	lda AreaDataHOffsets,y
	clc
	adc AreaAddrsOffset
	asl
	tay
	lda AreaDataOffsets+1,y
	sta AreaDataHigh
	lda AreaDataOffsets,y
	sta AreaDataLow
	ldy #$00
	lda (AreaData),y
	pha
	and #%00000111
	cmp #%00000100
	bcc StoreFore
	sta BackgroundColorCtrl
	lda #$00
StoreFore:
	sta ForegroundScenery
	pla
	pha
	and #%00111000
	lsr
	lsr
	lsr
	sta PlayerEntranceCtrl
	pla
	and #%11000000
	clc
	rol
	rol
	rol
	sta GameTimerSetting
	iny
	lda (AreaData),y
	pha
	and #%00001111
	sta TerrainControl
	pla
	pha
	and #%00110000
	lsr
	lsr
	lsr
	lsr
	sta BackgroundScenery
	pla
	and #$c0
	clc
	rol
	rol
	rol
	cmp #%00000011
	bne StoreStyle
	sta CloudTypeOverride
	lda #$00
StoreStyle:
	sta AreaStyle
	lda AreaDataLow
	clc
	adc #$02
	sta AreaDataLow
	lda AreaDataHigh
	adc #$00
	sta AreaDataHigh
	rts

;-------------------------------------------------------------------------------------
;GAME LEVELS DATA

WorldAddrOffsets:
    .byte World1Areas-AreaAddrOffsets, World2Areas-AreaAddrOffsets
    .byte World3Areas-AreaAddrOffsets, World4Areas-AreaAddrOffsets
    .byte World5Areas-AreaAddrOffsets, World6Areas-AreaAddrOffsets
    .byte World7Areas-AreaAddrOffsets, World8Areas-AreaAddrOffsets
    .byte World9Areas-AreaAddrOffsets
	
AreaAddrOffsets:
World1Areas:
	.byte $20, $29, $40, $21, $60
	
World2Areas:
	.byte $22, $23, $24, $61
	
World3Areas:
	.byte $25, $29, $00, $26, $62
	
World4Areas:
	.byte $27, $28, $2a, $63
	
World5Areas:	
	.byte $2b, $29, $43, $2c, $64
	
World6Areas:
	.byte $2d, $29, $01, $2e, $65
	
World7Areas:
	.byte $2f, $30, $31, $66
	
World8Areas:
	.byte $32, $35, $36, $67
	
World9Areas:
	.byte $38, $06, $68, $07
	
;something about loops:
	.byte $0c
	.byte $0c
	.byte $42
	.byte $42
	.byte $10
	.byte $10
	.byte $30
	.byte $30
	.byte $06
	.byte $0c
	.byte $54
	.byte $06
	
EnemyDataHOffsets:
	.byte (E_WaterAreas-EnemyDataOffsets)>>1, (E_GroundAreas-EnemyDataOffsets)>>1
	.byte (E_UndergroundAreas-EnemyDataOffsets)>>1, (E_CastleAreas-EnemyDataOffsets)>>1
	
EnemyDataOffsets:
E_CastleAreas:
	.word E_14, E_24, E_34, E_44, E_54, E_64, E_74, E_84, E_93, E_93_Cloud

E_GroundAreas:
	.word E_11, E_13, E_21, E_22, E_23, E_31, E_33, E_41, E_42, E_PipeIntro, E_43
	.word E_51, E_53, E_61, E_63, E_71, E_72, E_73, E_81
	.word E_1t4_Cloud, E_OverworldWarp, E_82, E_83, E_Unused_37
	.word E_91_Start, E_Unused_39, E_Unused_3A, E_OverworldExit, E_5t8_Cloud
	
E_UndergroundAreas:
	.word E_12, E_12_Warp, E_1t4_Bonus, E_52, E_5t8_Bonus
	
E_WaterAreas:
	.word E_32, E_62, E_41_Water, E_84_Water, E_5t8_Water, E_91, E_92, E_94
	
AreaDataHOffsets:
	.byte (L_WaterAreas-AreaDataOffsets)>>1, (L_GroundAreas-AreaDataOffsets)>>1
	.byte (L_UndergroundAreas-AreaDataOffsets)>>1, (L_CastleAreas-AreaDataOffsets)>>1

AreaDataOffsets:
L_CastleAreas:
	.word L_14, L_24, L_34, L_44, L_54, L_64, L_74, L_84, L_93, L_93_Cloud

L_GroundAreas:
	.word L_11, L_13, L_21, L_22, L_23, L_31, L_33, L_41, L_42, L_PipeIntro, L_43
	.word L_51, L_53, L_61, L_63, L_71, L_72, L_73, L_81
	.word L_1t4_Cloud, L_OverworldWarp, L_82, L_83, L_Unused_37
	.word L_91_Start, L_Unused_39, L_Unused_3A, L_OverworldExit, L_5t8_Cloud
	
L_UndergroundAreas:
	.word L_12, L_12_Warp, L_1t4_Bonus, L_52, L_5t8_Bonus
	
L_WaterAreas
	.word L_32, L_62, L_41_Water, L_84_Water, L_5t8_Water, L_91, L_92, L_94

	.word $ffff                                ;unused
	
	lda $06fc
	and #$10
	beq $c494
	lda #$00
	sta $07fa
	sta $07fc
	sta $07fb
	lda $d29f
	cmp #$08
	bcc $c491
	lda $06fc
	and #$80
	beq $c491
	inc $07fb
	jmp $c4ec

	lda $06fc
	cmp #$20
	beq $c4aa
	ldx $07a2
	bne $c4cf
	sta $0780
	jsr $c553
	bcs $c4dd
	bcc $c4d4
	lda $07a2
	beq $c4dd
	lda #$18
	sta $07a2
	lda $09
	and #$fe
	sta $09
	lda $0780
	bne $c4cf
	lda #$10
	sta $0780
	lda $0753
	eor #$01
	sta $0753
	jsr $c51b
	lda #$00
	sta $06fc
	jsr $7a47
	lda $0e
	cmp #$06
	bne $c510
	lda #$00
	sta $0770
	sta $0772
	sta $0722
	inc $0774
	rts

	lda $07a2
	beq $c4dd
	inc $0772
	jsr $c5ff
	lda #$00
	sta $075f
	lda #$00
	sta $075c
	lda #$00
	sta $0760
	ldx #$0b
	lda #$00
	sta $07dd,x
	dex
	bpl $c50a
	rts

	.byte $22
	.byte $4b
	.byte $83
	.byte $ce
	.byte $24
	.byte $24
	.byte $00
	.byte $ce
	.byte $24
	.byte $ce
	lda #$1c
	sta $0773
	ldy $0753
	lda $c518,y
	sta $c514
	lda $c519,y
	sta $c516
	rts

	.byte $01
	.byte $81
	.byte $01
	.byte $81
	.byte $01
	.byte $81
	.byte $02
	.byte $01
	.byte $81
	.byte $00
	.byte $81
	.byte $00
	.byte $80
	.byte $01
	.byte $81
	.byte $01
	.byte $00
	.byte $b0
	.byte $10
	.byte $10
	.byte $10
	.byte $28
	.byte $10
	.byte $28
	.byte $06
	.byte $10
	.byte $10
	.byte $0c
	.byte $80
	.byte $10
	.byte $28
	.byte $08
	.byte $90
	.byte $ff
	.byte $00
	ldx $0717
	lda $0718
	bne $c568
	inx
	inc $0717
	sec
	lda $c540,x
	sta $0718
	beq $c572
	lda $c52f,x
	sta $06fc
	dec $0718
	clc
	rts

	lda $0770
	bne $c58f
	ldx #$00
	sta $0300,x
	sta $0400,x
	dex
	bne $c57a
	jsr $c51b
	inc $073c
	rts

	lda #$fa
	jsr $87ef
	jmp $6326

	lda #$00
	sta $07fa
	sta $07fb
	sta $0753
	jsr $c5ff
	jsr $c520
	ldy #$33
	lda #$0c
	sta $00
	ldx #$00
	lda #$26
	cpx $d29f
	bcs $c5b4
	lda #$f1
	sta $c62b,y
	iny
	dec $00
	bne $c5be
	ldy #$4d
	inx
	cpx #$18
	bne $c5ab
	ldy #$6f
	jsr $6f08
	ldy #$1f
	sta $07b0,y
	dey
	bpl $c5ca
	lda #$18
	sta $07a2
	jsr $c2a4
	jmp $6e39

	lda #$01
	sta $0757
	sta $0754
	lda #$02
	sta $075a
	jmp $6eb9

	.byte $16
	.byte $0a
	.byte $1b
	.byte $12
	.byte $18
	.byte $15
	.byte $1e
	.byte $12
	.byte $10
	.byte $12
	.byte $22
	.byte $16
	.byte $27
	.byte $18
	.byte $22
	.byte $30
	.byte $27
	.byte $19
	.byte $04
	.byte $09
	ldy $0753
	lda $c5fd,y
	pha
	iny
	sty $00
	tay
	ldx #$04
	lda $c5eb,y
	sta $6675,x
	sta $6c4a,x
	dey
	dex
	bpl $c60c
	pla
	sec
	sbc $00
	tay
	ldx #$03
	lda $c5f5,y
	sta $6537,x
	dey
	dex
	bpl $c620
	rts

TitleScreenData:
	.byte $20
	.byte $84
	.byte $01
	.byte $44
	.byte $20
	.byte $85
	.byte $57
	.byte $48
	.byte $20
	.byte $9c
	.byte $01
	.byte $49
	.byte $20
	.byte $a4
	.byte $c9
	.byte $46
	.byte $20
	.byte $a5
	.byte $57
	.byte $26
	.byte $20
	.byte $bc
	.byte $c9
	.byte $4a
	.byte $20
	.byte $a5
	.byte $0a
	.byte $d0
	.byte $d1
	.byte $d8
	.byte $d8
	.byte $de
	.byte $d1
	.byte $d0
	.byte $da
	.byte $de
	.byte $d1
	.byte $20
	.byte $c5
	.byte $17
	.byte $d2
	.byte $d3
	.byte $db
	.byte $db
	.byte $db
	.byte $d9
	.byte $db
	.byte $dc
	.byte $db
	.byte $df
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $20
	.byte $e5
	.byte $17
	.byte $d4
	.byte $d5
	.byte $d4
	.byte $d9
	.byte $db
	.byte $e2
	.byte $d4
	.byte $da
	.byte $db
	.byte $e0
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $26
	.byte $21
	.byte $05
	.byte $57
	.byte $26
	.byte $21
	.byte $05
	.byte $0a
	.byte $d6
	.byte $d7
	.byte $d6
	.byte $d7
	.byte $e1
	.byte $26
	.byte $d6
	.byte $dd
	.byte $e1
	.byte $e1
	.byte $21
	.byte $25
	.byte $17
	.byte $d0
	.byte $e8
	.byte $d1
	.byte $d0
	.byte $d1
	.byte $de
	.byte $d1
	.byte $d8
	.byte $d0
	.byte $d1
	.byte $26
	.byte $de
	.byte $d1
	.byte $de
	.byte $d1
	.byte $d0
	.byte $d1
	.byte $d0
	.byte $d1
	.byte $26
	.byte $26
	.byte $d0
	.byte $d1
	.byte $21
	.byte $45
	.byte $17
	.byte $db
	.byte $42
	.byte $42
	.byte $db
	.byte $42
	.byte $db
	.byte $42
	.byte $db
	.byte $db
	.byte $42
	.byte $26
	.byte $db
	.byte $42
	.byte $db
	.byte $42
	.byte $db
	.byte $42
	.byte $db
	.byte $42
	.byte $26
	.byte $26
	.byte $db
	.byte $42
	.byte $21
	.byte $65
	.byte $46
	.byte $db
	.byte $21
	.byte $6b
	.byte $11
	.byte $df
	.byte $db
	.byte $db
	.byte $db
	.byte $26
	.byte $db
	.byte $df
	.byte $db
	.byte $df
	.byte $db
	.byte $db
	.byte $e4
	.byte $e5
	.byte $26
	.byte $26
	.byte $ec
	.byte $ed
	.byte $21
	.byte $85
	.byte $17
	.byte $db
	.byte $db
	.byte $db
	.byte $de
	.byte $43
	.byte $db
	.byte $e0
	.byte $db
	.byte $db
	.byte $db
	.byte $26
	.byte $db
	.byte $e3
	.byte $db
	.byte $e0
	.byte $db
	.byte $db
	.byte $e6
	.byte $e3
	.byte $26
	.byte $26
	.byte $ee
	.byte $ef
	.byte $21
	.byte $a5
	.byte $17
	.byte $db
	.byte $db
	.byte $db
	.byte $db
	.byte $42
	.byte $db
	.byte $db
	.byte $db
	.byte $d4
	.byte $d9
	.byte $26
	.byte $db
	.byte $d9
	.byte $db
	.byte $db
	.byte $d4
	.byte $d9
	.byte $d4
	.byte $d9
	.byte $e7
	.byte $26
	.byte $de
	.byte $da
	.byte $21
	.byte $c4
	.byte $19
	.byte $5f
	.byte $95
	.byte $95
	.byte $95
	.byte $95
	.byte $95
	.byte $95
	.byte $95
	.byte $95
	.byte $97
	.byte $98
	.byte $78
	.byte $95
	.byte $96
	.byte $95
	.byte $95
	.byte $97
	.byte $98
	.byte $97
	.byte $98
	.byte $95
	.byte $78
	.byte $95
	.byte $f0
	.byte $7a
	.byte $21
	.byte $ef
	.byte $0e
	.byte $cf
	.byte $01
	.byte $09
	.byte $08
	.byte $06
	.byte $24
	.byte $17
	.byte $12
	.byte $17
	.byte $1d
	.byte $0e
	.byte $17
	.byte $0d
	.byte $18
	.byte $22
	.byte $4d
	.byte $0a
	.byte $16
	.byte $0a
	.byte $1b
	.byte $12
	.byte $18
	.byte $24
	.byte $10
	.byte $0a
	.byte $16
	.byte $0e
	.byte $22
	.byte $8d
	.byte $0a
	.byte $15
	.byte $1e
	.byte $12
	.byte $10
	.byte $12
	.byte $24
	.byte $10
	.byte $0a
	.byte $16
	.byte $0e
	.byte $22
	.byte $eb
	.byte $04
	.byte $1d
	.byte $18
	.byte $19
	.byte $28
	.byte $22
	.byte $f5
	.byte $01
	.byte $00
	.byte $23
	.byte $c9
	.byte $47
	.byte $55
	.byte $23
	.byte $d1
	.byte $47
	.byte $55
	.byte $23
	.byte $d9
	.byte $47
	.byte $55
	.byte $23
	.byte $cc
	.byte $43
	.byte $f5
	.byte $23
	.byte $d6
	.byte $01
	.byte $dd
	.byte $23
	.byte $de
	.byte $01
	.byte $5d
	.byte $23
	.byte $e2
	.byte $04
	.byte $55
	.byte $aa
	.byte $aa
	.byte $aa
	.byte $23
	.byte $ea
	.byte $04
	.byte $95
	.byte $aa
	.byte $aa
	.byte $2a
	.byte $00
	.byte $ff
	.byte $ff
	
E_14:
	.byte $35, $9d, $55, $9b, $c9, $1b, $59, $9d, $45, $9b
	.byte $c5, $1b, $26, $80, $45, $1b, $b9, $1d, $f0, $15
	.byte $59, $9d, $0f, $08, $78, $2d, $96, $28, $90, $b5
	.byte $ff

E_24:
	.byte $74, $80, $f0, $38, $a0, $bb, $40, $bc, $8c, $1d
	.byte $c9, $9d, $05, $9b, $1c, $0c, $59, $1b, $b5, $1d
	.byte $2c, $8c, $40, $15, $7c, $1b, $dc, $1d, $6c, $8c
	.byte $bc, $0c, $78, $ad, $a5, $28, $90, $b5, $ff

E_34:
	.byte $0f, $04, $9c, $0c, $0f, $07, $c5, $1b, $65, $9d
	.byte $49, $9d, $5c, $8c, $78, $2d, $90, $b5, $ff

E_44:
	.byte $49, $9f, $67, $03, $79, $9d, $a0, $3a, $57, $9f
	.byte $bb, $1d, $d5, $25, $0f, $05, $18, $1d, $74, $00
	.byte $84, $00, $94, $00, $c6, $29, $49, $9d, $db, $05
	.byte $0f, $08, $05, $9b, $09, $1d, $b0, $38, $80, $95
	.byte $c0, $3c, $ec, $a8, $cc, $8c, $4a, $9b, $78, $2d
	.byte $90, $b5, $ff

E_11:
	.byte $07, $8e, $47, $03, $0f, $03, $10, $38, $1b, $80
	.byte $53, $06, $77, $0e, $83, $83, $a0, $3d, $90, $3b
	.byte $90, $b7, $60, $bc, $b7, $0e, $ee, $42, $00, $f7
	.byte $80, $6b, $83, $1b, $83, $ab, $06, $ff

E_13:
	.byte $96, $a4, $f9, $24, $d3, $83, $3a, $83, $5a, $03
	.byte $95, $07, $f4, $0f, $69, $a8, $33, $87, $86, $24
	.byte $c9, $24, $4b, $83, $67, $83, $17, $83, $56, $28
	.byte $95, $24, $0a, $a4, $ff

E_21:
	.byte $0f, $02, $47, $0e, $87, $0e, $c7, $0e, $f7, $0e
	.byte $27, $8e, $ee, $42, $25, $0f, $06, $ac, $28, $8c
	.byte $a8, $4e, $b3, $20, $8b, $8e, $f7, $90, $36, $90
	.byte $e5, $8e, $32, $8e, $c2, $06, $d2, $06, $e2, $06
	.byte $ff

E_22:
	.byte $15, $8e, $9b, $06, $e0, $37, $80, $bc, $0f, $04
	.byte $2b, $3b, $ab, $0e, $eb, $0e, $0f, $06, $f0, $37
	.byte $4b, $8e, $6b, $80, $bb, $3c, $4b, $bb, $ee, $42
	.byte $20, $1b, $bc, $cb, $00, $ab, $83, $eb, $bb, $0f
	.byte $0e, $1b, $03, $9b, $37, $d4, $0e, $a3, $86, $b3
	.byte $06, $c3, $06, $ff

E_23:
	.byte $c0, $be, $0f, $03, $38, $0e, $15, $8f, $aa, $83
	.byte $f8, $07, $0f, $07, $96, $10, $0f, $09, $48, $10
	.byte $ba, $03, $ff

E_31:
	.byte $87, $85, $a3, $05, $db, $83, $fb, $03, $93, $8f
	.byte $bb, $03, $ce, $42, $42, $9b, $83, $ae, $b3, $40
	.byte $db, $00, $f4, $0f, $33, $8f, $74, $0f, $10, $bc
	.byte $f5, $0f, $2e, $c2, $45, $b7, $03, $f7, $03, $c8
	.byte $90, $ff

E_33:
	.byte $80, $be, $83, $03, $92, $10, $4b, $80, $b0, $3c
	.byte $07, $80, $b7, $24, $0c, $a4, $96, $a9, $1b, $83
	.byte $7b, $24, $b7, $24, $97, $83, $e2, $0f, $a9, $a9
	.byte $38, $a9, $0f, $0b, $74, $8f, $ff

E_41:
	.byte $e2, $91, $0f, $03, $42, $11, $0f, $06, $72, $11
	.byte $0f, $08, $ee, $02, $60, $02, $91, $ee, $b3, $60
	.byte $d3, $86, $ff

E_42:
	.byte $0f, $02, $9b, $02, $ab, $02, $0f, $04, $13, $03
	.byte $92, $11, $60, $b7, $00, $bc, $00, $bb, $0b, $83
	.byte $cb, $03, $7b, $85, $9e, $c2, $60, $e6, $05, $0f
	.byte $0c, $62, $10, $ff

E_43:
	.byte $e6, $a9, $57, $a8, $b5, $24, $19, $a4, $76, $28
	.byte $a2, $0f, $95, $8f, $9d, $a8, $0f, $07, $09, $29
	.byte $55, $24, $8b, $17, $a9, $24, $db, $83, $04, $a9
	.byte $24, $8f, $65, $0f, $ff

E_1t4_Cloud:
	.byte $0a, $aa, $1e, $22, $29, $1e, $25, $49, $2e, $27
	.byte $66, $ff

E_12:
	.byte $0a, $8e, $de, $b4, $00, $e0, $37, $5b, $82, $2b
	.byte $a9, $aa, $29, $29, $a9, $a8, $29, $0f, $08, $f0
	.byte $3c, $79, $a9, $c5, $26, $cd, $26, $ee, $3b, $01
	.byte $67, $b4, $0f, $0c, $2e, $c1, $00, $ff

E_12_Warp:
	.byte $09, $a9, $19, $a9, $de, $42, $02, $7b, $83, $ff

E_1t4_Bonus:
	.byte $1e, $a0, $0a, $1e, $23, $2b, $1e, $28, $6b, $0f
	.byte $03, $1e, $40, $08, $1e, $25, $4e, $0f, $06, $1e
	.byte $22, $25, $1e, $25, $45, $ff

E_32:
	.byte $0f, $01, $2a, $07, $2e, $3b, $41, $e9, $07, $0f
	.byte $03, $6b, $07, $f9, $07, $b8, $80, $2a, $87, $4a
	.byte $87, $b3, $0f, $84, $87, $47, $83, $87, $07, $0a
	.byte $87, $42, $87, $1b, $87, $6b, $03, $ff

E_41_Water:
	.byte $1e, $a7, $6a, $5b, $82, $74, $07, $d8, $07, $e8
	.byte $02, $0f, $04, $26, $07, $ff

L_14:
	.byte $9b, $07, $05, $32, $06, $33, $07, $34, $33, $8e
	.byte $4e, $0a, $7e, $06, $9e, $0a, $ce, $06, $e3, $00
	.byte $ee, $0a, $1e, $87, $53, $0e, $8e, $02, $9c, $00
	.byte $c7, $0e, $d7, $37, $57, $8e, $6c, $05, $da, $60
	.byte $e9, $61, $f8, $62, $fe, $0b, $43, $8e, $c3, $0e
	.byte $43, $8e, $b7, $0e, $ee, $09, $fe, $0a, $3e, $86
	.byte $57, $0e, $6e, $0a, $7e, $06, $ae, $0a, $be, $06
	.byte $fe, $07, $15, $e2, $55, $62, $95, $62, $fe, $0a
	.byte $0d, $c4, $cd, $43, $ce, $09, $de, $0b, $dd, $42
	.byte $fe, $02, $5d, $c7, $fd

L_24:
	.byte $9b, $07, $05, $32, $06, $33, $07, $34, $03, $e2
	.byte $0e, $06, $1e, $0c, $7e, $0a, $8e, $05, $8e, $82
	.byte $8a, $8e, $8e, $0a, $ee, $02, $0a, $e0, $19, $61
	.byte $23, $06, $28, $62, $2e, $0b, $7e, $0a, $81, $62
	.byte $87, $30, $8e, $04, $a7, $31, $c7, $0e, $d7, $33
	.byte $fe, $03, $03, $8e, $0e, $0a, $11, $62, $1e, $04
	.byte $27, $32, $4e, $0a, $51, $62, $57, $0e, $5e, $04
	.byte $67, $34, $9e, $0a, $a1, $62, $ae, $03, $b3, $0e
	.byte $be, $0b, $ee, $09, $fe, $0a, $2e, $82, $7a, $0e
	.byte $7e, $0a, $97, $31, $be, $04, $da, $0e, $ee, $0a
	.byte $f1, $62, $fe, $02, $3e, $8a, $7e, $06, $ae, $0a
	.byte $ce, $06, $fe, $0a, $0d, $c4, $11, $53, $21, $52
	.byte $24, $0b, $51, $52, $61, $52, $cd, $43, $ce, $09
	.byte $dd, $42, $de, $0b, $fe, $02, $5d, $c7, $fd

L_34:
	.byte $5b, $09, $05, $34, $06, $35, $6e, $06, $7e, $0a
	.byte $ae, $02, $fe, $02, $0d, $01, $0e, $0e, $2e, $0a
	.byte $6e, $09, $be, $0a, $ed, $4b, $e4, $60, $ee, $0d
	.byte $5e, $82, $78, $72, $a4, $3d, $a5, $3e, $a6, $3f
	.byte $a3, $be, $a6, $3e, $a9, $32, $e9, $3a, $9c, $80
	.byte $a3, $33, $a6, $33, $a9, $33, $e5, $06, $ed, $4b
	.byte $f3, $30, $f6, $30, $f9, $30, $fe, $02, $0d, $05
	.byte $3c, $01, $57, $73, $7c, $02, $93, $30, $a7, $73
	.byte $b3, $37, $cc, $01, $07, $83, $17, $03, $27, $03
	.byte $37, $03, $64, $3b, $77, $3a, $0c, $80, $2e, $0e
	.byte $9e, $02, $a5, $62, $b6, $61, $cc, $02, $c3, $33
	.byte $ed, $4b, $03, $b7, $07, $37, $83, $37, $87, $37
	.byte $dd, $4b, $03, $b5, $07, $35, $5e, $0a, $8e, $02
	.byte $ae, $0a, $de, $06, $fe, $0a, $0d, $c4, $cd, $43
	.byte $ce, $09, $dd, $42, $de, $0b, $fe, $02, $5d, $c7
	.byte $fd

L_44:
	.byte $9b, $07, $05, $32, $06, $33, $07, $34, $4e, $03
	.byte $5c, $02, $0c, $f1, $27, $00, $3c, $74, $47, $0e
	.byte $fc, $00, $fe, $0b, $77, $8e, $ee, $09, $fe, $0a
	.byte $45, $b2, $55, $0e, $99, $32, $b9, $0e, $fe, $02
	.byte $0e, $85, $fe, $02, $16, $8e, $2e, $0c, $ae, $0a
	.byte $ee, $05, $1e, $82, $47, $0e, $07, $bd, $c4, $72
	.byte $de, $0a, $fe, $02, $03, $8e, $07, $0e, $13, $3c
	.byte $17, $3d, $e3, $03, $ee, $0a, $f3, $06, $f7, $03
	.byte $fe, $0e, $fe, $8a, $38, $e4, $4a, $72, $68, $64
	.byte $37, $b0, $98, $64, $a8, $64, $e8, $64, $f8, $64
	.byte $0d, $c4, $71, $64, $cd, $43, $ce, $09, $dd, $42
	.byte $de, $0b, $fe, $02, $5d, $c7, $fd

L_11:
	.byte $50, $31, $0f, $26, $13, $e4, $23, $24, $27, $23
	.byte $37, $07, $66, $61, $ac, $74, $c7, $01, $0b, $f1
	.byte $77, $73, $b6, $04, $db, $71, $5c, $82, $83, $2d
	.byte $a2, $47, $a7, $0a, $b7, $29, $4f, $b3, $87, $0b
	.byte $93, $23, $cc, $06, $e3, $2c, $3a, $e0, $7c, $71
	.byte $97, $01, $ac, $73, $e6, $61, $0e, $b1, $b7, $f3
	.byte $dc, $02, $d3, $25, $07, $fb, $2c, $01, $e7, $73
	.byte $2c, $f2, $34, $72, $57, $00, $7c, $02, $39, $f1
	.byte $bf, $37, $33, $e7, $cd, $41, $0f, $a6, $ed, $47
	.byte $fd

L_13:
	.byte $50, $11, $0f, $26, $fe, $10, $47, $92, $56, $40
	.byte $ac, $16, $af, $12, $0f, $95, $73, $16, $82, $44
	.byte $ec, $48, $bc, $c2, $1c, $b1, $b3, $16, $c2, $44
	.byte $86, $c0, $9c, $14, $9f, $12, $a6, $40, $df, $15
	.byte $0b, $96, $43, $12, $97, $31, $d3, $12, $03, $92
	.byte $27, $14, $63, $00, $c7, $15, $d6, $43, $ac, $97
	.byte $af, $11, $1f, $96, $64, $13, $e3, $12, $2e, $91
	.byte $9d, $41, $ae, $42, $df, $20, $cd, $c7, $fd

L_21:
	.byte $52, $21, $0f, $20, $6e, $64, $4f, $b2, $7c, $5f
	.byte $7c, $3f, $7c, $d8, $7c, $38, $83, $02, $a3, $00
	.byte $c3, $02, $f7, $16, $5c, $d6, $cf, $35, $d3, $20
	.byte $e3, $0a, $f3, $20, $25, $b5, $2c, $53, $6a, $7a
	.byte $8c, $54, $da, $72, $fc, $50, $0c, $d2, $39, $73
	.byte $5c, $54, $aa, $72, $cc, $53, $f7, $16, $33, $83
	.byte $40, $06, $5c, $5b, $09, $93, $27, $0f, $3c, $5c
	.byte $0a, $b0, $63, $27, $78, $72, $93, $09, $97, $03
	.byte $a7, $03, $b7, $22, $47, $81, $5c, $72, $2a, $b0
	.byte $28, $0f, $3c, $5f, $58, $31, $b8, $31, $28, $b1
	.byte $3c, $5b, $98, $31, $fa, $30, $03, $b2, $20, $04
	.byte $7f, $b7, $f3, $67, $8d, $c1, $bf, $26, $ad, $c7
	.byte $fd

L_22:
	.byte $54, $11, $0f, $26, $38, $f2, $ab, $71, $0b, $f1
	.byte $96, $42, $ce, $10, $1e, $91, $29, $61, $3a, $60
	.byte $4e, $10, $78, $74, $8e, $11, $06, $c3, $1a, $e0
	.byte $1e, $10, $5e, $11, $67, $63, $77, $63, $88, $62
	.byte $99, $61, $aa, $60, $be, $10, $0a, $f2, $15, $45
	.byte $7e, $11, $7a, $31, $9a, $e0, $ac, $02, $d9, $61
	.byte $d4, $0a, $ec, $01, $d6, $c2, $84, $c3, $98, $fa
	.byte $d3, $07, $d7, $0b, $e9, $61, $ee, $10, $2e, $91
	.byte $39, $71, $93, $03, $a6, $03, $be, $10, $e1, $71
	.byte $e3, $31, $5e, $91, $69, $61, $e6, $41, $28, $e2
	.byte $99, $71, $ae, $10, $ce, $11, $be, $90, $d6, $32
	.byte $3e, $91, $5f, $37, $66, $60, $d3, $67, $6d, $c1
	.byte $af, $26, $9d, $c7, $fd

L_23:
	.byte $54, $11, $0f, $26, $af, $32, $d8, $62, $e8, $62
	.byte $f8, $62, $fe, $10, $0c, $be, $f8, $64, $0d, $c8
	.byte $2c, $43, $98, $64, $ac, $39, $48, $e4, $6a, $62
	.byte $7c, $47, $fa, $62, $3c, $b7, $ea, $62, $fc, $4d
	.byte $f6, $02, $03, $80, $06, $02, $13, $02, $da, $62
	.byte $0d, $c8, $0b, $17, $97, $16, $2c, $b1, $33, $43
	.byte $6c, $31, $ac, $31, $17, $93, $73, $12, $cc, $31
	.byte $1a, $e2, $2c, $4b, $67, $48, $ea, $62, $0d, $ca
	.byte $17, $12, $53, $12, $be, $11, $1d, $c1, $3e, $42
	.byte $6f, $20, $4d, $c7, $fd

L_31:
	.byte $52, $b1, $0f, $20, $6e, $75, $53, $aa, $57, $25
	.byte $b7, $0a, $c7, $23, $0c, $83, $5c, $72, $87, $01
	.byte $c3, $00, $c7, $20, $dc, $65, $0c, $87, $c3, $22
	.byte $f3, $03, $03, $a2, $27, $7b, $33, $03, $43, $23
	.byte $52, $42, $9c, $06, $a7, $20, $c3, $23, $03, $a2
	.byte $0c, $02, $33, $09, $39, $71, $43, $23, $77, $06
	.byte $83, $67, $a7, $73, $5c, $82, $c9, $11, $07, $80
	.byte $1c, $71, $98, $11, $9a, $10, $f3, $04, $16, $f4
	.byte $3c, $02, $68, $7a, $8c, $01, $a7, $73, $e7, $73
	.byte $ac, $83, $09, $8f, $1c, $03, $9f, $37, $13, $e7
	.byte $7c, $02, $ad, $41, $ef, $26, $0d, $0e, $39, $71
	.byte $7f, $37, $f2, $68, $02, $e8, $12, $3a, $1c, $00
	.byte $68, $7a, $de, $3f, $6d, $c5, $fd

L_33:
	.byte $55, $10, $0b, $1f, $0f, $26, $d6, $12, $07, $9f
	.byte $33, $1a, $fb, $1f, $f7, $94, $53, $94, $71, $71
	.byte $cc, $15, $cf, $13, $1f, $98, $63, $12, $9b, $13
	.byte $a9, $71, $fb, $17, $09, $f1, $13, $13, $21, $42
	.byte $59, $0f, $eb, $13, $33, $93, $40, $06, $8c, $14
	.byte $8f, $17, $93, $40, $cf, $13, $0b, $94, $57, $15
	.byte $07, $93, $19, $f3, $c6, $43, $c7, $13, $d3, $03
	.byte $e3, $03, $33, $b0, $4a, $72, $55, $46, $73, $31
	.byte $a8, $74, $e3, $12, $8e, $91, $ad, $41, $ce, $42
	.byte $ef, $20, $dd, $c7, $fd

L_41:
	.byte $52, $21, $0f, $20, $6e, $63, $a9, $f1, $fb, $71
	.byte $22, $83, $37, $0b, $36, $50, $39, $51, $b8, $62
	.byte $57, $f3, $e8, $02, $f8, $02, $08, $82, $18, $02
	.byte $2d, $4a, $28, $02, $38, $02, $48, $00, $a8, $0f
	.byte $aa, $30, $bc, $5a, $6a, $b0, $4f, $b6, $b7, $04
	.byte $9a, $b0, $ac, $71, $c7, $01, $e6, $74, $0d, $09
	.byte $46, $02, $56, $00, $6c, $01, $84, $79, $86, $02
	.byte $96, $02, $a4, $71, $a6, $02, $b6, $02, $c4, $71
	.byte $c6, $02, $d6, $02, $39, $f1, $6c, $00, $77, $02
	.byte $a3, $09, $ac, $00, $b8, $72, $dc, $01, $07, $f3
	.byte $4c, $00, $6f, $37, $e3, $03, $e6, $03, $5d, $ca
	.byte $6c, $00, $7d, $41, $cf, $26, $9d, $c7, $fd

L_42:
	.byte $50, $a1, $0f, $26, $17, $91, $19, $11, $48, $00
	.byte $68, $11, $6a, $10, $96, $14, $d8, $0a, $e8, $02
	.byte $f8, $02, $dc, $81, $6c, $81, $89, $0f, $9c, $00
	.byte $c3, $29, $f8, $62, $47, $a7, $c6, $61, $0d, $07
	.byte $56, $74, $b7, $00, $b9, $11, $cc, $76, $ed, $4a
	.byte $1c, $80, $37, $01, $3a, $10, $de, $20, $e9, $0b
	.byte $ee, $21, $c8, $bc, $9c, $f6, $bc, $00, $cb, $7a
	.byte $eb, $72, $0c, $82, $39, $71, $b7, $63, $cc, $03
	.byte $e6, $60, $26, $e0, $4a, $30, $53, $31, $5c, $58
	.byte $ed, $41, $2f, $a6, $1d, $c7, $fd

L_43:
	.byte $50, $11, $0f, $26, $fe, $10, $8b, $93, $a9, $0f
	.byte $14, $c1, $cc, $16, $cf, $11, $2f, $95, $b7, $14
	.byte $c7, $96, $d6, $44, $2b, $92, $39, $0f, $72, $41
	.byte $a7, $00, $1b, $95, $97, $13, $6c, $95, $6f, $11
	.byte $a2, $40, $bf, $15, $c2, $40, $0b, $9f, $53, $16
	.byte $62, $44, $72, $c2, $9b, $1d, $b7, $e0, $ed, $4a
	.byte $03, $e0, $8e, $11, $9d, $41, $be, $42, $ef, $20
	.byte $cd, $c7, $fd

L_1t4_Cloud:
	.byte $00, $c1, $4c, $00, $03, $cf, $00, $d7, $23, $4d
	.byte $07, $af, $2a, $4c, $03, $cf, $3e, $80, $f3, $4a
	.byte $bb, $c2, $bd, $c7, $fd

L_12:
	.byte $48, $0f, $0e, $01, $5e, $02, $0a, $b0, $1c, $54
	.byte $6a, $30, $7f, $34, $c6, $64, $d6, $64, $e6, $64
	.byte $f6, $64, $fe, $00, $f0, $07, $00, $a1, $1e, $02
	.byte $47, $73, $7e, $04, $84, $52, $94, $50, $95, $0b
	.byte $96, $50, $a4, $52, $ae, $05, $b8, $51, $c8, $51
	.byte $ce, $01, $17, $f3, $45, $03, $52, $09, $62, $21
	.byte $6f, $34, $81, $21, $9e, $02, $b6, $64, $c6, $64
	.byte $c0, $0c, $d6, $64, $d0, $07, $e6, $64, $e0, $0c
	.byte $f0, $07, $fe, $0a, $0d, $06, $0e, $01, $4e, $04
	.byte $67, $73, $8e, $02, $b7, $0a, $bc, $03, $c4, $72
	.byte $c7, $22, $08, $f2, $2c, $02, $59, $71, $7c, $01
	.byte $96, $74, $bc, $01, $d8, $72, $fc, $01, $39, $f1
	.byte $4e, $01, $9e, $04, $a7, $52, $b7, $0b, $b8, $51
	.byte $c7, $51, $d7, $50, $de, $02, $3a, $e0, $3e, $0a
	.byte $9e, $00, $08, $d4, $18, $54, $28, $54, $48, $54
	.byte $6e, $06, $9e, $01, $a8, $52, $af, $47, $b8, $52
	.byte $c8, $52, $d8, $52, $de, $0f, $4d, $c7, $ce, $01
	.byte $dc, $01, $f9, $79, $1c, $82, $48, $72, $7f, $37
	.byte $f2, $68, $01, $e9, $11, $3a, $68, $7a, $de, $0f
	.byte $6d, $c5, $fd

L_12_Warp:
	.byte $0b, $0f, $0e, $01, $9c, $71, $b7, $00, $be, $00
	.byte $3e, $81, $47, $73, $5e, $00, $63, $42, $8e, $01
	.byte $a7, $73, $be, $00, $7e, $81, $88, $72, $f0, $59
	.byte $fe, $00, $00, $d9, $0e, $01, $39, $79, $a7, $03
	.byte $ae, $00, $b4, $03, $de, $0f, $0d, $05, $0e, $02
	.byte $68, $7a, $be, $01, $de, $0f, $6d, $c5, $fd

L_1t4_Bonus:
	.byte $08, $8f, $0e, $01, $17, $05, $2e, $02, $30, $07
	.byte $37, $03, $3a, $49, $44, $03, $58, $47, $df, $4a
	.byte $6d, $c7, $0e, $81, $00, $5a, $2e, $02, $87, $52
	.byte $97, $2f, $99, $4f, $0a, $90, $93, $56, $a3, $0b
	.byte $a7, $50, $b3, $55, $df, $4a, $6d, $c7, $0e, $81
	.byte $00, $5a, $2e, $00, $3e, $02, $41, $56, $57, $25
	.byte $56, $45, $68, $51, $7a, $43, $b7, $0b, $b8, $51
	.byte $df, $4a, $6d, $c7, $fd

L_32:
	.byte $41, $01, $03, $b4, $04, $34, $05, $34, $5c, $02
	.byte $83, $37, $84, $37, $85, $37, $09, $c2, $0c, $02
	.byte $1d, $49, $fa, $60, $09, $e1, $18, $62, $20, $63
	.byte $27, $63, $33, $37, $37, $63, $47, $63, $5c, $05
	.byte $79, $43, $fe, $06, $35, $d2, $46, $48, $91, $53
	.byte $d6, $51, $fe, $01, $0c, $83, $6c, $04, $b4, $62
	.byte $c4, $62, $d4, $62, $e4, $62, $f4, $62, $18, $d2
	.byte $79, $51, $f4, $66, $fe, $02, $0c, $8a, $1d, $49
	.byte $31, $55, $56, $41, $77, $41, $98, $41, $c5, $55
	.byte $fe, $01, $07, $e3, $17, $63, $27, $63, $37, $63
	.byte $47, $63, $57, $63, $67, $63, $78, $62, $89, $61
	.byte $9a, $60, $bc, $07, $ca, $42, $3a, $b3, $46, $53
	.byte $63, $34, $66, $44, $7c, $01, $9a, $33, $b7, $52
	.byte $dc, $01, $fa, $32, $05, $d4, $2c, $0d, $43, $37
	.byte $47, $35, $b7, $30, $c3, $64, $23, $e4, $29, $45
	.byte $33, $64, $43, $64, $53, $64, $63, $64, $73, $64
	.byte $9a, $60, $a9, $61, $b8, $62, $be, $0b, $d4, $31
	.byte $d5, $0d, $de, $0f, $0d, $ca, $7d, $47, $fd

L_41_Water:
	.byte $01, $01, $78, $52, $b5, $55, $da, $60, $e9, $61
	.byte $f8, $62, $fe, $0b, $fe, $81, $0a, $cf, $36, $49
	.byte $62, $43, $fe, $07, $36, $c9, $fe, $01, $0c, $84
	.byte $65, $55, $97, $52, $9a, $32, $a9, $31, $b8, $30
	.byte $c7, $63, $ce, $0f, $d5, $0d, $7d, $c7, $fd

;here is where the padding takes place
;push all this to the end of the PRG-RAM
    .pad FDS_BIOS_Start - (END_SM2MAIN_ - NbrGamesWon)

NbrGamesWon:
	.byte $00
EndSave

SoundEngine:
	lda OperMode
	bne SndOn
	sta SND_MASTERCTRL_REG
	rts

SndOn:
	lda #$ff
	sta JOYPAD_PORT2
	lda #%00001111
	sta SND_MASTERCTRL_REG
	lda PauseModeFlag
	bne InPause
	lda PauseSoundQueue
	cmp #$01
	bne RunSoundSubroutines
InPause:
	lda PauseSoundBuffer
	bne ContPau
	lda PauseSoundQueue
	beq SkipSoundSubroutines
	sta PauseSoundBuffer
	sta PauseModeFlag
	lda #$00
	sta SND_MASTERCTRL_REG
	sta Square1SoundBuffer
	sta Square2SoundBuffer
	sta NoiseSoundBuffer
	lda #%00001111
	sta SND_MASTERCTRL_REG
	lda #$2a
	sta Squ1_SfxLenCounter
PTone1F:	
	lda #%01000100
	bne PTRegC
ContPau:
	lda Squ1_SfxLenCounter
	cmp #$24
	beq PTone2F
	cmp #$1e
	beq PTone1F
	cmp #$18
	bne DecPauC
PTone2F:
	lda #%01100100
PTRegC:
	ldx #%10000100
	ldy #%01111111
	jsr PlaySqu1Sfx
DecPauC:
	dec Squ1_SfxLenCounter
	bne SkipSoundSubroutines
	lda #$00
	sta SND_MASTERCTRL_REG
	lda PauseSoundBuffer
	cmp #$02
	bne SkipPIn
	lda #$00
	sta PauseModeFlag
SkipPIn:
	lda #$00
	sta PauseSoundBuffer
	beq SkipSoundSubroutines

RunSoundSubroutines:
	jsr Square1SfxHandler
	jsr Square2SfxHandler
	jsr NoiseTriSfxHandler
	jsr MusicHandler
	lda #$00
	sta AreaMusicQueue
	sta EventMusicQueue
	
SkipSoundSubroutines:
	lda #$00
	sta Square1SoundQueue
	sta Square2SoundQueue
	sta NoiseSoundQueue
	sta PauseSoundQueue
	ldy DAC_Counter
	lda AreaMusicBuffer
	and #%00000011
	beq NoIncDAC
	inc DAC_Counter
	cpy #$30
	bcc StrWave
NoIncDAC:
	tya
	beq StrWave
	dec DAC_Counter
StrWave:
	sty SND_DELTA_REG+1
	rts

;-----------------------------------------------------------------------------------

Dump_Squ1_Regs:
	sty SND_SQUARE1_REG+1
	stx SND_SQUARE1_REG
	rts

PlaySqu1Sfx:
	jsr Dump_Squ1_Regs
	
SetFreq_Squ1:
	ldx #(SND_SQUARE1_REG - SND_REGISTER)
	
Dump_Freq_Regs:
	tay
	lda FreqRegLookupTbl+1,y
	beq NoTone
	sta SND_REGISTER+2,x
	lda FreqRegLookupTbl,y
	ora #%00001000
	sta SND_REGISTER+3,x
NoTone:
	rts

Dump_Squ2_Regs:
	stx SND_SQUARE2_REG
	sty SND_SQUARE2_REG+1
	rts

PlaySqu2Sfx:
	jsr Dump_Squ2_Regs

SetFreq_Squ2:
	ldx #(SND_SQUARE2_REG - SND_REGISTER)
	bne Dump_Freq_Regs

SetFreq_Tri:	
	ldx #(SND_TRIANGLE_REG - SND_REGISTER)
	bne Dump_Freq_Regs

;-----------------------------------------------------------------------------------
	
SwimStompEnvelopeData:
	.byte $9f, $9b, $98, $96, $95, $94, $92, $90, $90, $9a, $97, $95, $93, $92

PlayFlagpoleSlide:	
	lda #$40
	sta Squ1_SfxLenCounter
	lda #$62
	jsr SetFreq_Squ1
	ldx #%10011001
	bne FPS2nd
	
PlaySmallJump:
	lda #$26
	bne JumpRegContents
	
PlayBigJump:
	lda #$18

JumpRegContents:
	ldx #%10000010
	ldy #%10100111
	jsr PlaySqu1Sfx
	lda #$28
	sta Squ1_SfxLenCounter
	
ContinueSndJmp:
	lda Squ1_SfxLenCounter
	cmp #$25
	bne N2Prt
	ldx #%01011111
	ldy #%11110110
	bne DmpJpFPS
N2Prt:
	cmp #$20
	bne DecJpFPS
	ldx #%01001000
FPS2nd:
	ldy #%10111100
DmpJpFPS:
	jsr Dump_Squ1_Regs
	bne DecJpFPS

PlayFireballThrow:
	lda #$05
	ldy #%10011001
	bne FThrow
	
PlayBump:
	lda #$0a
	ldy #%10010011
FThrow:
	ldx #%10011110
	sta Squ1_SfxLenCounter
	lda #$0c
	jsr PlaySqu1Sfx
	
ContinueBumpThrow:
	lda Squ1_SfxLenCounter
	cmp #$06
	bne DecJpFPS
	lda #%10111011
	sta SND_SQUARE1_REG+1
DecJpFPS:
	bne BranchToDecLength1

Square1SfxHandler:
	ldy Square1SoundQueue
	beq CheckSfx1Buffer
	sty Square1SoundBuffer
	bmi PlaySmallJump
	lsr Square1SoundQueue
	bcs PlayBigJump
	lsr Square1SoundQueue
	bcs PlayBump
	lsr Square1SoundQueue
	bcs PlaySwimStomp
	lsr Square1SoundQueue
	bcs PlaySmackEnemy
	lsr Square1SoundQueue
	bcs PlayPipeDownInj
	lsr Square1SoundQueue
	bcs PlayFireballThrow
	lsr Square1SoundQueue
	bcs PlayFlagpoleSlide
	
CheckSfx1Buffer:
	lda Square1SoundBuffer
	beq ExS1H
	bmi ContinueSndJmp
	lsr
	bcs ContinueSndJmp
	lsr
	bcs ContinueBumpThrow
	lsr
	bcs ContinueSwimStomp
	lsr
	bcs ContinueSmackEnemy
	lsr
	bcs ContinuePipeDownInj
	lsr
	bcs ContinueBumpThrow
	lsr
	bcs DecrementSfx1Length
ExS1H:
	rts

PlaySwimStomp:
	lda #$0e
	sta Squ1_SfxLenCounter
	ldy #%10011100
	ldx #%10011110
	lda #$26
	jsr PlaySqu1Sfx
ContinueSwimStomp:
	ldy Squ1_SfxLenCounter
	lda SwimStompEnvelopeData-1,y
	sta SND_SQUARE1_REG
	cpy #$06
	bne BranchToDecLength1
	lda #$9e
	sta SND_SQUARE1_REG+2
BranchToDecLength1:
	bne DecrementSfx1Length
	
PlaySmackEnemy:
	lda #$0e
	ldy #%11001011
	ldx #%10011111
	sta Squ1_SfxLenCounter
	lda #$28
	jsr PlaySqu1Sfx
	bne DecrementSfx1Length
	
ContinueSmackEnemy:
	ldy Squ1_SfxLenCounter
	cpy #$08
	bne SmSpc
	lda #$a0
	sta SND_SQUARE1_REG+2
	lda #%10011111
	bne SmTick
SmSpc:
	lda #%10010000
SmTick:
	sta SND_SQUARE1_REG
	
DecrementSfx1Length:
	dec Squ1_SfxLenCounter
	bne ExSfx1
StopSquare1Sfx:
	ldx #$00
	stx Square1SoundBuffer
	ldx #%00001110
	stx SND_MASTERCTRL_REG
	ldx #%00001111
	stx SND_MASTERCTRL_REG
ExSfx1:
	rts

PlayPipeDownInj:
	lda #$2f
	sta Squ1_SfxLenCounter
	
ContinuePipeDownInj:
	lda Squ1_SfxLenCounter
	lsr
	bcs NoPDwnL
	lsr
	bcs NoPDwnL
	and #%00000010
	beq NoPDwnL
	ldy #$91
	ldx #%10011010
	lda #%01000100
	jsr PlaySqu1Sfx
NoPDwnL:
	jmp DecrementSfx1Length

;-----------------------------------------------------------------------------------

ExtraLifeFreqData:
	.byte $58, $02, $54, $56, $4e, $44

PowerUpGrabFreqData:
	.byte $4c, $52, $4c, $48, $3e, $36, $3e, $36, $30, $28
	.byte $4a, $50, $4a, $64, $3c, $32, $3c, $32, $2c, $24
	.byte $3a, $64, $3a, $34, $2c, $22, $2c

;residual frequency data (still?)
	.byte $22, $1c, $14

PUp_VGrow_FreqData:
	.byte $14, $04, $22, $24, $16, $04, $24, $26, $18, $04
	.byte $26, $28, $1a, $04, $28, $2a, $1c, $04, $2a, $2c
	.byte $1e, $04, $2c, $2e, $20, $04, $2e, $30, $22, $04
	.byte $30, $32

PlayCoinGrab:
	lda #$35
	ldx #%10001101
	bne CGrab_TTickRegL
	
PlayTimerTick:
	lda #$06
	ldx #%10011000

CGrab_TTickRegL:
	sta Squ2_SfxLenCounter
	ldy #%01111111
	lda #$42
	jsr PlaySqu2Sfx
	
ContinueCGrabTick:
	lda Squ2_SfxLenCounter
	cmp #$30
	bne N2Tone
	lda #$54
	sta SND_SQUARE2_REG+2
N2Tone:
	bne DecrementSfx2Length

PlayBlast:
	lda #$20
	sta Squ2_SfxLenCounter
	ldy #%10010100
	lda #$5e
	bne SBlasJ
	
ContinueBlast:
	lda Squ2_SfxLenCounter
	cmp #$18
	bne DecrementSfx2Length
	ldy #%10010011
	lda #$18
SBlasJ:
	bne BlstSJp

PlayPowerUpGrab:
	lda #$36
	sta Squ2_SfxLenCounter

ContinuePowerUpGrab:
	lda Squ2_SfxLenCounter
	lsr
	bcs DecrementSfx2Length
	tay
	lda PowerUpGrabFreqData-1,y
	ldx #%01011101
	ldy #%01111111

LoadSqu2Regs:
	jsr PlaySqu2Sfx
	
DecrementSfx2Length:
	dec Squ2_SfxLenCounter
	bne ExSfx2

EmptySfx2Buffer:
	ldx #$00
	stx Square2SoundBuffer
	
StopSquare2Sfx:
	ldx #%00001101
	stx SND_MASTERCTRL_REG
	ldx #%00001111
	stx SND_MASTERCTRL_REG
ExSfx2:
	rts

Square2SfxHandler:
	lda Square2SoundBuffer
	and #Sfx_ExtraLife
   	bne ContinueExtraLife
	ldy Square2SoundQueue
	beq CheckSfx2Buffer
	sty Square2SoundBuffer
	bmi PlayKoopaFall
	lsr Square2SoundQueue
	bcs PlayCoinGrab
	lsr Square2SoundQueue
	bcs PlayGrowPowerUp
	lsr Square2SoundQueue
	bcs PlayGrowVine
	lsr Square2SoundQueue
	bcs PlayBlast
	lsr Square2SoundQueue
	bcs PlayTimerTick
	lsr Square2SoundQueue
	bcs PlayPowerUpGrab
	lsr Square2SoundQueue
	bcs PlayExtraLife

CheckSfx2Buffer:
	lda Square2SoundBuffer
	beq ExS2H
	bmi ContinueKoopaFall
	lsr
	bcs Cont_CGrabTick
	lsr
	bcs ContinueGrowItems
	lsr
	bcs ContinueGrowItems
	lsr
	bcs ContinueBlast
	lsr
	bcs Cont_CGrabTick
	lsr
	bcs ContinuePowerUpGrab
	lsr
	bcs ContinueExtraLife
ExS2H:
	rts

Cont_CGrabTick:
	jmp ContinueCGrabTick

JumpToDecLength2:
	jmp DecrementSfx2Length

PlayKoopaFall:
	lda #$38
	sta Squ2_SfxLenCounter
	ldy #%11000100
	lda #$18
BlstSJp:
	bne PBFRegs

ContinueKoopaFall:
	lda Squ2_SfxLenCounter
	cmp #$08
	bne DecrementSfx2Length
	ldy #%10100100
	lda #$5a
PBFRegs:
	ldx #%10011111
EL_LRegs:
	bne LoadSqu2Regs

PlayExtraLife:
	lda #$30
	sta Squ2_SfxLenCounter

ContinueExtraLife:
	lda Squ2_SfxLenCounter
	ldx #$03
DivLLoop:
	lsr
	bcs JumpToDecLength2
	dex
	bne DivLLoop
	tay
	lda ExtraLifeFreqData-1,y
	ldx #%10000010
	ldy #%01111111
    bne EL_LRegs

PlayGrowPowerUp:
	lda #$10
	bne GrowItemRegs

PlayGrowVine:
	lda #$20
	
GrowItemRegs:
	sta Squ2_SfxLenCounter
	lda #%01111111
	sta SND_SQUARE2_REG+1
	lda #$00
	sta Sfx_SecondaryCounter

ContinueGrowItems:
	inc Sfx_SecondaryCounter
	lda Sfx_SecondaryCounter
	lsr
	tay
	cpy Squ2_SfxLenCounter
	beq StopGrowItems
	lda #%10011101
	sta SND_SQUARE2_REG
	lda PUp_VGrow_FreqData,y
	jsr SetFreq_Squ2
	rts

StopGrowItems:
	jmp EmptySfx2Buffer

;-----------------------------------------------------------------------------------

WindSfxData:
	.byte $37, $46, $55, $64, $74, $83, $93, $a2, $b1
	.byte $c0, $d0, $e0, $f1, $f1, $f2, $e2, $e2, $c3
	.byte $a3, $84, $64, $44, $35, $25

BrickShatterFreqData:
	.byte $01, $0e, $0e, $0d, $0b, $06, $0c, $0f, $0a
	.byte $09, $03, $0d, $08, $0d, $06, $0c

SlideFreqData:
	.byte $47, $49, $42, $4a, $43, $4b

PlaySlide:
	sty NoiseSoundBuffer
	lda #$06
	sta Noise_SfxLenCounter

ContinueSlide:
	lda Noise_SfxLenCounter
	tay
	lda SlideFreqData-1,y
	sta SND_TRIANGLE_REG+2
	lda #%00011000
	sta SND_TRIANGLE_REG
	sta SND_TRIANGLE_REG+3
	bne DecrementSfx3Length

PlayBrickShatter:
	sty NoiseSoundBuffer
	lda #$20
	sta Noise_SfxLenCounter

ContinueBrickShatter:
	lda Noise_SfxLenCounter
	lsr
	bcc DecrementSfx3Length
	tay
	ldx BrickShatterFreqData,y
	lda BrickShatterEnvData,y

PlayNoiseSfx:
	sta SND_NOISE_REG
	stx SND_NOISE_REG+2
	lda #%00011000
	sta SND_NOISE_REG+3

DecrementSfx3Length:
	dec Noise_SfxLenCounter
	bne ExSfx3
	lda #%11110000
	sta SND_NOISE_REG
	lda #$00
	sta SND_TRIANGLE_REG
	lda #$00
	sta NoiseSoundBuffer
ExSfx3:
	rts

NoiseTriSfxHandler:
	lda NoiseSoundBuffer
	bmi ContinueSlide
	ldy NoiseSoundQueue
	bmi PlaySlide
	lsr NoiseSoundQueue
	bcs PlayBrickShatter
	lsr
	bcs ContinueBrickShatter
	lsr NoiseSoundQueue
	bcs PlayKoopaFlame
	lsr
	bcs ContinueKoopaFlame
	lsr
	bcs ContinueWind
	lsr NoiseSoundQueue
	bcs PlayWind
ExNH:
	rts

PlayKoopaFlame:
	sty NoiseSoundBuffer
	lda #$40
	sta Noise_SfxLenCounter

ContinueKoopaFlame:
	lda Noise_SfxLenCounter
	lsr
	tay
	ldx #%00001111
	lda KoopaFlameEnvData-1,y

PlayWSfx:
	bne PlayNoiseSfx

PlayWind:
	sty NoiseSoundBuffer
	lda #$c0
	sta Noise_SfxLenCounter
	
ContinueWind:
	lsr NoiseSoundQueue
	bcc ExSfx3
	lda Noise_SfxLenCounter
	lsr
	lsr
	lsr
	tay
	lda WindSfxData,y
	and #%00001111
	ora #%00010000
	tax
	lda WindSfxData,y
	lsr
	lsr
	lsr
	lsr
	ora #%00010000
	bne PlayWSfx

;-----------------------------------------------------------------------------------

ContinueMusic:
	jmp HandleSquare2Music

MusicHandler:
	lda EventMusicQueue
	bne LoadEventMusic
	lda AreaMusicQueue
	bne LoadAreaMusic
	lda EventMusicBuffer
	ora AreaMusicBuffer
	bne ContinueMusic
	rts

LoadEventMusic:
	sta EventMusicBuffer
	cmp #DeathMusic
	bne NoStopSfx
	jsr StopSquare1Sfx
	jsr StopSquare2Sfx
NoStopSfx:
	ldx AreaMusicBuffer
	stx AreaMusicBuffer_Alt
	ldy #$00
	sty NoteLengthTblAdder
	sty AreaMusicBuffer
	cmp #TimeRunningOutMusic
	bne FindEventMusicHeader
	ldx #$08
	stx NoteLengthTblAdder
	bne FindEventMusicHeader
	
LoadAreaMusic:
	cmp #UndergroundMusic
	bne NoStop1
	jsr StopSquare1Sfx
NoStop1:
	ldy #(StartGroundHdrs - MHD)
GMLoopB:
	sty GroundMusicHeaderOfs
	
HandleAreaMusicLoopB:
	ldy #$00
	sty EventMusicBuffer
	sta AreaMusicBuffer
	cmp #GroundMusic
	bne FindAreaMusicHeader
	inc GroundMusicHeaderOfs
	ldy GroundMusicHeaderOfs
	cpy #(EndGroundHdrs - MHD + 1)
	bne LoadHeader
	ldy #(GroundLoopbackSectionStart - MHD)
	bne GMLoopB
	
FindAreaMusicHeader:
	ldy #(AreaMusicHdrs - MHD)
	sty MusicOffset_Square2
	
FindEventMusicHeader:
	iny
	lsr
	bcc FindEventMusicHeader
	
LoadHeader:
	lda MHD-1,y
	tay
	lda MHD,y
	sta NoteLenLookupTblOfs
	lda MHD+1,y
	sta MusicDataLow
	lda MHD+2,y
	sta MusicDataHigh
	lda MHD+3,y
	sta MusicOffset_Triangle
	lda MHD+4,y
	sta MusicOffset_Square1
	lda MHD+5,y
	sta MusicOffset_Noise
	sta NoiseDataLoopbackOfs
	lda #$01
	sta Squ2_NoteLenCounter
	sta Squ1_NoteLenCounter
	sta Tri_NoteLenCounter
	sta Noise_BeatLenCounter
	lda #$00
	sta MusicOffset_Square2
	sta AltRegContentFlag
	lda #%00001011
	sta SND_MASTERCTRL_REG
	lda #%00001111
	sta SND_MASTERCTRL_REG

HandleSquare2Music:
	dec Squ2_NoteLenCounter
	bne MiscSqu2MusicTasks
	ldy MusicOffset_Square2
	inc MusicOffset_Square2
	lda (MusicData),y
	beq EndOfMusicData
	bpl Squ2NoteHandler
	bne Squ2LengthHandler
	
EndOfMusicData:
	lda EventMusicBuffer
	cmp #TimeRunningOutMusic
	bne NotTRO
	lda AreaMusicBuffer_Alt
	bne MusicLoopBack
NotTRO:
	and #VictoryMusic                  ;residual from Mario 1
	bne VictoryMLoopBack_Old
	lda AreaMusicBuffer
	and #(%01111111 ^ PipeIntroMusic)
	bne MusicLoopBack
	lda #$00
	sta AreaMusicBuffer
	sta EventMusicBuffer
	sta SND_TRIANGLE_REG
	lda #%10010000
	sta SND_SQUARE1_REG
	sta SND_SQUARE2_REG
	rts

MusicLoopBack:
	jmp HandleAreaMusicLoopB

VictoryMLoopBack_Old:                 ;residual from Mario 1
	jmp LoadEventMusic

Squ2LengthHandler:
	jsr ProcessLengthData
	sta Squ2_NoteLenBuffer
	ldy MusicOffset_Square2
	inc MusicOffset_Square2
	lda (MusicData),y
	
Squ2NoteHandler:
	ldx Square2SoundBuffer
	bne SkipFqL1
	jsr SetFreq_Squ2
	beq Rest
	jsr LoadControlRegs
Rest:
	sta Squ2_EnvelopeDataCtrl
	jsr Dump_Squ2_Regs
SkipFqL1:
	lda Squ2_NoteLenBuffer
	sta Squ2_NoteLenCounter
	
MiscSqu2MusicTasks:
	lda Square2SoundBuffer
	bne HandleSquare1Music
	lda EventMusicBuffer
	and #(Silence | AltGameOverMusic | DeathMusic)
	bne HandleSquare1Music
	ldy Squ2_EnvelopeDataCtrl
	beq NoDecEnv1
	dec Squ2_EnvelopeDataCtrl
NoDecEnv1:
	jsr LoadEnvelopeData
	sta SND_SQUARE2_REG
	ldx #%01111111
	stx SND_SQUARE2_REG+1
	
HandleSquare1Music:
	ldy MusicOffset_Square1
	beq HandleTriangleMusic
	dec Squ1_NoteLenCounter
	bne MiscSqu1MusicTasks
	
FetchSqu1MusicData:
	ldy MusicOffset_Square1
	inc MusicOffset_Square1
	lda (MusicData),y
	bne Squ1NoteHandler
	lda #%10000011
	sta SND_SQUARE1_REG
	lda #%10010100
	sta SND_SQUARE1_REG+1
	sta AltRegContentFlag
	bne FetchSqu1MusicData

Squ1NoteHandler:
	jsr AlternateLengthHandler
	sta Squ1_NoteLenCounter
	ldy Square1SoundBuffer
	bne HandleTriangleMusic
	txa
	and #%00111110
	jsr SetFreq_Squ1
	beq SkipCtrlL
	jsr LoadControlRegs
SkipCtrlL
	sta $07b7
	jsr Dump_Squ1_Regs

MiscSqu1MusicTasks:
	lda Square1SoundBuffer
	bne HandleTriangleMusic
	lda EventMusicBuffer
	and #(Silence | AltGameOverMusic | DeathMusic)
	bne DeathMAltReg
	ldy Squ1_EnvelopeDataCtrl
	beq NoDecEnv2
	dec Squ1_EnvelopeDataCtrl
NoDecEnv2:
	jsr LoadEnvelopeData
	sta SND_SQUARE1_REG
DeathMAltReg:
	lda AltRegContentFlag
	bne DoAltLoad
	lda #%01111111
DoAltLoad:
	sta SND_SQUARE1_REG+1

HandleTriangleMusic:
	lda MusicOffset_Triangle
	dec Tri_NoteLenCounter
	bne HandleNoiseMusic
	ldy MusicOffset_Triangle
	inc MusicOffset_Triangle
	lda (MusicData),y
	beq LoadTriCtrlReg
	bpl TriNoteHandler
	jsr ProcessLengthData
	sta Tri_NoteLenBuffer
	lda #%00011111
	sta SND_TRIANGLE_REG
	ldy MusicOffset_Triangle
	inc MusicOffset_Triangle
	lda (MusicData),y
	beq LoadTriCtrlReg

TriNoteHandler:
	jsr SetFreq_Tri
	ldx Tri_NoteLenBuffer
	stx Tri_NoteLenCounter
	lda EventMusicBuffer
	and #(%01111111 ^ AltGameOverMusic ^ DeathMusic)
	bne NotDOrD4
	lda AreaMusicBuffer
	and #(CastleMusic | WaterMusic)
	beq HandleNoiseMusic
NotDOrD4:
	txa
	cmp #$12
	bcs LongN
	lda EventMusicBuffer
	and #EndOfCastleMusic
	beq MediN
	lda #%00001111
	bne LoadTriCtrlReg
MediN:
	lda #%00011111
	bne LoadTriCtrlReg
LongN:
	lda #%11111111

LoadTriCtrlReg:
	sta $4008

HandleNoiseMusic:
	lda AreaMusicBuffer
	and #(%11111111 ^ UndergroundMusic ^ CastleMusic)
	beq ExitMusicHandler
	dec Noise_BeatLenCounter
	bne ExitMusicHandler

FetchNoiseBeatData:
	ldy MusicOffset_Noise
	inc MusicOffset_Noise
	lda (MusicData),y
	bne NoiseBeatHandler
	lda NoiseDataLoopbackOfs
	sta MusicOffset_Noise
	bne FetchNoiseBeatData

NoiseBeatHandler:
	jsr AlternateLengthHandler
	sta Noise_BeatLenCounter
	txa
	and #%00111110
	beq SilentBeat
	cmp #%00110000
	beq LongBeat
	cmp #%00100000
	beq StrongBeat
	and #%00010000
	beq SilentBeat
	lda #%00011100
	ldx #$03
	ldy #$18
	bne PlayBeat

StrongBeat:
	lda #%00011100
	ldx #$0c
	ldy #$18
	bne PlayBeat

LongBeat:
	lda #%00011100
	ldx #$03
	ldy #$58
	bne PlayBeat

SilentBeat:
	lda #%00010000
	
PlayBeat:
	sta $400c
	stx $400e
	sty $400f

ExitMusicHandler:
	rts

AlternateLengthHandler:
	tax
	ror
	txa
	rol
	rol
	rol
	
ProcessLengthData:
	and #%00000111
	clc
	adc NoteLenLookupTblOfs
	adc NoteLengthTblAdder
	tay
	lda MLLT,y
	rts

LoadControlRegs:
	lda EventMusicBuffer
	and #EndOfCastleMusic
	beq NotECstlM
	lda #$04
	bne AllMus
NotECstlM:
	lda AreaMusicBuffer
	and #(%01111111 ^ WaterMusic)
	beq WaterMus
	lda #$08
	bne AllMus
WaterMus:
	lda #$28
AllMus:
	ldx #%10000010
	ldy #%01111111
	rts

LoadEnvelopeData:
	lda EventMusicBuffer
	and #EndOfCastleMusic
	beq LoadUsualEnvData
	lda EndOfCastleMusicEnvData,y
	rts

LoadUsualEnvData:
	lda AreaMusicBuffer
	and #(%01111111 ^ WaterMusic)
	beq LoadWaterEventMusEnvData
	lda AreaMusicEnvData,y
	rts

LoadWaterEventMusEnvData:
	lda WaterEventMusEnvData,y
	rts

;------------------------------------------------------------------------------------

MHD:
MusicHeaderData:
	.byte DeathMusHdr-MHD           ;event music
	.byte GameOverMusHdr-MHD
	.byte VicMusOldHdr-MHD
	.byte WinCastleMusHdr-MHD
	.byte GameOverMusHdr-MHD
	.byte EndOfLevelMusHdr-MHD
	.byte TimeRunningOutHdr-MHD
	.byte SilenceHdr-MHD

AreaMusicHdrs:
	.byte GMPart1Hdr-MHD   ;area music
	.byte WaterMusHdr-MHD
	.byte UndergroundMusHdr-MHD
	.byte CastleMusHdr-MHD
	.byte Star_CloudHdr-MHD
	.byte GMLeadInHdr-MHD
	.byte Star_CloudHdr-MHD
	.byte SilenceHdr-MHD

StartGroundHdrs:
	.byte GMLeadInHdr-MHD  ;ground level music layout
GroundLoopbackSectionStart:
	.byte GMPart1Hdr-MHD, GMPart1Hdr-MHD
	.byte GMPart2AHdr-MHD, GMPart2BHdr-MHD, GMPart2AHdr-MHD, GMPart2CHdr-MHD
	.byte GMPart2AHdr-MHD, GMPart2BHdr-MHD, GMPart2AHdr-MHD, GMPart2CHdr-MHD
	.byte GMPart3AHdr-MHD, GMPart3BHdr-MHD, GMPart3AHdr-MHD, GMLeadInHdr-MHD
	.byte GMPart1Hdr-MHD, GMPart1Hdr-MHD
	.byte GMPart4AHdr-MHD, GMPart4BHdr-MHD, GMPart4AHdr-MHD, GMPart4CHdr-MHD
	.byte GMPart4AHdr-MHD, GMPart4BHdr-MHD, GMPart4AHdr-MHD, GMPart4CHdr-MHD
	.byte GMPart3AHdr-MHD, GMPart3BHdr-MHD, GMPart3AHdr-MHD, GMLeadInHdr-MHD
	.byte GMPart4AHdr-MHD, GMPart4BHdr-MHD, GMPart4AHdr-MHD, GMPart4CHdr-MHD
EndGroundHdrs

TimeRunningOutHdr:
	.byte MLLT1 - MLLT
	.word TROMData
	.byte TROM_Tri - TROMData, TROM_Sq1 - TROMData
Star_CloudHdr:
	.byte MLLT4 - MLLT
	.word SCMData
	.byte SCM_Tri - SCMData, SCM_Sq1 - SCMData, SCM_Nos - SCMData
EndOfLevelMusHdr:
	.byte MLLT4 - MLLT
	.word WLMData
	.byte WLM_Tri - WLMData, WLM_Sq1 - WLMData
ResidualMusHdr:
	.byte MLLT4 - MLLT
	.word RMData
	.byte RM_Tri - RMData, RM_Sq1 - RMData
UndergroundMusHdr:
	.byte MLLT3 - MLLT
	.word UMData
	.byte 0, 0
SilenceHdr:
	.byte MLLT1 - MLLT
	.word SilenceData
	.byte 0
CastleMusHdr:
	.byte MLLT0 - MLLT
	.word CMData
	.byte CM_Tri - CMData, CM_Sq1 - CMData
VicMusOldHdr:
GameOverMusHdr:
	.byte MLLT3 - MLLT
	.word GOMData
	.byte GOM_Tri - GOMData, GOM_Sq1 - GOMData
WaterMusHdr:
	.byte MLLT1 - MLLT
	.word WaterMData
	.byte WaterM_Tri - WaterMData, WaterM_Sq1 - WaterMData, WaterM_Nos - WaterMData
WinCastleMusHdr:
	.byte MLLT1 - MLLT
	.word EOCMData
	.byte EOCM_Tri - EOCMData, EOCM_Sq1 - EOCMData
GMPart1Hdr:
	.byte MLLT3 - MLLT
	.word GMP1Data
	.byte GMP1_Tri - GMP1Data, GMP1_Sq1 - GMP1Data, GMP12_Nos - GMP1Data
GMPart2AHdr:
	.byte MLLT3 - MLLT
	.word GMP2AData
	.byte GMP2A_Tri - GMP2AData, GMP2A_Sq1 - GMP2AData, GMP12_Nos - GMP2AData
GMPart2BHdr:
	.byte MLLT3 - MLLT
	.word GMP2BData
	.byte GMP2B_Tri - GMP2BData, GMP2B_Sq1 - GMP2BData, GMP12_Nos - GMP2BData
GMPart2CHdr:
	.byte MLLT3 - MLLT
	.word GMP2CData
	.byte GMP2C_Tri - GMP2CData, GMP2C_Sq1 - GMP2CData, GMP12_Nos - GMP2CData
GMPart3AHdr:
	.byte MLLT3 - MLLT
	.word GMP3AData
	.byte GMP3_Tri - GMP3AData, GMP3A_Sq1 - GMP3AData, GMP3L_Nos - GMP3AData
GMPart3BHdr:
	.byte MLLT3 - MLLT
	.word GMP3BData
	.byte GMP3_Tri - GMP3BData, GMP3B_Sq1 - GMP3BData, GMP3L_Nos - GMP3BData
GMLeadInHdr:
	.byte MLLT3 - MLLT
	.word GMLdData
	.byte GMLd_Tri - GMLdData, GMLd_Sq1 - GMLdData, GMP3L_Nos - GMLdData
GMPart4AHdr:
	.byte MLLT3 - MLLT
	.word GMP4AData
	.byte GMP4A_Tri - GMP4AData, GMP4A_Sq1 - GMP4AData, GMP4DM_Nos - GMP4AData
GMPart4BHdr:
	.byte MLLT3 - MLLT
	.word GMP4BData
	.byte GMP4B_Tri - GMP4BData, GMP4B_Sq1 - GMP4BData, GMP4DM_Nos - GMP4BData
GMPart4CHdr:
	.byte MLLT3 - MLLT
	.word GMP4CData
	.byte GMP4C_Tri - GMP4CData, GMP4C_Sq1 - GMP4CData, GMP4DM_Nos - GMP4CData
DeathMusHdr:
	.byte MLLT3 - MLLT
	.word DMData
	.byte DM_Tri - DMData, DM_Sq1 - DMData, GMP4DM_Nos - DMData

;------------------------------------------------------------------------------------

SCMData:
	.byte $84, $2c, $2c, $2c, $82, $04, $2c, $04, $85, $2c, $84, $2c
	.byte $2c, $2a, $2a, $2a, $82, $04, $2a, $04, $85, $2a, $84, $2a
	.byte $2a, $00
SCM_Sq1:
	.byte $1f, $1f, $1f, $98, $1f, $1f, $98, $9e, $98, $1f, $1d, $1d
	.byte $1d, $94, $1d, $1d, $94, $9c, $94, $1d
SCM_Tri:
	.byte $86, $18, $85, $26, $30, $84, $04, $26, $30, $86, $14, $85
	.byte $22, $2c, $84, $04, $22, $2c
SCM_Nos:
	.byte $21, $d0, $c4, $d0, $31, $d0, $c4, $d0, $00

GMP1Data:
	.byte $85, $2c, $22, $1c, $84, $26, $2a, $82, $28, $26, $04, $87
	.byte $22, $34, $3a, $82, $40, $04, $36, $84, $3a, $34, $82, $2c
	.byte $30, $85, $2a
	
SilenceData:
	.byte $00
	
GMP1_Sq1:
	.byte $5d, $55, $4d, $15, $19, $96, $15, $d5, $e3, $eb, $2d, $a6
	.byte $2b, $27, $9c, $9e, $59
GMP1_Tri:
	.byte $85, $22, $1c, $14, $84, $1e, $22, $82, $20, $1e, $04, $87
	.byte $1c, $2c, $34, $82, $36, $04, $30, $34, $04, $2c, $04, $26
	.byte $2a, $85, $22

GMP2AData:
	.byte $84, $04, $82, $3a, $38, $36, $32, $04, $34, $04, $24, $26
	.byte $2c, $04, $26, $2c, $30, $00
GMP2A_Sq1:
	.byte $05, $b4, $b2, $b0, $2b, $ac, $84, $9c, $9e, $a2, $84, $94
	.byte $9c, $9e
GMP2A_Tri:
	.byte $85, $14, $22, $84, $2c, $85, $1e, $82, $2c, $84, $2c, $1e

GMP2BData:
	.byte $84, $04, $82, $3a, $38, $36, $32, $04, $34, $04, $64, $04
	.byte $64, $86, $64, $00
GMP2B_Sq1:
	.byte $05, $b4, $b2, $b0, $2b, $ac, $84, $37, $b6, $b6, $45
GMP2B_Tri:
	.byte $85, $14, $1c, $82, $22, $84, $2c, $4e, $82, $4e, $84, $4e
	.byte $22

GMP2CData:
	.byte $84, $04, $85, $32, $85, $30, $86, $2c, $04, $00
GMP2C_Sq1:
	.byte $05, $a4, $05, $9e, $05, $9d, $85
GMP2C_Tri:
	.byte $84, $14, $85, $24, $28, $2c, $82, $22, $84, $22, $14
GMP12_Nos:
	.byte $21, $d0, $c4, $d0, $31, $d0, $c4, $d0, $00

GMP3AData:
	.byte $82, $2c, $84, $2c, $2c, $82, $2c, $30, $04, $34, $2c, $04
	.byte $26, $86, $22, $00
GMP3A_Sq1:
	.byte $a4, $25, $25, $a4, $29, $a2, $1d, $9c, $95

GMP3BData:
	.byte $82, $2c, $2c, $04, $2c, $04, $2c, $30, $85, $34, $04, $04
	.byte $00
GMP3B_Sq1:
	.byte $a4, $25, $25, $a4, $a8, $63, $04
GMP3_Tri:
	.byte $85, $0e, $1a, $84, $24, $85, $22, $14, $84, $0c

GMLdData:
	.byte $82, $34, $84, $34, $34, $82, $2c, $84, $34, $86, $3a, $04
	.byte $00
GMLd_Sq1:
	.byte $a0, $21, $21, $a0, $21, $2b, $05, $a3
GMLd_Tri:
	.byte $82, $18, $84, $18, $18, $82, $18, $18, $04, $86, $3a, $22
GMP3L_Nos:
	.byte $31, $90, $31, $90, $31, $71, $31, $90, $90, $90, $00

GMP4AData:
	.byte $82, $34, $84, $2c, $85, $22, $84, $24, $82, $26, $36, $04
	.byte $36, $86, $26, $00
GMP4A_Sq1:
	.byte $ac, $27, $5d, $1d, $9e, $2d, $ac, $9f
GMP4A_Tri:
	.byte $85, $14, $82, $20, $84, $22, $2c, $1e, $1e, $82, $2c, $2c
	.byte $1e, $04

GMP4BData:
	.byte $87, $2a, $40, $40, $40, $3a, $36, $82, $34, $2c, $04, $26
	.byte $86, $22, $00
GMP4B_Sq1:
	.byte $e3, $f7, $f7, $f7, $f5, $f1, $ac, $27, $9e, $9d
GMP4B_Tri:
	.byte $85, $18, $82, $1e, $84, $22, $2a, $22, $22, $82, $2c, $2c
	.byte $22, $04

DMData:
	.byte $86, $04

GMP4CData:
	.byte $82, $2a, $36, $04, $36, $87, $36, $34, $30, $86, $2c, $04
	.byte $00
DM_Sq1:
	.byte $00, $68, $6a, $6c, $45
GMP4C_Sq1:
	.byte $a2, $31, $b0, $f1, $ed, $eb, $a2, $1d, $9c, $95
DM_Tri:
	.byte $86, $04
GMP4C_Tri:
	.byte $85, $22, $82, $22, $87, $22, $26, $2a, $84, $2c, $22, $86
	.byte $14
GMP4DM_Nos:
	.byte $51, $90, $31, $11, $00

CMData:
	.byte $80, $22, $28, $22, $26, $22, $24, $22, $26, $22, $28, $22
	.byte $2a, $22, $28, $22, $26, $22, $28, $22, $26, $22, $24, $22
	.byte $26, $22, $28, $22, $2a, $22, $28, $22, $26, $20, $26, $20
	.byte $24, $20, $26, $20, $28, $20, $26, $20, $28, $20, $26, $20
	.byte $24, $20, $26, $20, $24, $20, $26, $20, $28, $20, $26, $20
	.byte $28, $20, $26, $20, $24, $28, $30, $28, $32, $28, $30, $28
	.byte $2e, $28, $30, $28, $2e, $28, $2c, $28, $2e, $28, $30, $28
	.byte $32, $28, $30, $28, $2e, $28, $30, $28, $2e, $28, $2c, $28
	.byte $2e, $00
CM_Sq1:
	.byte $04, $70, $6e, $6c, $6e, $70, $72, $70, $6e, $70, $6e, $6c
	.byte $6e, $70, $72, $70, $6e, $6e, $6c, $6e, $70, $6e, $70, $6e
	.byte $6c, $6e, $6c, $6e, $70, $6e, $70, $6e, $6c, $76, $78, $76
	.byte $74, $76, $74, $72, $74, $76, $78, $76, $74, $76, $74, $72
	.byte $74
CM_Tri:
	.byte $84, $1a, $83, $18, $20, $84, $1e, $83, $1c, $28, $26, $1c
	.byte $1a, $1c

GOMData:
	.byte $82, $2c, $04, $04, $22, $04, $04, $84, $1c, $87, $26, $2a
	.byte $26, $84, $24, $28, $24, $80, $22, $00
GOM_Sq1:
	.byte $9c, $05, $94, $05, $0d, $9f, $1e, $9c, $98, $9d
GOM_Tri:
	.byte $82, $22, $04, $04, $1c, $04, $04, $84, $14, $86, $1e, $80
	.byte $16, $80, $14

TROMData:
	.byte $81, $1c, $30, $04, $30, $30, $04, $1e, $32, $04, $32, $32
	.byte $04, $20, $34, $04, $34, $34, $04, $36, $04, $84, $36, $00
TROM_Sq1:
	.byte $46, $a4, $64, $a4, $48, $a6, $66, $a6, $4a, $a8, $68, $a8
	.byte $6a, $44, $2b
TROM_Tri:
	.byte $81, $2a, $42, $04, $42, $42, $04, $2c, $64, $04, $64, $64
	.byte $04, $2e, $46, $04, $46, $46, $04, $22, $04, $84, $22

WLMData:
	.byte $87, $04, $06, $0c, $14, $1c, $22, $86, $2c, $22, $87, $04
	.byte $60, $0e, $14, $1a, $24, $86, $2c, $24

RMData:
	.byte $87, $04, $08, $10, $18, $1e, $28, $86, $30, $30, $80, $64
	.byte $00
WLM_Sq1:
	.byte $cd, $d5, $dd, $e3, $ed, $f5, $bb, $b5, $cf, $d5, $db, $e5
	.byte $ed, $f3, $bd, $b3
RM_Sq1:
	.byte $d1, $d9, $df, $e9, $f1, $f7, $bf, $ff, $ff, $ff, $34, $00
WLM_Tri:
	.byte $86, $04, $87, $14, $1c, $22, $86, $34, $84, $2c, $04, $04
	.byte $04, $87, $14, $1a, $24, $86, $32, $84, $2c, $04
RM_Tri:
	.byte $86, $04, $87, $18, $1e, $28, $86, $36, $87, $30, $30, $30
	.byte $80, $2c

UMData:
	.byte $82, $14, $2c, $62, $26, $10, $28, $80, $04, $82, $14, $2c
	.byte $62, $26, $10, $28, $80, $04, $82, $08, $1e, $5e, $18, $60
	.byte $1a, $80, $04, $82, $08, $1e, $5e, $18, $60, $1a, $86, $04
	.byte $83, $1a, $18, $16, $84, $14, $1a, $18, $0e, $0c, $16, $83
	.byte $14, $20, $1e, $1c, $28, $26, $87, $24, $1a, $12, $10, $62
	.byte $0e, $80, $04, $04, $00

WaterMData:
	.byte $82, $18, $1c, $20, $22, $26, $28, $81, $2a, $2a, $2a, $04
	.byte $2a, $04, $83, $2a, $82, $22, $86, $34, $32, $34, $81, $04
	.byte $22, $26, $2a, $2c, $30, $86, $34, $83, $32, $82, $36, $84
	.byte $34, $85, $04, $81, $22, $86, $30, $2e, $30, $81, $04, $22
	.byte $26, $2a, $2c, $2e, $86, $30, $83, $22, $82, $36, $84, $34
	.byte $85, $04, $81, $22, $86, $3a, $3a, $3a, $82, $3a, $81, $40
	.byte $82, $04, $81, $3a, $86, $36, $36, $36, $82, $36, $81, $3a
	.byte $82, $04, $81, $36, $86, $34, $82, $26, $2a, $36, $81, $34
	.byte $34, $85, $34, $81, $2a, $86, $2c, $00
WaterM_Nos:
	.byte $84, $90, $b0, $84, $50, $50, $b0, $00
WaterM_Sq1:
	.byte $98, $96, $94, $92, $94, $96, $58, $58, $58, $44, $5c, $44
	.byte $9f, $a3, $a1, $a3, $85, $a3, $e0, $a6, $23, $c4, $9f, $9d
	.byte $9f, $85, $9f, $d2, $a6, $23, $c4, $b5, $b1, $af, $85, $b1
	.byte $af, $ad, $85, $95, $9e, $a2, $aa, $6a, $6a, $6b, $5e, $9d
WaterM_Tri:
	.byte $84, $04, $04, $82, $22, $86, $22, $82, $14, $22, $2c, $12
	.byte $22, $2a, $14, $22, $2c, $1c, $22, $2c, $14, $22, $2c, $12
	.byte $22, $2a, $14, $22, $2c, $1c, $22, $2c, $18, $22, $2a, $16
	.byte $20, $28, $18, $22, $2a, $12, $22, $2a, $18, $22, $2a, $12
	.byte $22, $2a, $14, $22, $2c, $0c, $22, $2c, $14, $22, $34, $12
	.byte $22, $30, $10, $22, $2e, $16, $22, $34, $18, $26, $36, $16
	.byte $26, $36, $14, $26, $36, $12, $22, $36, $5c, $22, $34, $0c
	.byte $22, $22, $81, $1e, $1e, $85, $1e, $81, $12, $86, $14

EOCMData:
	.byte $81, $2c, $22, $1c, $2c, $22, $1c, $85, $2c, $04, $81, $2e
	.byte $24, $1e, $2e, $24, $1e, $85, $2e, $04, $81, $32, $28, $22
	.byte $32, $28, $22, $85, $32, $87, $36, $36, $36, $84, $3a, $00
EOCM_Sq1:
	.byte $5c, $54, $4c, $5c, $54, $4c, $5c, $1c, $1c, $5c, $5c, $5c
	.byte $5c, $5e, $56, $4e, $5e, $56, $4e, $5e, $1e, $1e, $5e, $5e
	.byte $5e, $5e, $62, $5a, $50, $62, $5a, $50, $62, $22, $22, $62
	.byte $e7, $e7, $e7, $2b
EOCM_Tri:
	.byte $86, $14, $81, $14, $80, $14, $14, $81, $14, $14, $14, $14
	.byte $86, $16, $81, $16, $80, $16, $16, $81, $16, $16, $16, $16
	.byte $81, $28, $22, $1a, $28, $22, $1a, $28, $80, $28, $28, $81
	.byte $28, $87, $2c, $2c, $2c, $84, $30

    .byte 255
    
;------------------------------------------------------------------------------------

FreqRegLookupTbl:
	.byte $00, $88, $00, $2f, $00, $00, $02, $a6
	.byte $02, $80, $02, $5c, $02, $3a, $02, $1a
	.byte $01, $df, $01, $c4, $01, $ab, $01, $93
	.byte $01, $7c, $01, $67, $01, $53, $01, $40
	.byte $01, $2e, $01, $1d, $01, $0d, $00, $fe
	.byte $00, $ef, $00, $e2, $00, $d5, $00, $c9
	.byte $00, $be, $00, $b3, $00, $a9, $00, $a0
	.byte $00, $97, $00, $8e, $00, $86, $00, $77
	.byte $00, $7e, $00, $71, $00, $54, $00, $64
	.byte $00, $5f, $00, $59, $00, $50, $00, $47
	.byte $00, $43, $00, $3b, $00, $35, $00, $2a
	.byte $00, $23, $04, $75, $03, $57, $02, $f9
	.byte $02, $cf, $01, $fc, $00, $6a

MLLT:
MusicLengthLookupTbl:
MLLT0:
	.byte $05, $0a, $14, $28, $50, $1e, $3c, $02
MLLT1:
	.byte $04, $08, $10, $20, $40, $18, $30, $0c
MLLT2:
	.byte $03, $06, $0c, $18, $30, $12, $24, $08
MLLT3:
	.byte $36, $03, $09, $06, $12, $1b, $24, $0c
MLLT4:
	.byte $24, $02, $06, $04, $0c, $12, $18, $08
MLLT5:
	.byte $12, $01, $03, $02, $06, $09, $0c, $04

EndOfCastleMusicEnvData:
	.byte $98, $99, $9a, $9b

AreaMusicEnvData:
	.byte $90, $94, $94, $95, $95, $96, $97, $98

WaterEventMusEnvData:
	.byte $90, $91, $92, $92, $93, $93, $93, $94
	.byte $94, $94, $94, $94, $94, $95, $95, $95
	.byte $95, $95, $95, $96, $96, $96, $96, $96
	.byte $96, $96, $96, $96, $96, $96, $96, $96
	.byte $96, $96, $96, $96, $95, $95, $94, $93

KoopaFlameEnvData:
	.byte $15, $16, $16, $17, $17, $18, $19, $19
	.byte $1a, $1a, $1c, $1d, $1d, $1e, $1e, $1f
	.byte $1f, $1f, $1f, $1e, $1d, $1c, $1e, $1f
	.byte $1f, $1e, $1d, $1c, $1a, $18, $16, $14

BrickShatterEnvData:
	.byte $15, $16, $16, $17, $17, $18, $19, $19
	.byte $1a, $1a, $1c, $1d, $1d, $1e, $1e, $1f
    
;vectors
	.word NMI, Start, IRQ

